/*
 * Stellarium360 astronomy simulation and visualization
 *
 * Copyright (C) 2006 Fabien Chereau
 * Copyright (C) 2009, 2010 Digitalis Education Solutions, Inc.
 * Copyright (C) 2013 of the LSS team
 * Copyright (C) 2014 of the LSS Team & Association Sirius
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Stellarium360 is a free open project of the LSS team
 * See the TRADEMARKS file for free open project usage requirements.
 *
 */

#include "app.hpp"
#include "viewport_distorter.hpp"
#include "stellastro.h"
#include "utility.hpp"
#include "callbacks.hpp"
//#include "shared_data.hpp"
//#include <fastdb/fastdb.h>
#include <exception>
#include <Magick++.h>
#include <iomanip>
//#include <boost/function.hpp>

#include <fstream>
#include <iostream>
#include <sstream>
#include <string>

#include "log.hpp"



//REGISTER(ObjectRecord);
//REGISTER(TZRecord);

//~ static ConnectCallback connectCb("Bowen");

App::App( SDLFacade* const sdl ) :
		frame(0), timefr(0), timeBase(0), fps(0), maxfps(10000.f),  FlagTimePause(0),
		is_mouse_moving_horiz(false), is_mouse_moving_vert(false), /*draw_mode(App::DM_NONE),*/
		initialized(0), GMT_shift(0), m_OutputFrameNumber(0) , flagMasterput(false)
{
	Magick::InitializeMagick(NULL);
	// Ensure shared data structures are initialized early
	//~ SharedData::Instance();
	//~ AudioPlayer::Instance();

	// Construct an a function object to be called when/if a connection to cover light controller succeeds
	//~ boost::function<void(void)> f = connectCb;
	//~ NamedSockets::Instance().SetConnectCallback( f );
	//~ if( AppSettings::Instance()->Digitarium() )
		//~ NamedSockets::Instance().CreateOnCurrentSubnet( "Bowen", "245", 6005 );
	//stellState=new StellState();
	m_sdl = sdl;
	flagEnableDragMouse = false;
	window=m_sdl->getWindow();
	SelectedScript = SelectedScriptDirectory = "";
	AppSettings* settings = AppSettings::Instance();
	core = new Core(sdl, settings->getlDir(), settings->getDataRoot(), mBoost::callback<void, string>(this, &App::recordCommand));
	ui = new UI(core, this);
//	message = new Message_Mgr();
	commander = new AppCommandInterface(core, this);
	audio= new Audio();
	scripts = new ScriptMgr(commander, core->getUserDir(), audio /*, stellState*/);
	time_multiplier = 1;
	distorter = 0;
	cout.precision(16);
    infoJoystick();
    enable_tcp= false;
    enable_mkfifo= false;
    mkfifo= new Mkfifo();
    tcp= new Tcp();
}

App::~App()
{
	delete tcp;
	delete mkfifo;
	delete ui;
	delete scripts;
	delete audio;
	delete commander;
	delete core;
	if (distorter) delete distorter;
	if (pixels) delete pixels;
	//delete stellState;
}


void App::infoJoystick()
{
	joystick = SDL_JoystickOpen(0); // we only take joystick number 0
	string strMsg= "Found " + Utility::intToString(SDL_NumJoysticks()) + " joystick(s)";
   	Log.write(strMsg, cLog::eLOG_TYPE_INFO);
    if (SDL_NumJoysticks() >0)
    {
		SDL_JoystickUpdate();
		strMsg.clear();
		strMsg ="\nJoystick name number 0 : "+ string(SDL_JoystickName(joystick));
		strMsg+="\nNumber of buttons : " + Utility::intToString(SDL_JoystickNumButtons(joystick) );
		strMsg+="\nNumber of axes : "+Utility::intToString( SDL_JoystickNumAxes(joystick) );
		strMsg+="\nNumber of hats : "+Utility::intToString( SDL_JoystickNumHats(joystick) );
		strMsg+="\nNumber of trackballs : "+Utility::intToString( SDL_JoystickNumBalls(joystick) );
		Log.write(strMsg, cLog::eLOG_TYPE_INFO);
		SDL_JoystickEventState(SDL_ENABLE);
	}
	haveJoystickEvent=false;
}

void App::setViewPortDistorterType(const string &type)
{
	Uint16 w, h;
	m_sdl->getResolution( &w, &h );

	if (distorter) {
		if (distorter->getType() == type) return;
		delete distorter;
		distorter = 0;
	}
	distorter = ViewportDistorter::create(type, w, h, core);
	InitParser conf;
	conf.load(AppSettings::Instance()->getConfigFile());
	distorter->init(conf);
}

string App::getViewPortDistorterType(void) const
{
	if (distorter) return distorter->getType();
	return "none";
}

string App::getVideoModeList(void) const {
	return m_sdl->getVideoModeList();
}

void App::print(string name, string value)
{
	stringstream oss;
	oss << "print " << name << " " << value;
	cout << oss.str() << endl;
	Log.write(oss.str(),  cLog::eLOG_TYPE_SCRIPT_WARNING);
}

void App::quit(void)
{
	static SDL_Event Q;						// Send a SDL_QUIT event
	Q.type = SDL_QUIT;						// To the SDL event queue
	//commander->execute_command( "external_viewer filename dummy.mp3 action stop");
	std::string action = "killall vlc &";
	if (!system(action.c_str())) Log.write("No action needed for VLC",cLog::eLOG_TYPE_ERROR);	
	if (SDL_PushEvent(&Q) == -1) {			// Try to send the event
		stringstream oss;
		oss << "SDL_QUIT event can't be pushed: " << SDL_GetError() ;
		Log.write(oss.str(),  cLog::eLOG_TYPE_ERROR);
		exit(-1);
	}
}

//! Load configuration from disk
void App::init(void)
{
	if (!initialized)
		Translator::initSystemLanguage();

	// Initialize video device and other sdl parameters
	InitParser conf;
	AppSettings::Instance()->loadAppSettings( &conf );

	// forced settings
	conf.set_double("video:horizontal_offset", 0);
	conf.set_double("video:vertical_offset", 0);
	//conf.set_double("navigation:init_fov", 180);


	conf.set_boolean("navigation:flag_enable_move_mouse", 0);
	conf.set_double("viewing:constellation_art_fade_duration", 2);
	conf.set_double("landscape:constellation_art_fade_duration", 4);
	//conf.set_double("gui:base_font_size", 12);

	// Clear screen, this fixes a strange artifact at loading time in the upper top corner.
	glClear(GL_COLOR_BUFFER_BIT);

	maxfps 				= conf.get_double ("video","maximum_fps",10000);
	minfps 				= conf.get_double ("video","minimum_fps",10000);
	videoRecordFps		= conf.get_double ("video","video_record_fps",30);
	string appLocaleName = conf.get_str("localization", "app_locale", "system");
	time_format = string_to_s_time_format(conf.get_str("localization:time_display_format"));
	date_format = string_to_s_date_format(conf.get_str("localization:date_display_format"));
	setAppLanguage(appLocaleName);
	scripts->set_allow_ui( conf.get_boolean("gui","flag_script_allow_ui",0) );

	joystickSensitivity	= conf.get_int("joystick", "sensitivity", 500);
	joystickIntensity	= conf.get_double("joystick", "intensity", 8000.0);
	// time_zone used to be in init_location section of config,
	// so use that as fallback when reading config - Rob
	string tzstr = conf.get_str("localization", "time_zone", conf.get_str("init_location", "time_zone", "system_default"));
	if (tzstr == "system_default") {
		time_zone_mode = S_TZ_SYSTEM_DEFAULT;
		// Set the program global intern timezones variables from the system locale
		tzset();
	} else {
		if (tzstr == "gmt+x") { // TODO : handle GMT+X timezones form
			time_zone_mode = S_TZ_GMT_SHIFT;
			// GMT_shift = x;
		} else {
			// We have a custom time zone name
			time_zone_mode = S_TZ_CUSTOM;
			set_custom_tz_name(tzstr);
		}
	}

	Uint16 w, h;
	m_sdl->getResolution( &w, &h );
	core->init(conf, w, h);

	pixels = new unsigned char[4 * h * h];

	string tmpstr = conf.get_str("projection:viewport");
	if (tmpstr=="maximized") core->setMaximizedViewport(w, h);
	else
		//	  if (tmpstr=="square" || tmpstr=="disk") {

		if (tmpstr=="square") {
			core->setSquareViewport(w, h,
			                        conf.get_int("video:horizontal_offset"),
			                        conf.get_int("video:horizontal_offset"));

		} else if (tmpstr=="disk") {
			core->setViewportMaskDisk(w/2, h/2, w, h, h/2);

		} else {
			cerr << "ERROR : Unknown viewport type : " << tmpstr << endl;
			Log.write("ERROR : Unknown viewport type : "+tmpstr, cLog::eLOG_TYPE_ERROR);
			exit(-1);
		}

	// Digitalis projector configuration type, MUST call after setting disk viewport above


	int projConf = conf.get_int("projection", "projector_configuration", 0);


	core->setProjectorConfiguration( projConf );

	core->setProjectorOffset( conf.get_double("projection", "projector_offset_x", 0),
	                          conf.get_double("projection", "projector_offset_y", 0));

	core->setProjectorShearHorz(conf.get_double("projection", "horizontal_shear", 1));

	core->setCenterHorizontalOffset(conf.get_int("video", "center_horizontal_offset", 0));



	// Navigation section
	PresetSkyTime 		= conf.get_double ("navigation","preset_sky_time",2451545.);
	StartupTimeMode 	= conf.get_str("navigation:startup_time_mode");	// Can be "now" or "preset"
	flagEnableDragMouse	= conf.get_boolean("navigation","flag_enable_drag_mouse");
	MouseZoom			= conf.get_int("navigation","mouse_zoom",30);

	DayKeyMode = conf.get_str("navigation","day_key_mode","calendar");  // calendar or sidereal
	Log.write("Read daykeymode as <" + DayKeyMode + ">", cLog::eLOG_TYPE_INFO);

	if (StartupTimeMode=="preset" || StartupTimeMode=="Preset")
		core->setJDay(PresetSkyTime - get_GMT_shift(PresetSkyTime) * JD_HOUR);
	else core->setTimeNow();

	// initialisation of the User Interface

	// TODO: Need way to update settings from config without reinitializing whole gui
	ui->init(conf);
	ui->localizeTui( );
	if (!initialized) ui->init_tui(); // don't reinit tui since probably called from there
	else ui->localizeTui();  // update translations/fonts as needed

	// Initialisation of the color scheme
	//draw_mode = App::DM_NONE;  // fool caching
	//setVisionModeNormal();

	//set all color 
	core->setColorScheme(AppSettings::Instance()->getConfigFile(), "color");
	ui->setColorScheme(AppSettings::Instance()->getConfigFile(), "color");

	//~ if (conf.get_boolean("viewing:flag_chart")) setVisionModeChart();
	//~ if (conf.get_boolean("viewing:flag_night")) setVisionModeNight();

	if (distorter == 0) {
		setViewPortDistorterType(conf.get_str("video","distorter","none"));
	}


	// system integration hook
	//~ if (!system((core->getDataDir() + "script_app_init").c_str())) 	message->m_cout(__FILE__, WARNING, "core->script_app_init have problem" );

	// This did not seem to be required
	// and should be done by script_mgr anyway
	//	remove(SCRIPT_IS_PLAYING_FILE);


	if (! initialized)
	{
		flagMasterput=conf.get_boolean("main:flag_masterput");
		enable_tcp=conf.get_boolean("io","enable_tcp", false);
		enable_mkfifo=conf.get_boolean("io","enable_mkfifo",false);
	
		stringstream oss;
		oss << "Masterput is : " << flagMasterput <<endl << "Tcp is : " << enable_tcp << endl << "Mkfifo is : " << enable_mkfifo;
		Log.write(oss.str(), cLog::eLOG_TYPE_INFO);
	
	
		if (enable_tcp) {
			int port = conf.get_int("io:tcp_port_in");
			int buffer_in_size=conf.get_int("io:tcp_buffer_in_size");	
			printf("buffer TCP taille %i\n",buffer_in_size);
			tcp->init(port, buffer_in_size);
		}
	
		if (enable_mkfifo) {
			string mkfifo_file_in = conf.get_str("io:mkfifo_file_in");
			int buffer_in_size=conf.get_int("io:mkfifo_buffer_in_size");
			//string mkfifo_file_out = conf.get_str("io:mkfifo_file_out");
			//int buffer_out_size=conf.get_int("io:mkfifo_buffer_out_size");
			printf("buffer MKFIFO taille %i\n",buffer_in_size);
			mkfifo->init_in(mkfifo_file_in, buffer_in_size);
		//	mkfifo->init_out(mkfifo_file_out, buffer_out_size);
		}
	}
	
		// play startup script, if available
	if (!initialized && scripts) scripts->play_startup_script();

	initialized = 1;
}



void App::updateSharedInData( void )
{
	if (enable_mkfifo)
	{
		string out;
		if (mkfifo->update_in(out))
		{
			printf("dans app.cpp j'ai obtenu du mkfifo: %s\n",out.c_str());
			commander->execute_command(out);
		}
	}
	if (enable_tcp)
	{
		string out;
		if (tcp->update(out))
		{
			printf("dans app.cpp j'ai obtenu du tcp : %s\n",out.c_str());
			commander->execute_command(out);
		}
	}
}


//~ void App::updateSharedOutData( void )
//~ {
	//~ if (enable_mkfifo)
	//~ {
		//~ string string_to_extract;
		//~ while( stellState->updateState(string_to_extract) )
		//~ {
			//~ printf("Echange de données avec le thread\t%s\n",string_to_extract.c_str());
			//~ mkfifo->get_for_outside(string_to_extract);
			//~ //memset(string_to_extract,'\0',sizeof(string_to_extract));
		//~ }
	//~ }
//~ }

//~ void App::UpdateSharedData( void ) {
	//~ string cmdData, response;
	//~ ISharedData* shared = SharedData::Instance();
//~
	//~ bool have_sych_data = shared->Read( cmdData );
	//~ unsigned long delay = 0;
	//~ if( have_sych_data ) {
		//~ // Execute command untrusted
		//~ if( !commander->execute_command(cmdData, delay, 1) ) {
			//~ response = commander->getErrorString();
			//~ if( response.empty() )
				//~ response = _("Unrecognized or malformed command name.");
		//~ }
		//~ else
			//~ response = _("Success");
//~
		//~ shared->Write(response);
	//~ }
//~
	//~ // Asynchronous channel
	//~ bool have_asynch_data = shared->ReadRT( cmdData );
	//~ if( have_asynch_data ) {
		//~ if( !m_skyCommander.execute_command(cmdData) ) {
			//~ commander->execute_command(cmdData, delay, 1);
		//~ }
	//~ }
//~
	//~ // Stellarium360 'dynamic' state push
	//~ shared->DataPush();
//~ }

void App::update(int delta_time)	//it take 8 ms to make a loop in my computer O.N. 2014/08/28
{
	++frame;
	timefr+=delta_time;
	if (timefr-timeBase > 1000) {
		fps=frame*1000.0/(timefr-timeBase);				// Calc the FPS rate
		frame = 0;
		timeBase+=1000;
	}

	// change time rate if needed to fast forward scripts
	delta_time *= time_multiplier;

	//~  OLD class audio
	// keep audio position updated if changing time multiplier
	//~ if (!scripts->is_paused()) commander->update(delta_time);

	// run command from a running script
	scripts->update(delta_time);
	if (!scripts->is_paused() || !scripts->is_faster() )	audio->update(delta_time);

	// run any incoming command from shared memory interface
	updateSharedInData();
	//~ updateSharedOutData();
	
	ui->gui_update_widgets(delta_time);
	ui->tui_update_widgets();

	if (!scripts->is_paused()) ImageMgr::getImageMgr("script").update(delta_time);

	core->update(delta_time); //update take 7 ms on my computer O.N. 2014/08/28
}

//! Main drawinf function called at each frame
double App::draw(int delta_time)
{
	Uint16 w, h;
	m_sdl->getResolution( &w, &h );
// - 20061020 - clear areas not covered by main viewport
	glDisable(GL_BLEND);
	glColor3f(0.f,0.f,0.f);
	set2DfullscreenProjection();
	glBegin(GL_QUADS);
	{
		glVertex2f(0,h);
		glVertex2f(w,h);
		glVertex2f(w,0);
		glVertex2f(0,0);
	}
	glEnd();
	restoreFrom2DfullscreenProjection();

	// Render all the main objects of the application
	double squaredDistance = core->draw(delta_time);

	// Draw the Graphical ui and the Text ui
	ui->draw();

	distorter->distort();

	return squaredDistance;
}

//! @brief Set the application locale. This apply to GUI, console messages etc..
void App::setAppLanguage(const std::string& newAppLocaleName)
{
	// Update the translator with new locale name
	Translator::globalTranslator = Translator(PACKAGE, core->getLocaleDir(), newAppLocaleName);
	//cout << "Application locale is " << Translator::globalTranslator.getLocaleName() << endl;
	Log.write("Application locale is " + Translator::globalTranslator.getLocaleName(), cLog::eLOG_TYPE_INFO);
	
	// update translations and font in tui
	
	//-- ui->localizeTui( ); font not initialised yet O.N.

	//~ SettingsState state;
	//~ strcpy( state.m_state.selected_language, newAppLocaleName.c_str() );
	//~ SharedData::Instance()->Settings( state );

	// TODO: GUI needs to be reinitialized to load new translations and/or fonts
}

//! Handle mouse clics
int App::handleClick(int x, int y, S_GUI_VALUE button, S_GUI_VALUE state)
{
	distorter->distortXY(x,y);
	return ui->handle_clic(x, y, button, state);
}

//! Handle mouse move
int App::handleMove(int x, int y)
{
	//~ if (FlagEnableMoveMouse) {
		//~ if (x == 0) {
			//~ core->turn_left(1);
			//~ is_mouse_moving_horiz = true;
		//~ } else if (x == core->getViewportWidth() - 1) {
			//~ core->turn_right(1);
			//~ is_mouse_moving_horiz = true;
		//~ } else if (is_mouse_moving_horiz) {
			//~ core->turn_left(0);
			//~ is_mouse_moving_horiz = false;
		//~ }
//~ 
		//~ if (y == 0) {
			//~ core->turn_up(1);
			//~ is_mouse_moving_vert = true;
		//~ } else if (y == core->getViewportHeight() - 1) {
			//~ core->turn_down(1);
			//~ is_mouse_moving_vert = true;
		//~ } else if (is_mouse_moving_vert) {
			//~ core->turn_up(0);
			//~ is_mouse_moving_vert = false;
		//~ }
	//~ }

	return ui->handle_move(x, y);

}

void App::handleJoyHat(SDL_JoyHatEvent E)
{
	// Mouvement d'un chapeau
	// Nous devons donc utiliser le champ jhat
	//printf("Mouvement du chapeau %d du joystick %d\n",E.jhat.hat,E.jhat.which);
	switch (E.value)
	{
	case SDL_HAT_CENTERED:
		core->turn_up(0);
		core->turn_right(0);
		core->turn_down(0);
		core->turn_left(0);
		//printf("Retour position neutre\n");
		break;
	case SDL_HAT_DOWN :
		core->turn_down(1);
		//commander->execute_command( "timerate rate 0");
		break;
	case SDL_HAT_LEFT :
		core->turn_left(1);
		//commander->execute_command( "timerate action decrement");
		break;
	case SDL_HAT_RIGHT :
		core->turn_right(1);
		//commander->execute_command( "timerate action increment");
		break;
	case SDL_HAT_UP :
		core->turn_up(1);
		//commander->execute_command( "timerate rate 1");
		break;
	default: break;
	}
}

void App::handleJoyDeal()
{
	if (!haveJoystickEvent)	return; //ce boolean evite de lire les axes si rien ne se passe.
	int distZ;
	double latimem;
	stringstream oss;
	if (abs(joystickEvent[AXIS0])>0.01) {
		distZ=sqrt((posMouseX-core->getDisplayWidth() /2)*(posMouseX-core->getDisplayWidth() /2)+(posMouseY-core->getDisplayHeight() /2)*(posMouseY-core->getDisplayHeight() /2));
		if (distZ<1)
			distZ=1;
		posMouseX=posMouseX+(joystickEvent[AXIS0])*(posMouseY-core->getDisplayHeight()/2)/distZ;
		posMouseY=posMouseY-(joystickEvent[AXIS0])*(posMouseX-core->getDisplayWidth()/2)/distZ;
		SDL_WarpMouseInWindow(window, posMouseX, posMouseY);
		handleMove(posMouseX , posMouseY);
	}
	if (abs(joystickEvent[AXIS1])>0.01) {
		distZ=sqrt((posMouseX-core->getDisplayWidth() /2)*(posMouseX-core->getDisplayWidth() /2)+(posMouseY-core->getDisplayHeight() /2)*(posMouseY-core->getDisplayHeight() /2));		
		if (distZ<1)
			distZ=1;
		posMouseX=posMouseX+(joystickEvent[AXIS1])*(posMouseX-core->getDisplayWidth() /2-1)/distZ;
		posMouseY=posMouseY+(joystickEvent[AXIS1])*(posMouseY-core->getDisplayHeight() /2-1)/distZ;
		SDL_WarpMouseInWindow(window, posMouseX, posMouseY);
		handleMove(posMouseX , posMouseY);
	}
	if (abs(joystickEvent[AXIS2])>0.05) {
	        //printf("axe %f\n", joystickEvent[AXIS2]);
		if (SDL_JoystickNumButtons(joystick)==16)
		  core->moveRelLonObserver(joystickEvent[AXIS2],0.1);
		else
		  core->moveRelLonObserver(-joystickEvent[AXIS2],0.1);		  
	}
	if (abs(joystickEvent[AXIS3])>0.05) {
	        //printf("axe %f\n", joystickEvent[AXIS3]);
		if (SDL_JoystickNumButtons(joystick)==16)
		  core->moveRelLatObserver(joystickEvent[AXIS3],0.1);
		else
		  core->moveRelLatObserver(-joystickEvent[AXIS3],0.1);
	}
	if ((joystickEvent[AXIS4]>-0.45)&&(joystickEvent[AXIS4]!=0)) {
					latimem = core->getObservatory()->get_altitude();
					//printf("axe %f\n", joystickEvent[AXIS4]);
					latimem = latimem/(1.6+joystickEvent[AXIS4]);
					oss << "moveto altitude " << latimem << " duration 1";
					commander->execute_command(oss.str());  
	}
	if ((joystickEvent[AXIS5]>-0.45)&&(joystickEvent[AXIS5]!=0)) {
					latimem = core->getObservatory()->get_altitude();
					//printf("axe %f\n", joystickEvent[AXIS5]);
					latimem = latimem*(1.6+joystickEvent[AXIS5]);
					oss << "moveto altitude " << latimem << " duration 1";
					commander->execute_command(oss.str());        
	}
}

void App::handleJoyAxis(SDL_JoyAxisEvent E)
{
  
	// Mouvement d'un axe
	// Nous devons donc utiliser le champ jaxis
	//
	// les infos contenant la valeur de E.value de l'axe sont envoyées dans le tableau joystickEvent
	// ce dernier tableau sera lu par la fonction handleJoyDeal
	// cette fonction executée à tous les tours n'est active que si le boolean havejoystick est à true.
	//

	if  ( abs(E.value)> joystickSensitivity )
	{   //printf("Mouvement de l'axe %d\n", E.axis);
		//printf("%d est la nouvelle valeur de l'axe %d pour le joystick %d\n", E.jaxis.value, E.jaxis.axis, E.jaxis.which);
		switch (E.axis) {
		case 0:
			joystickEvent[AXIS0]=E.value/joystickIntensity;
			haveJoystickEvent=true;
			break;
		case 1:
				joystickEvent[AXIS1]=E.value/joystickIntensity;
			haveJoystickEvent=true;
			break;
		case 2:
			if (SDL_JoystickNumButtons(joystick)==16)
			  joystickEvent[AXIS2]=E.value/(joystickIntensity*8.0);
			else  
			  joystickEvent[AXIS4]=E.value/(joystickIntensity*8.0);
			haveJoystickEvent=true;
			break;
		case 3:
			if (SDL_JoystickNumButtons(joystick)==16)
			  joystickEvent[AXIS3]=-E.value/(joystickIntensity*8.0);
			else
			  joystickEvent[AXIS2]=-E.value/(joystickIntensity*8.0);
			haveJoystickEvent=true;
			break;
		case 4:
			if (SDL_JoystickNumButtons(joystick)==16)		  
			  joystickEvent[AXIS4]=E.value/(joystickIntensity*8.0);
			else
			  joystickEvent[AXIS3]=E.value/(joystickIntensity*8.0);
			haveJoystickEvent=true;
			break;
		case 5:
			joystickEvent[AXIS5]=E.value/(joystickIntensity*8.0);
			haveJoystickEvent=true;
			break;
		break;
		}
	}
	else
	{
		haveJoystickEvent=false;
		joystickEvent[AXIS0]=0;
		joystickEvent[AXIS1]=0;
		joystickEvent[AXIS2]=0;
		joystickEvent[AXIS3]=0;
		joystickEvent[AXIS4]=0;
		joystickEvent[AXIS5]=0;
	}
}


void App::handleJoyButtonUp (SDL_JoyButtonEvent E)
{
    stringstream oss;
    switch(E.button) {
		case 0: // TRIANGLE / A
			if (SDL_JoystickNumButtons(joystick)==16)		  
			core->zoom_in(0);
			break;
		case 1: // CIRCLE / B
			break;
		case 2: // CROSS / X
			break;
		case 3: // SQUARE / Y
			if (SDL_JoystickNumButtons(joystick)==16)		  
			core->zoom_out(0);
			// VOID
			break;
		case 4: // L2 / LB
			if (SDL_JoystickNumButtons(joystick)==16)		  
			core->lower_height(0);
			break;
		case 5: // R2 / RB
			if (SDL_JoystickNumButtons(joystick)==16)		  
			core->raise_height(0);
			break;
		case 6: // L1 / Back
			if (SDL_JoystickNumButtons(joystick)!=16)		  
			core->zoom_out(0);
			break;
		case 7: // R1 / Start
			if (SDL_JoystickNumButtons(joystick)!=16)		  
			core->zoom_in(0);
			break;
		case 8: // START / XBOX button
			break;
		case 9: // SELECT / stick left click
			if (SDL_JoystickNumButtons(joystick)!=16)		  
			SDL_WarpMouseInWindow( core->get_Window(), core->getDisplayWidth() /2, core->getDisplayHeight() /2);
			break;
		case 10: // L3 / stick right click
			break;
		case 11: // R3
			if (SDL_JoystickNumButtons(joystick)!=16)
			    core->turn_left(0);
			break;
		case 12: // UP
			if (SDL_JoystickNumButtons(joystick)==16)
				core->turn_up(0);
			else
			core->turn_right(0);
			break;
		case 13 : // RIGHT
			if (SDL_JoystickNumButtons(joystick)==16)
				core->turn_right(0);
			else
			core->turn_up(0);
			break;
		case 14 : // DOWN
			if (SDL_JoystickNumButtons(joystick)==16)
				core->turn_down(0);
			else
			core->turn_down(0);
			break;
		case 15 : // LEFT
			if (SDL_JoystickNumButtons(joystick)==16)
				core->turn_left(0);
			break;
		default:
			break;
  }
}

void App::handleJoyButtonDown(SDL_JoyButtonEvent E)
{
    stringstream oss;
    switch(E.button) {
				case 0: // TRIANGLE / A
					if (SDL_JoystickNumButtons(joystick)==16)		  
					core->zoom_in(1);
					else {
					handleClick(posMouseX, posMouseY, S_GUI_MOUSE_LEFT, S_GUI_PRESSED);
					handleClick(posMouseX, posMouseY, S_GUI_MOUSE_LEFT, S_GUI_RELEASED);
					}
					break;
				case 1: // CIRCLE / B
					handleClick(posMouseX, posMouseY, S_GUI_MOUSE_RIGHT, S_GUI_PRESSED);
					handleClick(posMouseX, posMouseY, S_GUI_MOUSE_RIGHT, S_GUI_RELEASED);
					break;
				case 2: // CROSS / X
					if (SDL_JoystickNumButtons(joystick)==16) {
					  handleClick(posMouseX, posMouseY, S_GUI_MOUSE_LEFT, S_GUI_PRESSED);
					  handleClick(posMouseX, posMouseY, S_GUI_MOUSE_LEFT, S_GUI_RELEASED);
					} else 
					  commander->execute_command( "zoom auto out manual 1");
					break;
				case 3: // SQUARE / Y
					if (SDL_JoystickNumButtons(joystick)==16)		  
					  core->zoom_out(1);
					else 
					  commander->execute_command( "zoom auto in manual 1");
					break;
				case 4: // L2 / LB
					if (SDL_JoystickNumButtons(joystick)==16)		  
					core->lower_height(1);
					else { if(scripts->is_playing())
					scripts->slower_script();
					else
					commander->execute_command( "timerate action decrement");
					}
					break;
				case 5: // R2 / RB
					if (SDL_JoystickNumButtons(joystick)==16)		  
					core->raise_height(1);
					else { if(scripts->is_playing())
					scripts->faster_script();
					else
					commander->execute_command( "timerate action increment");
					}
					break;
				case 6: // L1 / Back
					if (SDL_JoystickNumButtons(joystick)==16) {
					if(scripts->is_playing())
					scripts->slower_script();
					else
					commander->execute_command( "timerate action decrement");
					} else core->zoom_out(1);
					break;
				case 7: // R1 / Start
					if (SDL_JoystickNumButtons(joystick)==16) {
					if(scripts->is_playing())
					scripts->faster_script();
					else
					commander->execute_command( "timerate action increment");
					} else core->zoom_in(1);
					break;
				case 8: // START / XBOX button 
					if ( scripts->is_playing() ) {
					  commander->execute_command( "script action pause");
					  time_multiplier = 1; }
					else
					  commander->execute_command( "timerate rate 1");;
					break;
				case 9: // SELECT / click left click
					if (SDL_JoystickNumButtons(joystick)==16)
					commander->execute_command( "zoom auto initial");
					else
					//core->zoom_out(1);
					break;
				case 10: // L3 / stick right click
					if (SDL_JoystickNumButtons(joystick)==16)
					SDL_WarpMouseInWindow( core->get_Window(), core->getDisplayWidth() /2, core->getDisplayHeight() /2);
					else {
					commander->execute_command("select planet home_planet duration 5");
					commander->execute_command("flag track_object on");
					}
					break;
		case 11: // R3
			if (SDL_JoystickNumButtons(joystick)==16) {
			  commander->execute_command("select planet home_planet duration 5");
			  commander->execute_command("flag track_object on");
		        } else
		          core->turn_left(1);
			break;
		case 12: // UP
			if (SDL_JoystickNumButtons(joystick)==16)
			  core->turn_up(1);
			else
			  core->turn_right(1);
			break;
		case 13 : // RIGHT
			if (SDL_JoystickNumButtons(joystick)==16)
			  core->turn_right(1);
			else
			  core->turn_up(1);
			break;
		case 14 : // DOWN
			if (SDL_JoystickNumButtons(joystick)==16)
			  core->turn_down(1);
			else
			  core->turn_down(1);
			break;
		case 15 : // LEFT
			core->turn_left(1);
			break;
		default:
			break;
  }
}

//! Handle key press and release
int App::handleKeys(SDL_Scancode key, Uint16 mod, Uint16 unicode, s_gui::S_GUI_VALUE state)
{
	int retVal = 0;

	s_tui::S_TUI_VALUE tuiv;
	if (state == s_gui::S_GUI_PRESSED) tuiv = s_tui::S_TUI_PRESSED;
	else tuiv = s_tui::S_TUI_RELEASED;
	if (ui->FlagShowTuiMenu) {

		if (state==S_GUI_PRESSED && key==SDL_SCANCODE_SEMICOLON) {
			// leave tui menu
			ui->FlagShowTuiMenu = false;

			// If selected a script in tui, run that now
			if (SelectedScript!="")
// add to svn to allow spaces in scripts and dirs
				//~ commander->execute_command("script action play filename " +  SelectedScript
				                           //~ + " path " + SelectedScriptDirectory);
				scripts->play_script(SelectedScriptDirectory+SelectedScript, SelectedScriptDirectory);
			// clear out now
			SelectedScriptDirectory = SelectedScript = "";
			return 1;
		}

		if( ui->handle_keys_tui(key, tuiv) )
			return 1;

		return 0;
	}

	if (ui->handle_keys(key, mod, unicode, state))
		retVal = 1;

	if (state == S_GUI_PRESSED) {
		// Direction and zoom deplacements
		if (key==SDL_SCANCODE_LEFT) core->turn_left(1);
		if (key==SDL_SCANCODE_RIGHT) core->turn_right(1);
		if (key==SDL_SCANCODE_UP) {
			if (mod & KMOD_CTRL) core->zoom_in(1);
			else core->turn_up(1);
		}
		if (key==SDL_SCANCODE_DOWN) {
			if (mod & KMOD_CTRL) core->zoom_out(1);
			else core->turn_down(1);
		}
		if ((key==SDL_SCANCODE_PAGEUP) && !(ui->FlagConfig)) core->zoom_in(1);
		if ((key==SDL_SCANCODE_PAGEDOWN) && !(ui->FlagConfig)) core->zoom_out(1);
	} else {
		// When a deplacement key is released stop moving
		if (key==SDL_SCANCODE_LEFT) core->turn_left(0);
		if (key==SDL_SCANCODE_RIGHT) core->turn_right(0);
		if (mod & KMOD_CTRL) {
			if (key==SDL_SCANCODE_UP) core->zoom_in(0);
			if (key==SDL_SCANCODE_DOWN) core->zoom_out(0);
		} else {
			if (key==SDL_SCANCODE_UP) core->turn_up(0);
			if (key==SDL_SCANCODE_DOWN) core->turn_down(0);
		}
		if (key==SDL_SCANCODE_PAGEUP) core->zoom_in(0);
		if (key==SDL_SCANCODE_PAGEDOWN) core->zoom_out(0);
	}
	return retVal;
}


//! Set the drawing mode in 2D for drawing in the full screen
void App::set2DfullscreenProjection(void) const
{
	Uint16 w, h;
	m_sdl->getResolution( &w, &h );
	glViewport(0, 0, w, h);
	glMatrixMode(GL_PROJECTION);		// projection matrix mode
	glPushMatrix();						// store previous matrix
	glLoadIdentity();
	gluOrtho2D(	0, w,
	            0, h);			// set a 2D orthographic projection
	glMatrixMode(GL_MODELVIEW);			// modelview matrix mode
	glPushMatrix();
	glLoadIdentity();
}

//! Restore previous projection mode
void App::restoreFrom2DfullscreenProjection(void) const
{
	glMatrixMode(GL_PROJECTION);		// Restore previous matrix
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

//~ //! Set flag for activating night vision mode
//~ void App::setVisionModeNight(void)
//~ {
	//~ if (!getVisionModeNight()) {
		//~ core->setColorScheme(AppSettings::Instance()->getConfigFile(), "night_color");
		//~ ui->setColorScheme(AppSettings::Instance()->getConfigFile(), "night_color");
	//~ }
	//~ draw_mode=DM_NIGHT;
//~ }
//~ 
//~ //! Set flag for activating chart vision mode
//~ void App::setVisionModeChart(void)
//~ {
	//~ if (!getVisionModeChart()) {
		//~ core->setColorScheme(AppSettings::Instance()->getConfigFile(), "chart_color");
		//~ ui->setColorScheme(AppSettings::Instance()->getConfigFile(), "chart_color");
	//~ }
	//~ draw_mode=DM_CHART;
//~ }
//~ 
//~ //! Set flag for activating chart vision mode
//~ // ["color" section name used for easier backward compatibility for older configs - Rob]
//~ void App::setVisionModeNormal(void)
//~ {
	//~ if (!getVisionModeNormal()) {
		//~ core->setColorScheme(AppSettings::Instance()->getConfigFile(), "color");
		//~ ui->setColorScheme(AppSettings::Instance()->getConfigFile(), "color");
	//~ }
	//~ draw_mode=DM_NORMAL;
//~ }

double App::getMouseCursorTimeout()
{
	return ui->getMouseCursorTimeout();
}

//! For use by TUI - saves all current settings
//! @todo Put in stel_core?
void App::saveCurrentConfig(const string& confFile)
{

	// No longer resaves everything, just settings user can change through UI

	//cout << "Saving configuration file " << confFile << " ..." << endl;
	Log.write("Saving configuration file " + confFile + " ...", cLog::eLOG_TYPE_INFO);
	InitParser conf;
	conf.load(confFile);

	// Main section
	conf.set_str	("main:version", string(VERSION));
    conf.set_boolean("main:flag_optoma", core->getFlagVpOptoma());
    conf.set_boolean("main:flag_masterput", getFlagMasterput());

	// localization section
	conf.set_str    ("localization:sky_culture", core->getSkyCultureDir());
	conf.set_str    ("localization:sky_locale", core->getSkyLanguage());
	conf.set_str    ("localization:app_locale", getAppLanguage());
	conf.set_str	("localization:time_display_format", get_time_format_str());
	conf.set_str	("localization:date_display_format", get_date_format_str());
	if (time_zone_mode == S_TZ_CUSTOM) {
		conf.set_str("localization:time_zone", custom_tz_name);
	}
	if (time_zone_mode == S_TZ_SYSTEM_DEFAULT) {
		conf.set_str("localization:time_zone", "system_default");
	}
	if (time_zone_mode == S_TZ_GMT_SHIFT) {
		conf.set_str("localization:time_zone", "gmt+x");
	}

	// Rendering section
	conf.set_boolean("rendering:flag_antialias_lines", core->getFlagAntialiasLines());
	conf.set_double("rendering:line_width", core->getLineWidth());

	// viewing section
	conf.set_boolean("viewing:flag_constellation_drawing", core->getFlagConstellationLines());
	conf.set_boolean("viewing:flag_constellation_name", core->getFlagConstellationNames());
	conf.set_boolean("viewing:flag_constellation_art", core->getFlagConstellationArt());
	conf.set_boolean("viewing:flag_constellation_boundaries", core->getFlagConstellationBoundaries());
	conf.set_boolean("viewing:flag_constellation_pick", core->getFlagConstellationIsolateSelected());
	conf.set_double("viewing:moon_scale", core->getMoonScale());
	conf.set_double("viewing:sun_scale", core->getSunScale());
	//conf.set_boolean("viewing:use_common_names", FlagUseCommonNames);
	conf.set_boolean("viewing:flag_equatorial_grid", core->getFlagEquatorGrid());
	conf.set_boolean("viewing:flag_ecliptic_grid", core->getFlagEclipticGrid());
	conf.set_boolean("viewing:flag_galactic_grid", core->getFlagGalacticGrid());
	conf.set_boolean("viewing:flag_azimutal_grid", core->getFlagAzimutalGrid());
	conf.set_boolean("viewing:flag_equator_line", core->getFlagEquatorLine());
	conf.set_boolean("viewing:flag_ecliptic_line", core->getFlagEclipticLine());
	conf.set_boolean("viewing:flag_cardinal_points", core->getFlagCardinalsPoints());
	conf.set_boolean("viewing:flag_zenith_line", core->getFlagZenithLine());
	conf.set_boolean("viewing:flag_polar_circle", core->getFlagPolarCircle());
	conf.set_boolean("viewing:flag_ecliptic_center", core->getFlagEclipticCenter());
	conf.set_boolean("viewing:flag_galactic_pole", core->getFlagGalacticPole());
	conf.set_boolean("viewing:flag_galactic_center", core->getFlagGalacticCenter());
	conf.set_boolean("viewing:flag_vernal_points", core->getFlagVernalPoints());
	conf.set_boolean("viewing:flag_analemma", core->getFlagAnalemma());
	conf.set_boolean("viewing:flag_analemma_line", core->getFlagAnalemmaLine());
	conf.set_boolean("viewing:flag_aries_line", core->getFlagAriesLine());
	conf.set_boolean("viewing:flag_zodiac", core->getFlagZodiac());
	conf.set_boolean("viewing:flag_personal", core->getFlagPersonal());
	conf.set_boolean("viewing:flag_personeq", core->getFlagPersoneq());
	conf.set_boolean("viewing:flag_greenwich_line", core->getFlagGreenwichLine());
	conf.set_boolean("viewing:flag_vertical_line", core->getFlagVerticalLine());
	conf.set_boolean("viewing:flag_meridian_line", core->getFlagMeridianLine());
	conf.set_boolean("viewing:flag_precession_circle", core->getFlagPrecessionCircle());
	conf.set_boolean("viewing:flag_circumpolar_circle", core->getFlagCircumpolarCircle());
	conf.set_boolean("viewing:flag_tropic_lines", core->getFlagTropicLines());
	conf.set_boolean("viewing:flag_moon_scaled", core->getFlagMoonScaled());
	conf.set_boolean("viewing:flag_sun_scaled", core->getFlagSunScaled());
	conf.set_double ("viewing:constellation_art_intensity", core->getConstellationArtIntensity());
	conf.set_double ("viewing:constellation_art_fade_duration", core->getConstellationArtFadeDuration());
	conf.set_double("viewing:light_pollution_limiting_magnitude", core->getLightPollutionLimitingMagnitude());

	// Landscape section
	conf.set_boolean("landscape:flag_landscape", core->getFlagLandscape());
	conf.set_boolean("landscape:flag_atmosphere", core->getFlagAtmosphere());
	conf.set_boolean("landscape:flag_fog", core->getFlagFog());
	//	conf.set_double ("viewing:atmosphere_fade_duration", core->getAtmosphereFadeDuration());

	// Star section
	conf.set_double ("stars:star_scale", core->getStarScale());
	conf.set_double ("stars:star_mag_scale", core->getStarMagScale());
	conf.set_boolean("stars:flag_point_star", core->getFlagPointStar());
	conf.set_double("stars:max_mag_star_name", core->getMaxMagStarName());
	conf.set_boolean("stars:flag_star_twinkle", core->getFlagStarTwinkle());
	conf.set_double("stars:star_twinkle_amount", core->getStarTwinkleAmount());
	conf.set_double("stars:star_limiting_mag", core->getStarLimitingMag());

	// Color section
	conf.set_str    ("color:azimuthal_color", Utility::vec3f_to_str(core->getColorAzimutalGrid()));
	conf.set_str    ("color:equatorial_color", Utility::vec3f_to_str(core->getColorEquatorGrid()));
	conf.set_str    ("color:ecliptic_color", Utility::vec3f_to_str(core->getColorEclipticGrid()));
	conf.set_str    ("color:equator_color", Utility::vec3f_to_str(core->getColorEquatorLine()));
	conf.set_str    ("color:ecliptic_color", Utility::vec3f_to_str(core->getColorEclipticLine()));
	conf.set_str    ("color:meridian_color", Utility::vec3f_to_str(core->getColorMeridianLine()));
	conf.set_str    ("color:zenith_color", Utility::vec3f_to_str(core->getColorZenithLine()));
	conf.set_str    ("color:polar_color", Utility::vec3f_to_str(core->getColorPolarCircle()));
	conf.set_str    ("color:ecliptic_center_color", Utility::vec3f_to_str(core->getColorEclipticCenter()));
	conf.set_str    ("color:galactic_pole_color", Utility::vec3f_to_str(core->getColorGalacticPole()));
	conf.set_str    ("color:galactic_center_color", Utility::vec3f_to_str(core->getColorGalacticCenter()));
	conf.set_str    ("color:vernal_points_color", Utility::vec3f_to_str(core->getColorVernalPoints()));
	conf.set_str    ("color:analemma_color", Utility::vec3f_to_str(core->getColorAnalemma()));
	conf.set_str    ("color:analemma_line_color", Utility::vec3f_to_str(core->getColorAnalemmaLine()));
	conf.set_str    ("color:aries_color", Utility::vec3f_to_str(core->getColorAriesLine()));
	conf.set_str    ("color:zodiac_color", Utility::vec3f_to_str(core->getColorZodiac()));
	conf.set_str    ("color:personal_color", Utility::vec3f_to_str(core->getColorPersonal()));
	conf.set_str    ("color:personeq_color", Utility::vec3f_to_str(core->getColorPersoneq()));
	conf.set_str    ("color:greenwich_color", Utility::vec3f_to_str(core->getColorGreenwichLine()));
	conf.set_str    ("color:vertical_line", Utility::vec3f_to_str(core->getColorVerticalLine()));
	conf.set_str    ("color:const_lines_color", Utility::vec3f_to_str(core->getColorConstellationLine()));
	conf.set_str    ("color:const_names_color", Utility::vec3f_to_str(core->getColorConstellationNames()));
	conf.set_str    ("color:const_art_color", Utility::vec3f_to_str(core->getColorConstellationArt()));
	conf.set_str    ("color:const_boundary_color", Utility::vec3f_to_str(core->getColorConstellationBoundaries()));
	conf.set_str	("color:nebula_label_color", Utility::vec3f_to_str(core->getColorNebulaLabels()));
	conf.set_str	("color:nebula_circle_color", Utility::vec3f_to_str(core->getColorNebulaCircle()));
	conf.set_str	("color:precession_circle_color", Utility::vec3f_to_str(core->getColorPrecessionCircle()));
	conf.set_str    ("color:cardinal_color", Utility::vec3f_to_str(core->getColorCardinalPoints()));
	conf.set_str    ("color:planet_names_color", Utility::vec3f_to_str(core->getDefaultColorPlanetNames()));
	conf.set_str    ("color:planet_orbits_color", Utility::vec3f_to_str(core->getDefaultColorPlanetOrbits()));
	conf.set_str    ("color:object_trails_color", Utility::vec3f_to_str(core->getDefaultColorPlanetTrails()));
	//  Are these used?
	//	conf.set_str    ("color:star_label_color", Utility::vec3f_to_str(core->getColorStarNames()));
	//  conf.set_str    ("color:star_circle_color", Utility::vec3f_to_str(core->getColorStarCircles()));

	// gui section
	conf.set_double("gui:mouse_cursor_timeout",getMouseCursorTimeout());

	// not user settable yet
	// conf.set_str	("gui:gui_base_color", Utility::vec3f_to_str(GuiBaseColor));
	// conf.set_str	("gui:gui_text_color", Utility::vec3f_to_str(GuiTextColor));
	// conf.set_double ("gui:base_font_size", FontSizeGeneral);

	// Text ui section
	conf.set_boolean("tui:flag_show_gravity_ui", ui->getFlagShowGravityUi());
	conf.set_boolean("tui:flag_show_tui_datetime", ui->getFlagShowTuiDateTime());
	conf.set_boolean("tui:flag_show_tui_short_obj_info", ui->getFlagShowTuiShortObjInfo());

	// Navigation section
	conf.set_boolean("navigation:flag_manual_zoom", core->getFlagManualAutoZoom());
	conf.set_double ("navigation:auto_move_duration", core->getAutoMoveDuration());
	conf.set_double ("navigation:zoom_speed", core->getZoomSpeed());
	conf.set_double ("navigation:flight_duration", core->getFlightDuration());
	conf.set_double ("navigation:preset_sky_time", PresetSkyTime);
	conf.set_str	("navigation:startup_time_mode", StartupTimeMode);
	conf.set_str	("navigation:day_key_mode", DayKeyMode);

	conf.set_double ("navigation:heading", core->getHeading());
	conf.set_double ("navigation:view_offset", core->getViewOffset());

	// Astro section
	conf.set_boolean("astro:flag_object_trails", core->getFlagPlanetsTrails());
	conf.set_boolean("astro:flag_bright_nebulae", core->getFlagBrightNebulae());
	conf.set_boolean("astro:flag_stars", core->getFlagStars());
	conf.set_boolean("astro:flag_star_name", core->getFlagStarName());
	conf.set_boolean("astro:flag_nebula", core->getFlagNebula());
	conf.set_boolean("astro:flag_nebula_names", core->getFlagNebulaNames());
	conf.set_boolean("astro:flag_nebula_hints", core->getFlagNebulaHints());
	conf.set_double("astro:max_mag_nebula_name", core->getNebulaMaxMagHints());
	conf.set_boolean("astro:flag_planets", core->getFlagPlanets());
	conf.set_boolean("astro:flag_planets_hints", core->getFlagPlanetsHints());
	conf.set_boolean("astro:flag_planets_orbits", core->getFlagPlanetsOrbits());
	conf.set_boolean("astro:flag_light_travel_time", core->getFlagLightTravelTime());

	conf.set_boolean("astro:flag_milky_way", core->getFlagMilkyWay());
	conf.set_double("astro:milky_way_intensity", core->getMilkyWayIntensity());

	conf.set_double("astro:star_size_limit", core->getStarSizeLimit());
	conf.set_double("astro:planet_size_marginal_limit", core->getPlanetsSizeLimit());

// projection section
	conf.set_int("projection:projector_configuration", core->getProjectorConfiguration());
	conf.set_double("projection:projector_offset_x", core->getProjectorOffsetX());
	conf.set_double("projection:projector_offset_y", core->getProjectorOffsetY());

	conf.set_double("projection:horizontal_shear", core->getProjectorShearHorz() );

	conf.set_int("video:center_horizontal_offset", core->getCenterHorizontalOffset() );


	// Get landscape and other observatory info
	// TODO: shouldn't observer already know what section to save in?
	(core->getObservatory())->setConf(conf, "init_location");

	conf.save(confFile);

}


void App::recordCommand(string commandline)
{
	scripts->record_command(commandline);
}


//! Return a string with the UTC date formated according to the date_format variable
string App::get_printable_date_UTC(double JD) const
{
	struct tm time_utc;
	NShadeDateTime::TimeTmFromJulian(JD, &time_utc);

	static char date[255];
	switch (date_format) {
	case S_DATE_SYSTEM_DEFAULT :
		NShadeDateTime::my_strftime(date, 254, "%x", &time_utc);
		break;
	case S_DATE_MMDDYYYY :
		NShadeDateTime::my_strftime(date, 254, "%m/%d/%Y", &time_utc);
		break;
	case S_DATE_DDMMYYYY :
		NShadeDateTime::my_strftime(date, 254, "%d/%m/%Y", &time_utc);
		break;
	case S_DATE_YYYYMMDD :
		NShadeDateTime::my_strftime(date, 254, "%Y-%m-%d", &time_utc);
		break;
	}
	return date;
}

//! Return a string with the UTC time formated according to the time_format variable
//! @fixme for some locales (french) the %p returns nothing
string App::get_printable_time_UTC(double JD) const
{
	struct tm time_utc;
	NShadeDateTime::TimeTmFromJulian(JD, &time_utc);

	static char heure[255];
	switch (time_format) {
	case S_TIME_SYSTEM_DEFAULT :
		NShadeDateTime::my_strftime(heure, 254, "%X", &time_utc);
		break;
	case S_TIME_24H :
		NShadeDateTime::my_strftime(heure, 254, "%H:%M:%S", &time_utc);
		break;
	case S_TIME_12H :
		NShadeDateTime::my_strftime(heure, 254, "%I:%M:%S %p", &time_utc);
		break;
	}
	return heure;
}

//! Return the time in ISO 8601 format that is : %Y-%m-%d %H:%M:%S
string App::get_ISO8601_time_local(double JD) const
{
	struct tm time_local;
	if (time_zone_mode == S_TZ_GMT_SHIFT)
		NShadeDateTime::TimeTmFromJulian(JD + GMT_shift, &time_local);
	else
		NShadeDateTime::TimeTmFromJulian(JD + NShadeDateTime::GMTShiftFromSystem(JD)*0.041666666666, &time_local);

	static char isotime[255];
	NShadeDateTime::my_strftime(isotime, 254, "%Y-%m-%d %H:%M:%S", &time_local);
	return isotime;
}


//! Return a string with the local date formated according to the date_format variable
string App::get_printable_date_local(double JD) const
{
	struct tm time_local;

	if (time_zone_mode == S_TZ_GMT_SHIFT)
		NShadeDateTime::TimeTmFromJulian(JD + GMT_shift, &time_local);
	else
		NShadeDateTime::TimeTmFromJulian(JD + NShadeDateTime::GMTShiftFromSystem(JD)*0.041666666666, &time_local);

	static char date[255];
	switch (date_format) {
	case S_DATE_SYSTEM_DEFAULT :
		NShadeDateTime::my_strftime(date, 254, "%x", &time_local);
		break;
	case S_DATE_MMDDYYYY :
		NShadeDateTime::my_strftime(date, 254, "%m/%d/%Y", &time_local);
		break;
	case S_DATE_DDMMYYYY :
		NShadeDateTime::my_strftime(date, 254, "%d/%m/%Y", &time_local);
		break;
	case S_DATE_YYYYMMDD :
		NShadeDateTime::my_strftime(date, 254, "%Y-%m-%d", &time_local);
		break;
	}

	return date;
}

//! Return a string with the local time (according to time_zone_mode variable) formated according to the time_format variable
string App::get_printable_time_local(double JD) const
{
	struct tm time_local;

	if (time_zone_mode == S_TZ_GMT_SHIFT)
		NShadeDateTime::TimeTmFromJulian(JD + GMT_shift, &time_local);
	else
		NShadeDateTime::TimeTmFromJulian(JD + NShadeDateTime::GMTShiftFromSystem(JD)*0.041666666666, &time_local);

	static char heure[255];
	switch (time_format) {
	case S_TIME_SYSTEM_DEFAULT :
		NShadeDateTime::my_strftime(heure, 254, "%X", &time_local);
		break;
	case S_TIME_24H :
		NShadeDateTime::my_strftime(heure, 254, "%H:%M:%S", &time_local);
		break;
	case S_TIME_12H :
		NShadeDateTime::my_strftime(heure, 254, "%I:%M:%S %p", &time_local);
		break;
	}
	return heure;
}

string App::get_printable_time_nav() const
{
	ostringstream os;
    double sidereal;
    double T;
    double Le;
    double LST;
    double /*edt,*/c,r,l,lct,m;
    double jd = core->getJDay();
    
    T = (jd - 2451545.0) / 36525.0;
    Le = core->getObservatory()->get_longitude();    
    /* calc mean angle */
    sidereal = 280.46061837 + (360.98564736629 * (jd - 2451545.0)) + (0.000387933 * T * T) - (T * T * T / 38710000.0);
    while (sidereal>=360) sidereal-=360;
    while (sidereal<0)    sidereal+=360;
    LST=sidereal+Le;
    while (LST>=360) LST-=360;
    while (LST<0)    LST+=360;
    m=357.5291+0.98560028*(jd-2451545);
    c=1.9148*sin(m*3.1415926/180)+0.02*sin(2*m*3.1415926/180)+0.0003*sin(3*m*3.1415926/180)/12;
    l=280.4665+0.98564736*(jd-2451545)+c;
    r=-2.468*sin(2*l*3.1415926/180)+0.053*sin(4*l*3.1415926/180)-0.0014*sin(6*l*3.1415926/180);
  //  edt=c+r;
    lct=-Le+c+r;
    while (lct>360) lct-=360;
    while (lct<0) lct+=360;
    
		os << get_printable_date_local(jd) << " " << get_printable_time_local(jd) << "  UTC " << get_printable_time_UTC(jd) << " GST " << Utility::printAngleHMS(sidereal*3.1415926/180) << " LST " << Utility::printAngleHMS(LST*3.1415926/180);
		os << " LCT " << get_printable_time_UTC(jd+Le/360);
		//if (lct<=180) os << " (TE=" << Utility::printAngleHMS(lct*3.1415926/360) << ")";
		//if (lct>180) os << " (TE=-" << Utility::printAngleHMS((360-lct)*3.1415926/360) << ")";
		os << "@" << "Lat: " << Utility::printAngleDMS(core->getObservatory()->get_latitude()*3.1415926/180) << " Lon: " << Utility::printAngleDMS(core->getObservatory()->get_longitude()*3.1415926/180)<< " Eq. Time: ";
                if ((c+r)>=0) os << Utility::printAngleHMS((c+r)*3.1415926/180);
                if ((c+r)<0) os << "-" << Utility::printAngleHMS(-(c+r)*3.1415926/180);
		os << " TLocalPass: " << Utility::printAngleHMS(3.1415926+(c+r)*3.1415926/180-core->getObservatory()->get_longitude()*3.1415926/180) << " TGMPass: " << Utility::printAngleHMS(3.1415926+(c+r)*3.1415926/180); 
	return os.str();
}



//! Convert the time format enum to its associated string
App::S_TIME_FORMAT App::string_to_s_time_format(const string& tf) const
{
	if (tf == "system_default") return S_TIME_SYSTEM_DEFAULT;
	if (tf == "24h") return S_TIME_24H;
	if (tf == "12h") return S_TIME_12H;
	//cout << "ERROR : unrecognized time_display_format : " << tf << " system_default used." << endl;
	Log.write(" unrecognized time_display_format : " + tf + " system_default used.", cLog::eLOG_TYPE_ERROR);
	return S_TIME_SYSTEM_DEFAULT;
}

//! Convert string to its associated time format enum 
string App::s_time_format_to_string(S_TIME_FORMAT tf) const
{
	if (tf == S_TIME_SYSTEM_DEFAULT) return "system_default";
	if (tf == S_TIME_24H) return "24h";
	if (tf == S_TIME_12H) return "12h";
	//cout << "ERROR : unrecognized time_display_format value : " << tf << " system_default used." << endl;
	stringstream oss;
	oss << "ERROR : unrecognized time_display_format value : " << tf << " system_default used.";
	Log.write(oss.str(),  cLog::eLOG_TYPE_ERROR);
	return "system_default";
}

//! Convert the date format enum to its associated string
App::S_DATE_FORMAT App::string_to_s_date_format(const string& df) const
{
	if (df == "system_default") return S_DATE_SYSTEM_DEFAULT;
	if (df == "mmddyyyy") return S_DATE_MMDDYYYY;
	if (df == "ddmmyyyy") return S_DATE_DDMMYYYY;
	if (df == "yyyymmdd") return S_DATE_YYYYMMDD;  // iso8601
	//cout << "ERROR : unrecognized date_display_format : " << df << " system_default used." << endl;
	Log.write(" unrecognized date_display_format : " + df + " system_default used.",cLog::eLOG_TYPE_ERROR );
	return S_DATE_SYSTEM_DEFAULT;
}

//! Convert string to its associated date format enum
string App::s_date_format_to_string(S_DATE_FORMAT df) const
{
	if (df == S_DATE_SYSTEM_DEFAULT) return "system_default";
	if (df == S_DATE_MMDDYYYY) return "mmddyyyy";
	if (df == S_DATE_DDMMYYYY) return "ddmmyyyy";
	if (df == S_DATE_YYYYMMDD) return "yyyymmdd";
	//cout << "ERROR : unrecognized date_display_format value : " << df << " system_default used." << endl;
	stringstream oss;
	oss << "ERROR : unrecognized date_display_format value : " << df << " system_default used.";
	Log.write(oss.str(), cLog::eLOG_TYPE_WARNING );
	return "system_default";
}

void App::set_custom_tz_name(const string& tzname)
{
	custom_tz_name = tzname;
	time_zone_mode = S_TZ_CUSTOM;

	if ( custom_tz_name != "") {
		//~ ObserverState state;
		//~ strncpy( state.tz, tzname.c_str(), sizeof(state.tz) );
		//~ state.tz[sizeof(state.tz)-1] = '\0';
		//~ SharedData::Instance()->Observer(state);

		// set the TZ environement variable and update c locale stuff
		putenv(strdup((string("TZ=") + custom_tz_name).c_str()));
		tzset();
	}
}

float App::get_GMT_shift(double JD, bool _local) const
{
	if (time_zone_mode == S_TZ_GMT_SHIFT) return GMT_shift;
	else return NShadeDateTime::GMTShiftFromSystem(JD,_local);
}

//! Terminate the application
void App::terminateApplication(void)
{
	static SDL_Event Q;						// Send a SDL_QUIT event
	Q.type = SDL_QUIT;						// To the SDL event queue
	//commander->execute_command( "external_viewer filename dummy.mp3 action stop");
	std::string action = "killall vlc &";
	if (!system(action.c_str())) Log.write("No action needed for VLC",cLog::eLOG_TYPE_ERROR);	
	if (SDL_PushEvent(&Q) == -1) {			// Try to send the event
		//printf("SDL_QUIT event can't be pushed: %s\n", SDL_GetError() );
		stringstream oss;
		oss << "SDL_QUIT event can't be pushed: " << SDL_GetError();
		Log.write( oss.str(), cLog::eLOG_TYPE_ERROR );
		exit(-1);
	}
}

enum {
	USER_EVENT_TICK
};

Uint32 mytimer_callback(Uint32 interval, void *param)
{
	SDL_Event event;
	event.type = SDL_USEREVENT;
	event.user.type = SDL_USEREVENT;
	event.user.code = USER_EVENT_TICK;
	event.user.data1 = NULL;
	event.user.data2 = NULL;
	if (SDL_PushEvent(&event) == -1) {
		stringstream oss;
		oss << "User tick event can't be pushed: " << SDL_GetError();
	    Log.write(oss.str(), cLog::eLOG_TYPE_ERROR);
		//printf("User tick event can't be pushed: %s\n", SDL_GetError() );
		exit(-1);
	}

	// End this timer.
	return 0;
}

//! Give the first line of a file 
//! @deprecated
string rmc_givedata_fromfile(std::string filetoread)
{
	std::ifstream rmc( filetoread.c_str());
	std::string gettexttoread;
	if (rmc)
	{
 	  std::getline(rmc,gettexttoread);
 	  rmc.close();
 	 }
	return gettexttoread;
}

//! Masterput script launch
//! @todo rewriting with pipe no kluge!
//! @deprecated
void App::masterput(string homeDir)
{
	//std::string action = "/home/planetarium/.spacecrafter/ftp/pub/masterput.launch";
	std::string action = "/home/planetarium/.spacecrafter/ftp/pub/masterput.launch";
	FILE * tempFile = fopen(action.c_str(),"r");
	if (tempFile) {
		fclose(tempFile);
		Log.write("MASTERPUT is in action", cLog::eLOG_TYPE_INFO);
		//std::string action = homeDir+"/."+APP_LOWER_NAME+"/ftp/pub/erase.sh &";
		//if (!system(action.c_str())) Log.write("MASTERPUT can't execute erase.sh",cLog::eLOG_TYPE_ERROR);
		unlink(action.c_str());
		scripts->play_script("/home/planetarium/.spacecrafter/ftp/pub/script.sts", "/home/planetarium/.spacecrafter/ftp/pub/");
	}

	std::string mpname = "/home/planetarium/.spacecrafter/ftp/pub/masterparam.txt";
	FILE *fpm = fopen(mpname.c_str(), "r");
	if (fpm) {
	  fclose(fpm);
	} else {
	FILE *fpm = fopen(mpname.c_str(), "w");
	fprintf(fpm,"%2.2f;",core->getObservatory()->get_latitude());
	fprintf(fpm,"%3.2f;",core->getObservatory()->get_longitude());
	fprintf(fpm,"%10.2f;",core->getObservatory()->get_altitude());
	fprintf(fpm,"%10.6f;",core->getJDay());
	fprintf(fpm,"%10.6f;",core->getHeading());
	fclose(fpm);
	}
}


void App::start_main_loop()
{
	bool AppVisible = true;		// At The Beginning, Our App Is Visible
	enum S_GUI_VALUE bt;
	Uint32 last_event_time = SDL_GetTicks();
	// How fast the objects on-screen are moving, in pixels/millisecond.
	double animationSpeed = 0;
	SDL_Event	E;
	// Hold the value of SDL_GetTicks at start of main loop (set 0 time)
	static Uint32 LastCount = SDL_GetTicks();
	static Uint32 TickCount = 0;

    //for joystick use
	//double latimem;
    std::ostringstream oss;

    //center mouse in middle screen
    SDL_WarpMouseInWindow( window, core->getDisplayWidth() /2, core->getDisplayHeight() /2);

	// Start the main loop
	while (1) {

    string homeDir = getenv("HOME");

	if (flagMasterput==true) masterput(homeDir);

	if (SDL_PollEvent(&E)) {	// Fetch The First Event Of The Queue
			if (E.type != SDL_USEREVENT) {
				last_event_time = SDL_GetTicks();
			}
			switch (E.type) {	// And Processing It
			case SDL_QUIT:
				ImageMgr::cleanUp();
				return;
				break;

			case SDL_JOYAXISMOTION :
				handleJoyAxis(E.jaxis);
				break;

			case SDL_JOYBUTTONUP:
				handleJoyButtonUp(E.jbutton);
				break;

			case SDL_JOYBUTTONDOWN:
				handleJoyButtonDown(E.jbutton);
				break;

			case SDL_JOYHATMOTION :
				handleJoyHat(E.jhat);
				break;

			case SDL_WINDOWEVENT:
               switch(E.window.event)
               {
                    //~ case SDL_WINDOWEVENT_RESIZED:
                         //~ if(E.window.data1&&E.window.data2)
                             //~ core->setViewportSize(E.window.data1,E.window.data2);
                    //break;

                    case SDL_WINDOWEVENT_FOCUS_GAINED:
                        AppVisible = true; // Activity's been gained
                    break;

                    case SDL_WINDOWEVENT_FOCUS_LOST:
                         AppVisible = true; //switch true because L.R. need it
                    break;
                }
            break;

			case SDL_MOUSEMOTION:
				posMouseX=E.motion.x;
				posMouseY=E.motion.y;
				//cout <<"(x,y): "<< posMouseX << ":" << posMouseY << endl;
				handleMove(E.motion.x,E.motion.y);
				break;

			case SDL_MOUSEBUTTONDOWN:
				// Convert the name from GLU to my GUI
				switch (E.button.button) {
				case SDL_BUTTON_RIGHT :
					bt=S_GUI_MOUSE_RIGHT;
					break;
				case SDL_BUTTON_LEFT :
					bt=S_GUI_MOUSE_LEFT;
					break;
				case SDL_BUTTON_MIDDLE :
					bt=S_GUI_MOUSE_MIDDLE;
					break;
				default :
					bt=S_GUI_MOUSE_LEFT;
				}
				handleClick(E.button.x,E.button.y,bt,S_GUI_PRESSED);
				break;

            case SDL_MOUSEWHEEL: //patch sdl2.0
                if(E.wheel.y>0)
                   bt=S_GUI_MOUSE_WHEELUP; 
                   else
                //~ if(E.wheel.y<0)
					bt=S_GUI_MOUSE_WHEELDOWN;
                handleClick(E.button.x,E.button.y,bt,S_GUI_PRESSED);
                break;

			case SDL_MOUSEBUTTONUP:
				// Convert the name from GLU to my GUI
				switch (E.button.button) {
				case SDL_BUTTON_RIGHT :
					bt=S_GUI_MOUSE_RIGHT;
					break;
				case SDL_BUTTON_LEFT :
					bt=S_GUI_MOUSE_LEFT;
					break;
				case SDL_BUTTON_MIDDLE :
					bt=S_GUI_MOUSE_MIDDLE;
					break;

				default :
					bt=S_GUI_MOUSE_LEFT;
				}
				handleClick(E.button.x,E.button.y,bt,S_GUI_RELEASED);
				break;

			case SDL_KEYDOWN:
				// shift-ctrl-v starts recording video
				//~ if (E.key.keysym.sym==SDLK_v && (E.key.keysym.mod & COMPATIBLE_KMOD_CTRL) && (E.key.keysym.mod & KMOD_SHIFT)) {
					//~ if(m_OutputFrameNumber) {
						//~ m_OutputFrameNumber = 0;
						//~ //ui->show_message(_("Stopped recording video frames."), 500);  // TODO not drawing
						//~ message->m_cout(__FILE__, INFO, string("Stopped recording video frames.") );
					//~ } else {
						//~ m_OutputFrameNumber = 1;
						//~ //ui->show_message(_("Now recording video frames!\nPress CTRL-SHIFT-V to stop."), 500);
						//~ message->m_cout(__FILE__, INFO, string("Begin recording video frames."));
					//~ }
					//~ break;
				//~ }

				// shift-ctrl-s take screenshot
				//~ if (E.key.keysym.scancode==SDL_SCANCODE_S && (E.key.keysym.mod & COMPATIBLE_KMOD_CTRL) && (E.key.keysym.mod & KMOD_SHIFT)) {
					//~ string tempName = getNextScreenshotFilename();
					//~ writeScreenshot(tempName);
					//~ message->m_cout(__FILE__, INFO, string("Saved screenshot to file "+ tempName ) );
					//~ break;
				//~ }

				// shift-ctrl-p reload allocate.ini
				//~ if (E.key.keysym.scancode==SDL_SCANCODE_P && (E.key.keysym.mod & COMPATIBLE_KMOD_CTRL) && (E.key.keysym.mod & KMOD_SHIFT)) {
					//~ ui->get_a_tab_parser();
					//~ break;
				//~ }

				// Rescue escape in case of lock : CTRL + ESC forces brutal quit
				if (E.key.keysym.scancode==SDL_SCANCODE_ESCAPE && (SDL_GetModState() & KMOD_CTRL)) {
					terminateApplication();
					break;
				}

				// Send the event to the gui and stop if it has been intercepted
				// use unicode translation, since not keyboard dependent
				// however, for non-printing keys must revert to just keysym... !
				//if (!handleKeys(E.key.keysym.scancode,E.key.keysym.mod,E.key.keysym.scancode,S_GUI_PRESSED) )
					handleKeys(E.key.keysym.scancode,E.key.keysym.mod,E.key.keysym.sym,S_GUI_PRESSED);
					// why not just switch the scan/keycode one time???

				break;

			case SDL_KEYUP:
				handleKeys(E.key.keysym.scancode,E.key.keysym.mod,E.key.keysym.sym,S_GUI_RELEASED);
				break;
			}
		} else { // No events to poll
			handleJoyDeal();
			
			// If the application is not visible
			if (!AppVisible) {
				// Leave the CPU alone, don't waste time, simply wait for an event
				SDL_WaitEvent(NULL);
			} else {
				// Compute how many fps we should run at to get 1 pixel movement each frame.
				double frameRate = 1000. * animationSpeed;
				// If there was user action in the last 2.5 seconds, shoot for the max framerate.
				if (SDL_GetTicks() - last_event_time < 2500 || frameRate > maxfps) {
					frameRate = maxfps;
				}
				if (frameRate < minfps) {
					frameRate = minfps;
				}
					frameRate = maxfps;

				TickCount = SDL_GetTicks();			// Get present ticks
				// Wait a while if drawing a frame right now would exceed our preferred framerate.
				if (TickCount-LastCount < 1000./frameRate) {
					unsigned int delay = (unsigned int) (1000./frameRate) - (TickCount-LastCount);
//					printf("delay=%d\n", delay);
					if (delay < 15) {
						// Less than 15ms, just do a dumb wait.
						SDL_Delay(delay);
					} else {
						// A longer delay. Use this timer song and dance so
						// that the app is still responsive if the user does something.
						if (!SDL_AddTimer(delay, mytimer_callback, NULL)) {
							stringstream oss;
							oss << "Error: couldn't create an SDL timer: " << SDL_GetError();
							cerr << oss.str() << endl;
							Log.write( oss.str(), cLog::eLOG_TYPE_ERROR );
						}
						SDL_WaitEvent(NULL);
					}
				}

				TickCount = SDL_GetTicks();			// Get present ticks

				// If outputting video frames, force output frame rate
				if(m_OutputFrameNumber) {
					TickCount = LastCount + 1000./videoRecordFps;
				}

				this->update(TickCount-LastCount);	// And update the motions and data
				double squaredDistance = this->draw(TickCount-LastCount);	// Do the drawings!

				// write out video frame if recording video
				if(m_OutputFrameNumber) {
					char c[6];
					snprintf(c,6,"%05u",m_OutputFrameNumber);
					writeScreenshot(getScreenshotDirectory() + ".stellarium360/vframes/" + APP_LOWER_NAME + "-" + string(c) + ".jpg");
					m_OutputFrameNumber++;
				}

				animationSpeed = sqrt(squaredDistance) / (TickCount-LastCount);
				LastCount = TickCount;				// Save the present tick probing
				SDL_GL_SwapWindow(window);  		// And swap the buffers
			}
		}
	}
}


//! Write current video frame to a specified file
void App::writeScreenshot(string filename)
{
	Uint16 rh=m_sdl->getCurrentResH();
	Uint16 rw=m_sdl->getCurrentResW();

  //unsigned char *pixels = new unsigned char[4 * rh * rh];
  if (pixels == NULL)
		return;

	glReadPixels( (rw-rh)/2, 0, rh, rh, GL_RGBA, GL_UNSIGNED_BYTE, pixels);
    //glReadPixels(0, 0, m_sdl->getScreen()->w, m_sdl->getScreen()->h, GL_RGBA, GL_UNSIGNED_BYTE, pixels);

  
  Magick::Image image( rh, rh, "RGBA", Magick::CharPixel, pixels );
  image.quality(95);
  image.flip();
  image.write(filename.c_str());

  //if (pixels) delete pixels;
}

//! Determine where screenshot files should go on different platforms
string App::getScreenshotDirectory()
{
	string shotdir;
#if defined(WIN32)
	char path[MAX_PATH];
	path[MAX_PATH-1] = '\0';
	if (SUCCEEDED(SHGetFolderPath(NULL, CSIDL_DESKTOPDIRECTORY, NULL, 0, path))) {
		shotdir = string(path)+"\\";
	} else {
		if (getenv("USERPROFILE")!=NULL) {
			//for Win XP etc.
			shotdir = string(getenv("USERPROFILE")) + "\\My Documents\\";
		} else {
			//for Win 98 etc.
			shotdir = "C:\\My Documents\\";
		}
	}
#else
	shotdir = string(getenv("HOME")) + "/";
#endif
#ifdef MACOSX
	shotdir += "/Desktop/";
#endif

	return shotdir;
}


//! Return the next sequential screenshot filename to use
string App::getNextScreenshotFilename()
{
	string tempName;
	char c[3];
	FILE *fp;

	string shotdir = getScreenshotDirectory();

	for (int j=0; j<=100; ++j) {
		snprintf(c,3,"%d",j);

		tempName = shotdir + APP_LOWER_NAME + c + ".jpg";
		fp = fopen(tempName.c_str(), "r");
		if (fp == NULL)
			break;
		else
			fclose(fp);
	}

	return tempName;
}
