/*
 * Stellarium360 astronomy simulation and visualization
 *
 * Copyright (C) 2002 Fabien Chereau
 * Copyright (C) 2009-2010 Digitalis Education Solutions, Inc.
 * Copyright (C) 2014 of the LSS Team & Association Sirius
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Stellarium360 is a free open project of of LSS team
 * See the TRADEMARKS file for free open project usage requirements.
 *
 */
 
#include <iostream>
#include <iomanip>

#include "night_shader.hpp"
#include "bump_shader.hpp"
#include "ring_shader.hpp"
#include "ringed_shader.hpp"
#include "entity.hpp"
#include "navigator.hpp"
#include "projector.hpp"
#include "s_font.hpp"
//#include "s_gui.hpp"
#include "stellastro.h" // just for get_apparent_sidereal_time
#include "log.hpp"


ArtificialPlanet::ArtificialPlanet(Entity &orig) : 
	Entity(0, "", UNKNOWN, 0, 0, orig.getRadius(), orig.getOblateness(), 
		   Vec3f(0,0,0), Vec3f(0,0,0), Vec3f(0,0,0), Vec3f(0,0,0), 1.0, 0, "", "", "",
		   NULL, false, "", NULL, true, false, 0,
		   "", "", "", "", ""), dest(0),
	orig_name(orig.getEnglishName()), orig_name_i18n(orig.getNameI18n()),
	orig_radius(orig.getRadius())
{
	if (orig.get_parent()) {
		parent = orig.getParent();
		while (parent->getParent()) parent = parent->getParent();
	} else {
		parent = &orig; // sun
	}

	re = orig.getRotationElements();
	setRotEquatorialToVsop87(orig.getRotEquatorialToVsop87());
	set_heliocentric_ecliptic_pos(orig.get_heliocentric_ecliptic_pos());
}

void ArtificialPlanet::setDest(Entity &dest) 
{
	ArtificialPlanet::dest = &dest;

	englishName = orig_name + "->" + dest.getEnglishName();
	nameI18 = orig_name_i18n + "->" + dest.getNameI18n();
	
	// rotation:
	const RotationElements &r(dest.getRotationElements());
	//  lastJD = StelApp::getInstance().getCore()->getNavigation()->getJDay();
	lastJD = dest.getLastJD();
  
	re.offset = r.offset + fmod(re.offset - r.offset
								+ 360.0*( (lastJD-re.epoch)/re.period
                                          - (lastJD-r.epoch)/r.period),
								360.0);
	
	re.epoch = r.epoch;
	re.period = r.period;
	if (re.offset - r.offset < -180.f) re.offset += 360.f; else
		if (re.offset - r.offset >  180.f) re.offset -= 360.f;

	// Assume not much planet movement while travelling
	startPos = get_heliocentric_ecliptic_pos();

	orig_rotation = getRot(this);
}

void ArtificialPlanet::setRot(const Vec3d &r)
{
	const double ca = cos(r[0]);
	const double sa = sin(r[0]);
	const double cd = cos(r[1]);
	const double sd = sin(r[1]);
	const double cp = cos(r[2]);
	const double sp = sin(r[2]);
	Mat4d m;
	m.r[ 0] = cd*cp;
	m.r[ 4] = sd;
	m.r[ 8] = cd*sp;
	m.r[12] = 0;
	m.r[ 1] = -ca*sd*cp -sa*sp;
	m.r[ 5] =  ca*cd;
	m.r[ 9] = -ca*sd*sp +sa*cp;
	m.r[13] = 0;
	m.r[ 2] =  sa*sd*cp -ca*sp;
	m.r[ 6] = -sa*cd;
	m.r[10] =  sa*sd*sp +ca*cp;
	m.r[14] = 0;
	m.r[ 3] = 0;
	m.r[ 7] = 0;
	m.r[11] = 0;
	m.r[15] = 1.0;
	setRotEquatorialToVsop87(m);
}

Vec3d ArtificialPlanet::getRot(Entity *p) 
{
	const Mat4d m(p->getRotEquatorialToVsop87());
	const double cos_r1 = sqrt(m.r[0]*m.r[0]+m.r[8]*m.r[8]);
	Vec3d r;
	r[1] = atan2(m.r[4],cos_r1);
    // not well defined if cos(r[1])==0:
	if (cos_r1 <= 0.0) {
		r[2] = 0.0;
		r[0] = atan2(m.r[9],m.r[10]);
	} else {
		r[0] = atan2(-m.r[6],m.r[5]);
		r[2] = atan2( m.r[8],m.r[0]);
	}
	return r;
}

void ArtificialPlanet::computeAverage(float t) 
{
	float f1, f2;

	// TODO a smoothing function would be better
	if(t < 0.5) f2 = 2*.95*t;
	else if(t < 0.75) f2 = .95 + 4*.04*(t-.5);
	else if(t < .875 ) f2 = .99 + 8*.009*(t-.75);
	else if(t < .9375 ) f2 = .999 + 16*.0009*(t-.875);
	else f2 = .9999 + 16*.0001*(t-.9375);

	f1 = 1 - f2;
	//	printf("%.4f\t%.4f\n", t, f1);

	set_heliocentric_ecliptic_pos(startPos*f1
								  + dest->get_heliocentric_ecliptic_pos()*f2);

	// finish rotation and radius adjustments before final travel
	if(t < 0.9) f2 = t / 0.9;
	else f2 = 1.0;
	f1 = 1 - f2;

	radius = orig_radius * f1 + dest->getRadius() * f2;

	// 3 Euler angles:
	// TODO: Seems like rotates excessively...
	Vec3d a1 = orig_rotation;
	const Vec3d a2(getRot(dest));
	if (a1[0]-a2[0] >  M_PI) a1[0] -= 2.0*M_PI; else
		if (a1[0]-a2[0] < -M_PI) a1[0] += 2.0*M_PI;

	if (a1[2]-a2[2] >  M_PI) a1[2] -= 2.0*M_PI; else
		if (a1[2]-a2[2] < -M_PI) a1[2] += 2.0*M_PI;

	setRot(a1*f1 + a2*f2);

    // rotation offset
	re.offset = f1*re.offset + f2*dest->getRotationElements().offset;
}
