/*
 * Stellarium360 astronomy simulation and visualization
 *
 * Copyright (C) 2005 Robert Spearman
 * Copyright (C) 2009 Digitalis Education Solutions, Inc.
 * Copyright (C) 2014 of the LSS Team & Association Sirius
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Stellarium360 is a free open project of of LSS team
 * See the TRADEMARKS file for free open project usage requirements.
 *
 */

/**
 * @class Audio
 * @brief manage audio tracks and sounds with SDL mixer
 * there is two functions type
 * \li working with sounds
 * \li working with tracks
 * We can only play 1 track but multiple sounds
*/

#include <iostream>
#include "audio.hpp"
#include "stellarium360.hpp"
#include "translator.hpp"
#include <sstream>
#include "log.hpp"

Audio::Audio()
{
	track = NULL;
	music_is_playing = false;
	music_length=0.0;
	master_volume=SDL_MIX_MAXVOLUME/3*2;
	elapsed_seconds=0.0;
	Log.write("Audio initialization successful", cLog::eLOG_TYPE_INFO);
	for (int i=0;i<MAX_CHANNELS;i++)
	sound[i]=NULL;
}

Audio::~Audio()
{
	if(track) {
		music_is_playing=false;
		Mix_HaltMusic(); // stop playing
		Mix_FreeMusic(track);  // free memory
	}
	for (int i=0;i<MAX_CHANNELS;i++)
	{
		if(sound[i] !=NULL)
			Mix_FreeChunk(sound[i]);
	}
	Mix_CloseAudio();
	Log.write("Audio support end", cLog::eLOG_TYPE_INFO);
}

float Audio::music_get_length(std::string filename)
{
	FILE *plop;
	string test="avconv -i "+ filename + " 2>&1 | grep 'Duration' | cut -d ' ' -f 4";
	plop=popen (test.c_str(), "r");
	float h, m, s;
	if ( fscanf(plop, "%f:%f:%f", &h,&m,&s) == 0) { // an
		pclose(plop);
		return 0.0;
	}
	pclose (plop);
	//printf("taille :%f h %f min %f s\n durée %f\n",h,m,s,(h*60+m)*60+s);
	return ((h*60+m)*60+s);
}

void Audio::loadMusic(std::string filename) 
{
	//printf("Fichier : %s\n",filename.c_str());
	track = Mix_LoadMUS(filename.c_str());
	if (track == NULL) {
		music_is_playing = false;
		Log.write("Could not load audio file " +filename, cLog::eLOG_TYPE_WARNING);
	} 
	else music_is_playing = 1;
	music_name = filename;
	music_length=music_get_length(filename);
	elapsed_seconds=0.0;
}


void Audio::playMusic(bool loop)
{
	//std::cout << "playmusic" << endl;
	// TODO check for load errors
	if (track != NULL) {
		music_is_playing = true;
		elapsed_seconds = 0.0;
		//cout << "Play Mid\n" << endl;
		//~ MediaState state;
		//~ state.playStateAudio = MediaState::PLAYING;
		//~ SharedData::Instance()->Media( state );
		if (loop)
			Mix_PlayMusic(track, -1);
		else
			Mix_PlayMusic(track, 0);
	//printf("File : %s durée %f\n",music_name.c_str(), music_length);
	}
}

// used solely to track elapsed seconds of play
void Audio::update(int delta_time)
{
	if (track ) elapsed_seconds += delta_time/1000.f;
}

// sychronize with elapsed time
// no longer starts playback if paused or disabled
void Audio::syncMusic()
{
	//~ printf("syncMusic\n");
	if (track==NULL) return;

	if (music_is_playing) 
		Mix_PauseMusic();
	
	//special case from mp3 format music
	if (Mix_GetMusicType(NULL) == MUS_MP3)
		Mix_RewindMusic();
	
	//~ if (elapsed_seconds<music_length) {	
		Mix_SetMusicPosition(elapsed_seconds); 
		Mix_ResumeMusic();
	//~ } else //music length too short
		//~ haltMusic();
}

void Audio::rewindMusic()
{
	Mix_RewindMusic();
	//~ printf("RewindMusic\n");
	//~ music_is_playing=1;
	//~ MediaState state;
	//~ state.playStateAudio = MediaState::PLAYING;
	//~ SharedData::Instance()->Media( state );
	//~ syncMusic();
}



void Audio::pauseMusic()
{
	if (track !=NULL) {
		Mix_PauseMusic();
		//~ MediaState state;
		//~ state.playStateAudio = MediaState::PAUSED;
		//~ SharedData::Instance()->Media( state );
		//~ printf("audio:pauseMusic() music_is_playing\n");
	}
}

void Audio::resumeMusic()
{
	if (track !=NULL)
		{
			Mix_ResumeMusic();
			//~ printf("audio:resumeMusic() music_is_playing\n");
			music_is_playing=1;
		}
	//~ else
		//~ printf("audio:resumeMusic() not music_is_playing\n");
	//~ MediaState state;
	//~ state.playStateAudio = MediaState::PLAYING;
	//~ SharedData::Instance()->Media( state );
	//syncMusic();
}

void Audio::haltMusic()
{
	Mix_HaltMusic();
	//~ MediaState state;
	//~ state.playStateAudio = MediaState::STOPPED;
	//~ SharedData::Instance()->Media( state );
	music_is_playing=0;
	elapsed_seconds=0.0;
}

void Audio::dropMusic()
{
	if (track !=NULL)
	{
	Mix_HaltMusic();
	//~ MediaState state;
	//~ state.playStateAudio = MediaState::STOPPED;
	//~ SharedData::Instance()->Media( state );
	Mix_FreeMusic(track);
	}
	track=NULL;
	music_is_playing=0;
	elapsed_seconds=0.0;
}

void Audio::decrement_volumeMusic(int value)
{
	master_volume -= value;
	if (master_volume < 0) master_volume = 0;
		Mix_VolumeMusic(master_volume);
	//~ MediaState state;
	//~ state.volume = master_volume;
	//~ SharedData::Instance()->Media( state );
//~ 
	//~ ScriptState sstate;
	//~ sstate.volume = master_volume;
	//~ SharedData::Instance()->Script( sstate );	
}

void Audio::increment_volumeMusic(int value)
{
	master_volume += value;
	if (master_volume > SDL_MIX_MAXVOLUME) master_volume = SDL_MIX_MAXVOLUME;
		Mix_VolumeMusic(master_volume);
	//~ MediaState state;
	//~ state.volume = master_volume;
	//~ SharedData::Instance()->Media( state );
//~ 
	//~ ScriptState sstate;
	//~ sstate.volume = master_volume;
	//~ SharedData::Instance()->Script( sstate );
}

void Audio::set_volumeMusic(int _value)
{
	master_volume=_value;
	if (master_volume > SDL_MIX_MAXVOLUME) master_volume = SDL_MIX_MAXVOLUME;	
	if (master_volume < 0) master_volume = 0;	
		Mix_VolumeMusic(master_volume);
	//~ MediaState state;
	//~ state.volume = master_volume;
	//~ SharedData::Instance()->Media( state );	
}






void Audio::loadChunk(std::string filename)
{
	for (int i=0;i<MAX_CHANNELS;i++)
	{
		if(sound[i] ==NULL)
			{
				sound[i]=Mix_LoadWAV(filename.c_str());
				if (sound[i] == NULL) {
					Log.write("Nothing to do because no existing sound with this name", cLog::eLOG_TYPE_ERROR);
					return;
				}
				sound_name[i]=filename.c_str();
				return;
			}
	}	
    Log.write("Nothing to do because no channel free", cLog::eLOG_TYPE_WARNING);
}

void Audio::playChunk()
{
	for (int i=0;i<MAX_CHANNELS;i++)
	{
		if(sound[i] !=NULL)
				Mix_PlayChannel(-1,sound[i],0);
	}	
}

void Audio::pauseChunk()
{
	Mix_Pause(-1);
}
	
void Audio::dropChunk(std::string filename)
{
	for (int i=0;i<MAX_CHANNELS;i++)
	{
		if(sound_name[i] == filename.c_str())
			{
				Mix_FreeChunk(sound[i]);
				sound[i]=NULL;
				sound_name[i] =="";
				return;
			}
	}	
    Log.write("Nothing to do because no existing sound with this name", cLog::eLOG_TYPE_WARNING);
}

void Audio::resumeChunk()
{
	Mix_Resume(-1);
}

void Audio::haltChunk()
{
	Mix_HaltChannel(-1);
}	
