/*
 * Copyright (C) 2003 Fabien Chereau
 * Copyright (C) 2009 Digitalis Education Solutions, Inc.
 * Copyright (C) 2013 of the LSS team
 * Copyright (C) 2014 of the LSS Team & Association Sirius
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Stellarium360 is a free open project of the LSS team
 * See the TRADEMARKS file for free open project usage requirements.
 *
 */

#ifndef _CORE_H_
#define _CORE_H_

// default font sizes
//~ #define FontSizeCardinalPoints 30.
//~ #define FontSizeSolarSystem 14.
//~ #define FontSizeGeneral 14.
//~ #define FontSizeConstellations 16.

#include <string>

#include "solarsystem.hpp"
#include "navigator.hpp"
#include "observer.hpp"
#include "projector.hpp"
#include "object.hpp"
#include "constellation_mgr.hpp"
#include "nebula_mgr.hpp"
#include "illuminate_mgr.hpp"
#include "atmosphere.hpp"
#include "tone_reproductor.hpp"
#include "utility.hpp"
#include "init_parser.hpp"
//#include "draw.hpp"
#include "landscape.hpp"
#include "meteor_mgr.hpp"
#include "sky_localizer.hpp"
//#include "loadingbar.hpp"
#include "image_mgr.hpp"
#include "callbacks.hpp"
#include "geodesic_grid.hpp"

#include "text_mgr.hpp"
#include "mCity_mgr.hpp"
#include "backup_mgr.hpp"
#include "audio.hpp"
#include "sdl_facade.hpp"
#include "milkyway.hpp"
#include "cardinals.hpp"
#include "skyline.hpp"
#include "skygrid.hpp"
#include "skyperson.hpp"
#include "model3D.hpp" //class model3D_mgr

#include "suntrace.hpp"
#include "shader.hpp"

//!  @brief Main class for application core processing.
//!
//! Manage all the objects to be used in the program.
//! This class is the main API of the program. It must be documented using doxygen.
class Core
{
public:
	//! Possible mount modes
	enum MOUNT_MODE { MOUNT_ALTAZIMUTAL, MOUNT_EQUATORIAL };

	//! Inputs are the locale directory and root directory and callback function for recording actions
	Core(SDLFacade *sdl, const string& LDIR, const string& DATA_ROOT, const mBoost::callback <void, string> & recordCallback);
	virtual ~Core();

	//! Init and load all main core components from the passed config file.
	void init(const InitParser& conf, const int viewW, const int viewH);

	//! Update all the objects with respect to the time.
	//! @param delta_time the time increment in ms.
	void update(int delta_time);

	//! Execute all the drawing functions
	//! @param delta_time the time increment in ms.
	//! @return Returns the max squared distance in pixels any single object has moved since the previous update.
	double draw(int delta_time);

	//! Get the name of the directory containing the data
	const string getDataDir(void) const {
		return dataRoot + "/data/";
	}
	
	//! Get the fullname of the directory containing the data user
	const string getUserDir(void) const {
		string homeDir = getenv("HOME");
		string CDIR = homeDir + "/." + APP_LOWER_NAME + "/";
		return CDIR;
	}

	//! Get the SDL_Window window
	SDL_Window *get_Window() {
		return window;
	}

	//! Get the fullname of the directory containing the fonts user
	const string getUserFontDir(void) const {
		string homeDir = getenv("HOME");
		string CDIR = homeDir + "/." + APP_LOWER_NAME + "/fonts/";
		return CDIR;
	}

	//! Get the name of the local script directory
	const string getScriptDir(void) const {
		return dataRoot + "/data/scripts/";
	}

	//! Get the name of the directory containing the data
	const string getLocaleDir(void) const {
		return localeDir;
	}

	//! Get the name of the root directory i.e the one containing the other main directories
	const string& getDataRoot() const {
		return dataRoot;
	}

	//! Set the sky culture from I18 name
	//! Returns false and doesn't change if skyculture is invalid
	bool setSkyCulture(const string& cultureName);

	//! Set the current sky culture from the passed directory
	bool setSkyCultureDir(const string& culturedir);

	string getSkyCultureDir() {
		return skyCultureDir;
	}

	//! Get the current sky culture I18 name
	string getSkyCulture() const {
		return skyloc->directoryToSkyCultureI18(skyCultureDir);
	}

	void setInitialSkyCulture() {
		printf("Culture %s\n",mBackup.initial_skyCulture.c_str());
		setSkyCultureDir(mBackup.initial_skyCulture);
	}
	
	void setInitialSkyLocale() {
		printf("Locale %s\n",mBackup.initial_skyLocale.c_str());
		setSkyLanguage(mBackup.initial_skyLocale);
	}

	//! Get the I18 available sky culture names
	string getSkyCultureListI18() const {
		return skyloc->getSkyCultureListI18();
	}

	string getSkyCultureHash() const {
		return skyloc->getSkyCultureHash();
	}

	bool loadSkyCulture(const string& culturePath);

	//! Set the landscape
	bool setLandscape(const string& new_landscape_name);

	//! Load a landscape based on a hash of parameters mirroring the landscape.ini file
	//! and make it the current landscape
	bool loadLandscape(stringHash_t& param);

	//! @brief Set the sky language and reload the sky objects names with the new translation
	//! This function has no permanent effect on the global locale
	//!@param newSkyLocaleName The name of the locale (e.g fr) to use for sky object labels
	void setSkyLanguage(const string& newSkyLocaleName);

	//! Get the current sky language used for sky object labels
	//! @return The name of the locale (e.g fr)
	string getSkyLanguage() {
		return skyTranslator.getLocaleName();
	}

	///////////////////////////////////////////////////////////////////////////////////////
	// Navigation
	//! Set time speed in JDay/sec
	void setTimeSpeed(double ts) {
		navigation->set_time_speed(ts);
	}
	//! Get time speed in JDay/sec
	double getTimeSpeed(void) const {
		return navigation->get_time_speed();
	}

	//! Set simulation time to current real world time
	void setTimeNow();
	//! Get wether the current simulation time is the real world time
	bool getIsTimeNow(void) const;


	//! Set the current date in Julian Day
	void setJDay(double JD) {
		navigation->set_JDay(JD);
	}
	//! Get the current date in Julian Day
	double getJDay(void) const {
		return navigation->get_JDay();
	}

	//! Set object tracking
	void setFlagTracking(bool b);
	//! Get object tracking
	bool getFlagTracking(void) {
		return navigation->get_flag_traking();
	}

	//! Set whether sky position is to be locked
	void setFlagLockSkyPosition(bool b) {
		navigation->set_flag_lock_equ_pos(b);
	}
	//! Set whether sky position is locked
	bool getFlagLockSkyPosition(void) {
		return navigation->get_flag_lock_equ_pos();
	}

	//! Set current mount type
	void setMountMode(MOUNT_MODE m) {
		navigation->set_viewing_mode((m==MOUNT_ALTAZIMUTAL) ? Navigator::VIEW_HORIZON : Navigator::VIEW_EQUATOR);
	}
	//! Get current mount type
	MOUNT_MODE getMountMode(void) {
		return ((navigation->get_viewing_mode()==Navigator::VIEW_HORIZON) ? MOUNT_ALTAZIMUTAL : MOUNT_EQUATORIAL);
	}
	//! Toggle current mount mode between equatorial and altazimutal
	void toggleMountMode(void) {
		if (getMountMode()==MOUNT_ALTAZIMUTAL) setMountMode(MOUNT_EQUATORIAL);
		else setMountMode(MOUNT_ALTAZIMUTAL);
	}

	// TODO!
	void loadObservatory();

	//! Go to the selected object
	void gotoSelectedObject(void) {
		if (selected_object)
			navigation->move_to(
			    selected_object.get_earth_equ_pos(navigation),
			    auto_move_duration);
	}

	//! Move view in alt/az (or equatorial if in that mode) coordinates
	void panView(double delta_az, double delta_alt)	{
		setFlagTracking(0);
		navigation->update_move(projection, delta_az, delta_alt, projection->get_fov());
	}


	void setViewOffset(double offset);
	double getViewOffset();
	void setHeading(double heading, int duration);
	double getHeading();

	//! Set flight duration in seconds between bodies
	void setFlightDuration(float f) {
		flight_duration = f;
		//~ SettingsState state;
		//~ state.m_state.flyto_duration = f;
		//~ SharedData::Instance()->Settings( state );
	}
	//! Get flight duration in seconds
	float getFlightDuration(void) const {
		return flight_duration;
	}


	//! Set automove duration in seconds
	void setAutomoveDuration(float f) {
		auto_move_duration = f;
		//~ SettingsState state;
		//~ state.m_state.auto_move_duration = f;
		//~ SharedData::Instance()->Settings( state );
	}
	//! Get automove duration in seconds
	float getAutomoveDuration(void) const {
		return auto_move_duration;
	}

	//! Zoom to the given FOV (in degree)
	void zoomTo(double aim_fov, float move_duration = 1.) {
		projection->zoom_to(aim_fov, move_duration);
	}

	//! Get current FOV (in degree)
	float getFov(void) const {
		return projection->get_fov();
	}

	//! If is currently zooming, return the target FOV, otherwise return current FOV
	double getAimFov(void) const {
		return projection->getAimFov();
	}

	//! Set the current FOV (in degree)
	void setFov(double f) {
		projection->set_fov(f);
	}

	//! Set the maximum FOV (in degree)
	void setMaxFov(double f) {
		projection->setMaxFov(f);
	}

	//! Go and zoom temporarily to the selected object.
	void autoZoomIn(float move_duration = 1.f, bool allow_manual_zoom = 1);

	//! Unzoom to the previous position
	void autoZoomOut(float move_duration = 1.f, bool full = 0, bool allow_manual_zoom = 0);

	//! Set whether auto zoom can go further than normal
	void setFlagManualAutoZoom(bool b) {
		FlagManualZoom = b;
		//~ SettingsState state;
		//~ state.m_state.manual_zoom = b;
		//~ SharedData::Instance()->Settings( state );

		//~ ObjectsState ostate;
		//~ ostate.m_state.manual_zoom = b;
		//~ SharedData::Instance()->Objects( ostate );
	}
	//! Get whether auto zoom can go further than normal
	bool getFlagManualAutoZoom(void) {
		return FlagManualZoom;
	}

	// Viewing direction function : 1 move, 0 stop.
	void turn_right(int);
	void turn_left(int);
	void turn_up(int);
	void turn_down(int);
	void zoom_in(int);
	void zoom_out(int);
	void raise_height(int);
	void lower_height(int);

	//! Make the first screen position correspond to the second (useful for mouse dragging)
	void dragView(int x1, int y1, int x2, int y2);

	//! Find and select an object near given equatorial position
	//! @return true if a object was found at position (this does not necessarily means it is selected)
	bool findAndSelect(const Vec3d& pos);

	//! Find and select an object near given screen position
	//! @return true if a object was found at position (this does not necessarily means it is selected)
	bool findAndSelect(int x, int y);

	//! Find and select an object from its translated name
	//! @param nameI18n the case sensitive object translated name
	//! @return true if a object was found with the passed name
	bool findAndSelectI18n(const string &nameI18n);

	//! Find and select an object based on selection type and standard name or number
	//! @return true if an object was selected
	bool selectObject(const string &type, const string &id);


	//! Find and return the list of at most maxNbItem objects auto-completing the passed object I18n name
	//! @param objPrefix the case insensitive first letters of the searched object
	//! @param maxNbItem the maximum number of returned object names
	//! @return a vector of matching object name by order of relevance, or an empty vector if nothing match
	vector<string> listMatchingObjectsI18n(const string& objPrefix, unsigned int maxNbItem=5) const;

	//! Return whether an object is currently selected
	bool getFlagHasSelected(void) {
		return selected_object;
	}

	//! Deselect selected object if any
	//! Does not deselect selected constellation
	void unSelect(void) {
		selected_object=NULL;
		//asterisms->setSelected(NULL);
		ssystem->setSelected(Object());
		// do not:  setFlagTracking(0);  // do not put in svn 200703
		// stay looking at same area of sky
	}

	void unsetSelectedConstellation(string constellation) {
		asterisms->unsetSelected(constellation);
	}

	void deselect(void);

	//! Set whether a pointer is to be drawn over selected object
	void setFlagSelectedObjectPointer(bool b) {
		object_pointer_visibility = b;
	}

	string getSelectedPlanetEnglishName() const;

	//! Get a multiline string describing the currently selected object
	string getSelectedObjectInfo(void) const {
		return selected_object.getInfoString(navigation);
	}

	void getDeRa(double *ra, double *de) const {
		selected_object.getRaDeValue(navigation,ra,de);
	}

	//! Get a 1 line string briefly describing the currently selected object
	string getSelectedObjectShortInfo(void) const {
		return selected_object.getShortInfoString(navigation);
	}

	//! Get a 1 line string briefly describing the currently NAV eddition selected object
	string getSelectedObjectShortInfoNav(void) const {
		return selected_object.getShortInfoNavString(navigation);
	}


	//! Get a color used to display info about the currently selected object
	Vec3f getSelectedObjectInfoColor(void) const;


	///////////////////////////////////////////////////////////////////////////////////////
	// Rendering settings

	//! Set rendering flag of antialiased lines
	void setFlagAntialiasLines(bool b) {
		FlagAntialiasLines = b;
		//~ SettingsState state;
		//~ state.m_state.antialias_lines = b;
		//~ SharedData::Instance()->Settings( state );

		if(b) glEnable(GL_LINE_SMOOTH);
		else glDisable(GL_LINE_SMOOTH);
	}
	//! Get display flag of constellation lines
	bool getFlagAntialiasLines(void) {
		return FlagAntialiasLines;
	}

	void setLineWidth(float w) {
		m_lineWidth = w;
		//~ SettingsState state;
		//~ state.m_state.line_width = w;
		//~ SharedData::Instance()->Settings( state );
	}
	float getLineWidth() {
		return m_lineWidth;
	}

	//~ //! Enable shaders
	//~ void setFlagShaders(bool b) {
		//~ FlagShaders = b;
		//~ ssystem->setFlagShaders(b);
	//~ }
	//~ //! Get flag of shaders
	//~ bool getFlagShaders(void) {
		//~ return FlagShaders;
	//~ }

	///////////////////////////////////////////////////////////////////////////////////////
	// Constellations methods
	//! Set display flag of constellation lines
	void setFlagConstellationLines(bool b) {
		asterisms->setFlagLines(b);
	}
	//! Get display flag of constellation lines
	bool getFlagConstellationLines(void) {
		return asterisms->getFlagLines();
	}

	//! Set display flag of constellation art
	void setFlagConstellationArt(bool b) {
		asterisms->setFlagArt(b);
	}
	//! Get display flag of constellation art
	bool getFlagConstellationArt(void) {
		return asterisms->getFlagArt();
	}

	//! Set display flag of constellation names
	void setFlagConstellationNames(bool b) {
		asterisms->setFlagNames(b);
	}
	//! Get display flag of constellation names
	bool getFlagConstellationNames(void) {
		return asterisms->getFlagNames();
	}

	//! Set display flag of constellation boundaries
	void setFlagConstellationBoundaries(bool b) {
		asterisms->setFlagBoundaries(b);
	}
	//! Get display flag of constellation boundaries
	bool getFlagConstellationBoundaries(void) {
		return asterisms->getFlagBoundaries();
	}
	Vec3f getColorConstellationBoundaries(void) const {
		return asterisms->getBoundaryColor();
	}

	//! Set constellation art intensity
	void setConstellationArtIntensity(float f) {
		asterisms->setArtIntensity(f);
	}
	//! Get constellation art intensity
	float getConstellationArtIntensity(void) const {
		return asterisms->getArtIntensity();
	}

	//! Set constellation art intensity
	void setConstellationArtFadeDuration(float f) {
		asterisms->setArtFadeDuration(f);
	}
	//! Get constellation art intensity
	float getConstellationArtFadeDuration(void) const {
		return asterisms->getArtFadeDuration();
	}

	//! Set whether selected constellation is drawn alone
	void setFlagConstellationIsolateSelected(bool b) {
		asterisms->setFlagIsolateSelected(b);
	}
	//! Get whether selected constellation is drawn alone
	bool getFlagConstellationIsolateSelected(void) {
		return asterisms->getFlagIsolateSelected();
	}

	//! Get constellation line color
	Vec3f getColorConstellationLine() const {
		return asterisms->getLineColor();
	}
	//! Set constellation line color
	void setColorConstellationLine(const Vec3f& v) {
		asterisms->setLineColor(v);
	}

	//! Get constellation names color
	Vec3f getColorConstellationNames() const {
		return asterisms->getLabelColor();
	}
	//! Set constellation names color
	void setColorConstellationNames(const Vec3f& v) {
		asterisms->setLabelColor(v);
	}

	//! Get constellation art color
	Vec3f getColorConstellationArt() const {
		return asterisms->getArtColor();
	}
	//! Set constellation line color
	void setColorConstellationArt(const Vec3f& v) {
		asterisms->setArtColor(v);
	}

	///////////////////////////////////////////////////////////////////////////////////////
	// Stars methods
	//! Set display flag for Stars
	void setFlagStars(bool b);
	//! Get display flag for Stars
	bool getFlagStars(void) const;

	//! Set display flag for Star names. Also make sure that stars are on if want labels
	void setFlagStarName(bool b);
	//! Get display flag for Star names
	bool getFlagStarName(void) const;

	//! Set display flag for Star Scientific names
	void setFlagStarSciName(bool b);
	//! Get display flag for Star Scientific names
	bool getFlagStarSciName(void) const;

	//! Set flag for Star twinkling
	void setFlagStarTwinkle(bool b);
	//! Get flag for Star twinkling
	bool getFlagStarTwinkle(void) const;

	//! Set flag for displaying Star as GLpoints (faster but not so nice)
	void setFlagPointStar(bool b);
	//! Get flag for displaying Star as GLpoints (faster but not so nice)
	bool getFlagPointStar(void) const;

	//! Set maximum magnitude at which stars names are displayed
	void setMaxMagStarName(float f);
	//! Get maximum magnitude at which stars names are displayed
	float getMaxMagStarName(void) const;

	//! Set maximum magnitude at which stars scientific names are displayed
	void setMaxMagStarSciName(float f);
	//! Get maximum magnitude at which stars scientific names are displayed
	float getMaxMagStarSciName(void) const;

	//! Set base stars display scaling factor
	void setStarScale(float f);
	//! Get base stars display scaling factor
	float getStarScale(void) const;

	//! set the time hipstars fader
	void setStarsDuration(float f);
	
	//! Set base planets display scaling factor
	//! This is additive to star size limit above
	//! since makes no sense to be less
	//! ONLY SET THROUGH THIS METHOD
	void setPlanetsSizeLimit(float f);

	//! Get base planets display scaling factor
	float getPlanetsSizeLimit(void) const {
		return (ssystem->getSizeLimit()-getStarSizeLimit());
	}

	float getStarSizeLimit(void) const;
	void setStarSizeLimit(float);

	float getStarLimitingMag(void) const;
	void setStarLimitingMag(float);

	//! Set stars display scaling factor wrt magnitude
	void setStarMagScale(float f);
	//! Get base stars display scaling factor wrt magnitude
	float getStarMagScale(void) const;

	//! Set stars twinkle amount
	void setStarTwinkleAmount(float f);
	//! Get stars twinkle amount
	float getStarTwinkleAmount(void) const;

	float getMagConverterMagShift(void) const;
	void setMagConverterMagShift(float f);
	float getMagConverterMaxScaled60DegMag(void) const;
	void setMagConverterMaxScaled60DegMag(float f);
	float getMagConverterMaxFov(void) const;
	void setMagConverterMaxFov(float f);
	float getMagConverterMinFov(void) const;
	void setMagConverterMinFov(float f);
	float getMagConverterMaxMag(void) const;
	void setMagConverterMaxMag(float f);

	Vec3f getColorStarNames(void) const;
	Vec3f getColorStarCircles(void) const;
	void setColorStarNames(const Vec3f &v);
	void setColorStarCircles(const Vec3f &v);

	///////////////////////////////////////////////////////////////////////////////////////
	// Planets flags
	//! Set flag for displaying Planets
	void setFlagPlanets(bool b) {
		ssystem->setFlagPlanets(b);
	}
	//! Get flag for displaying Planets
	bool getFlagPlanets(void) const {
		return ssystem->getFlagPlanets();
	}

	//! Set flag for displaying Planets Trails
	void setFlagPlanetsTrails(bool b) {
		ssystem->setFlagTrails(b);
	}
	//! Get flag for displaying Planets Trails
	bool getFlagPlanetsTrails() const {
		return ssystem->getFlagTrails();
	}

	//! Set flag for displaying Planets Axis
	void setFlagPlanetsAxis(bool b) {
		ssystem->setFlagAxis(b);
	}
	//! Get flag for displaying Planets Axis
	bool getFlagPlanetsAxis(void) const {
		return ssystem->getFlagAxis();
	}


	//! Set flag for displaying Planets Hints
	void setFlagPlanetsHints(bool b) {
		ssystem->setFlagHints(b);
	}
	//! Get flag for displaying Planets Hints
	bool getFlagPlanetsHints(void) const {
		return ssystem->getFlagHints();
	}

	//! Set flag for displaying Planets Orbits
	void setFlagPlanetsOrbits(bool b) {
		ssystem->setFlagPlanetsOrbits(b);
	}
	//! Get flag for displaying Planets Orbits
	bool getFlagPlanetsOrbits(void) const {
		return ssystem->getFlagOrbits();
	}
	
	//! Set flag for displaying Satellites Orbits
	void setFlagSatellitesOrbits(bool b) {
		ssystem->setFlagSatellitesOrbits(b);
	}
	
	//! Get flag for displaying Satellites Orbits
	bool getFlagSatellitesOrbits(void) const {
		return ssystem->getFlagSatellitesOrbits();
	}
		//! Set flag for displaying Planets & Satellites Orbits
	void setFlagPlanetOrbits(bool b) {
		ssystem->setFlagOrbits(b);
	}
	
	//! Get flag for displaying Planets & Satellites Orbits
	bool getFlagPlanetOrbits(void) const {
		return ssystem->getFlagOrbits();
	}
	
	void setFlagLightTravelTime(bool b) {
		ssystem->setFlagLightTravelTime(b);
	}
	bool getFlagLightTravelTime(void) const {
		return ssystem->getFlagLightTravelTime();
	}

// set & get planet trail, orbit & label

	void setColorPlanetOrbits(string englishName, Vec3f c) const {
		ssystem->setPlanetOrbitColor(englishName, c);
	}

	Vec3f getColorPlanetOrbits(string englishName) const {
		return ssystem->getPlanetOrbitColor(englishName);
	}

	void setColorPlanetName(string englishName, Vec3f c) const {
		ssystem->setPlanetLabelColor(englishName, c);
	}

	Vec3f getColorPlanetName(string englishName) const {
		return ssystem->getPlanetLabelColor(englishName);
	}

	void setColorPlanetTrail(string englishName, Vec3f c) const {
		ssystem->setPlanetTrailColor(englishName, c);
	}

	Vec3f getColorPlanetTrail(string englishName) const {
		return ssystem->getPlanetTrailColor(englishName);
	}
	
//default planet param's color

	void setDefaultColorPlanetOrbits(const Vec3f& c)  {
		ssystem->setDefaultPlanetOrbitColor(c);
	}

	Vec3f getDefaultColorPlanetOrbits(void ) const {
		return ssystem->getDefaultPlanetOrbitColor();
	}

	void setDefaultColorPlanetNames(const Vec3f& c)  {
		ssystem->setDefaultPlanetLabelColor(c);
	}

	Vec3f getDefaultColorPlanetNames(void) const {
		return ssystem->getDefaultPlanetLabelColor();
	}

	void setDefaultColorPlanetTrails(const Vec3f& c)  {
		ssystem->setDefaultPlanetTrailColor(c);
	}

	Vec3f getDefaultColorPlanetTrails(void) const {
		return ssystem->getDefaultPlanetTrailColor();
	}


	//! Start/stop displaying planets Trails
	void startPlanetsTrails(bool b) {
		ssystem->startTrails(b);
	}

	//! Set base planets display scaling factor
	void setPlanetsScale(float f) {
		ssystem->setScale(f);
	}
	//! Get base planets display scaling factor
	float getPlanetsScale(void) const {
		return ssystem->getScale();
	}

	//! Set planets viewer scaling factor
	void setPlanetSizeScale(string name, float f) {
		ssystem->setPlanetSizeScale(name, f);
	}
	
	//! Get planets viewer scaling factor
	float getPlanetSizeScale(string name) {
		return ssystem->getPlanetSizeScale(name);
	}

	//! Set selected planets by englishName
	//! @param englishName The planet name or "" to select no planet
	void setPlanetsSelected(const string& englishName) {
		ssystem->setSelected(englishName);
	}

	string getPlanetHashString(void);

	bool setHomePlanet(string planet, float transit_time=0.f);

	//! Set flag for displaying a scaled Moon
	void setFlagMoonScaled(bool b) {
		ssystem->setFlagMoonScale(b);
	}
	//! Get flag for displaying a scaled Moon
	bool getFlagMoonScaled(void) const {
		return ssystem->getFlagMoonScale();
	}

	//! Set flag for displaying a scaled Sun
	void setFlagSunScaled(bool b) {
		ssystem->setFlagSunScale(b);
	}
	//! Get flag for displaying a scaled Sun
	bool getFlagSunScaled(void) const {
		return ssystem->getFlagSunScale();
	}
	
	//! Set Moon scale
	void setMoonScale(float f) {
		if (f<0) ssystem->setMoonScale(1.);
		else ssystem->setMoonScale(f);
	}
	//! Get Moon scale
	float getMoonScale(void) const {
		return ssystem->getMoonScale();
	}

	//! Set Sun scale
	void setSunScale(float f) {
		if (f<0) ssystem->setSunScale(1.);
		else ssystem->setSunScale(f);
	}
	//! Get Moon scale
	float getSunScale(void) const {
		return ssystem->getSunScale();
	}


	//! Set flag for displaying clouds (planet rendering feature)
	void setFlagClouds(bool b) {
		ssystem->setFlagClouds(b);
	}
	//! Get flag for displaying Atmosphere
	bool getFlagClouds(void) const {
		return ssystem->getFlagClouds();
	}


	void setJDayRelativeYear(int yy);

// for adding planets
	string addSolarSystemBody(stringHash_t& param);

	string removeSolarSystemBody(string name);

	string removeSupplementalSolarSystemBodies();

	void initialSolarSystemBodies() {
		return ssystem->initialSolarSystemBodies();
	}

	void setPlanetHidden(string name, bool planethidden);

	///////////////////////////////////////////////////////////////////////////////////////
	// Grid and lines
	//! Set flag for displaying Azimutal Grid
	void setFlagAzimutalGrid(bool b) {
		azi_grid->setFlagshow(b);
	}
	//! Get flag for displaying Azimutal Grid
	bool getFlagAzimutalGrid(void) const {
		return azi_grid->getFlagshow();
	}
	Vec3f getColorAzimutalGrid(void) const {
		return azi_grid->getColor();
	}

	//! Set flag for displaying Equatorial Grid
	void setFlagEquatorGrid(bool b) {
		equ_grid->setFlagshow(b);
	}
	void setFlagEclipticGrid(bool b) {
		ecliptic_grid->setFlagshow(b);
	}
    void setFlagGalacticGrid(bool b) {
		galactic_grid->setFlagshow(b);
	}


	//! Get flag for displaying Equatorial Grid
	bool getFlagEquatorGrid(void) const {
		return equ_grid->getFlagshow();
	}
	Vec3f getColorEquatorGrid(void) const {
		return equ_grid->getColor();
	}

    bool getFlagEclipticGrid(void) const {
		return ecliptic_grid->getFlagshow();
	}
	Vec3f getColorEclipticGrid(void) const {
		return ecliptic_grid->getColor();
	}
	bool getFlagGalacticGrid(void) const {
		return galactic_grid->getFlagshow();
	}
	Vec3f getColorGalacticGrid(void) const {
		return galactic_grid->getColor();
	}

	//! Set flag for displaying Equatorial Line
	void setFlagEquatorLine(bool b) {
	//~ if (flagNav)
		//~ equator_nav_line->setFlagshow(b);
	//~ else
		equator_line->setFlagshow(b);
	}
	void setFlagGalacticLine(bool b) {
		galactic_line->setFlagshow(b);
	}
	//! Get flag for displaying Equatorial Line
	bool getFlagEquatorLine(void) const {
		//~ if (flagNav)
			//~ return equator_nav_line->getFlagshow();
		//~ else
			return equator_line->getFlagshow();
	}
	bool getFlagGalacticLine(void) const {
		return galactic_line->getFlagshow();
	}
	Vec3f getColorEquatorLine(void) const {
		return equator_line->getColor();
	}
	Vec3f getColorGalacticLine(void) const {
		return galactic_line->getColor();
	}

	//! Set flag for displaying Ecliptic Line
	void setFlagEclipticLine(bool b) {
		ecliptic_line->setFlagshow(b);
	}
	//! Get flag for displaying Ecliptic Line
	bool getFlagEclipticLine(void) const {
		return ecliptic_line->getFlagshow();
	}
	Vec3f getColorEclipticLine(void) const {
		return ecliptic_line->getColor();
	}

	//! Set flag for displaying Precession Circle
	void setFlagPrecessionCircle(bool b) {
		precession_circle->setFlagshow(b);
	}
	//! Get flag for displaying Precession Circle
	bool getFlagPrecessionCircle(void) const {
		return precession_circle->getFlagshow();
	}
	Vec3f getColorPrecessionCircle(void) const {
		return precession_circle->getColor();
	}

	//! Set flag for displaying Circumpolar Circle
	void setFlagCircumpolarCircle(bool b) {
		circumpolar_circle->setFlagshow(b);
	}
	//! Get flag for displaying Circumpolar Circle
	bool getFlagCircumpolarCircle(void) const {
		return circumpolar_circle->getFlagshow();
	}
	Vec3f getColorCircumpolarCircle(void) const {
		return circumpolar_circle->getColor();
	}

	//! Set flag for displaying Tropic Lines
	void setFlagTropicLines(bool b) {
		tropic_line->setFlagshow(b);
	}
	//! Get flag for displaying Tropic Lines
	bool getFlagTropicLines(void) const {
		return tropic_line->getFlagshow();
	}
	Vec3f getColorTropicLine(void) const {
		return tropic_line->getColor();
	}

	//! Set flag for displaying Meridian Line
	void setFlagMeridianLine(bool b) {
		meridian_line->setFlagshow(b);
	}
	//! Get flag for displaying Meridian Line
	bool getFlagMeridianLine(void) const {
		return meridian_line->getFlagshow();
	}
	Vec3f getColorMeridianLine(void) const {
		return meridian_line->getColor();
	}

	//! Set flag for displaying zenith
	void setFlagZenithLine(bool b) {
		zenith_line->setFlagshow(b);
	}
	//! Get flag for displaying zenit 
	bool getFlagZenithLine(void) const {
		return zenith_line->getFlagshow();
	}	
	Vec3f getColorZenithLine(void) const {
		return zenith_line->getColor();
	}

	//! Set flag for displaying polar circle
	void setFlagPolarCircle(bool b) {
		polar_circle->setFlagshow(b);
	}
	//! Get flag for displaying polar circle
	bool getFlagPolarCircle(void) const {
		return polar_circle->getFlagshow();
	}		
	Vec3f getColorPolarCircle(void) const {
		return polar_circle->getColor();
	}

	void setFlagEclipticCenter(bool b) {
		ecliptic_center->setFlagshow(b);
	}
	//! Get flag for displaying polar circle
	bool getFlagEclipticCenter(void) const {
		return ecliptic_center->getFlagshow();
	}		
	Vec3f getColorEclipticCenter(void) const {
		return ecliptic_center->getColor();
	}
	void setFlagGalacticPole(bool b) {
		galactic_pole->setFlagshow(b);
	}
	//! Get flag for displaying polar circle
	bool getFlagGalacticPole(void) const {
		return galactic_pole->getFlagshow();
	}		
	Vec3f getColorGalacticPole(void) const {
		return galactic_pole->getColor();
	}
    void setFlagGalacticCenter(bool b) {
		galactic_center->setFlagshow(b);
	}
	//! Get flag for displaying polar circle
	bool getFlagGalacticCenter(void) const {
		return galactic_center->getFlagshow();
	}		
	Vec3f getColorGalacticCenter(void) const {
		return galactic_center->getColor();
	}
	void setFlagVernalPoints(bool b) {
		vernal_points->setFlagshow(b);
	}
	//! Get flag for displaying polar circle
	bool getFlagVernalPoints(void) const {
		return vernal_points->getFlagshow();
	}		
	Vec3f getColorVernalPoints(void) const {
		return vernal_points->getColor();
	}


	//! Set flag for displaying zenit
	void setFlagAnalemmaLine(bool b) {
		analemma_line->setFlagshow(b);
	}
	//! Get flag for displaying zenit 
	bool getFlagAnalemmaLine(void) const {
		return analemma_line->getFlagshow();
	}
	Vec3f getColorAnalemmaLine(void) const {
		return analemma_line->getColor();
	}

	//! Set flag for displaying polar circle
	void setFlagAnalemma(bool b) {
		analemma->setFlagshow(b);
	}
	//! Get flag for displaying polar circle
	bool getFlagAnalemma(void) const {
		return analemma->getFlagshow();
	}
	Vec3f getColorAnalemma(void) const {
		return analemma->getColor();
	}

	//! Set flag for displaying Aries Line
	void setFlagAriesLine(bool b) {
		aries_line->setFlagshow(b);
	}
	//! Get flag for displaying Aries Line
	bool getFlagAriesLine(void) const {
		return aries_line->getFlagshow();
	}
	Vec3f getColorAriesLine(void) const {
		return aries_line->getColor();
	}

	//! Set flag for displaying Zodiac Line
	void setFlagZodiac(bool b) {
		zodiac->setFlagshow(b);
	}
	//! Get flag for displaying Zodiac Line
	bool getFlagZodiac(void) const {
		return zodiac->getFlagshow();
	}
	Vec3f getColorZodiac(void) const {
		return zodiac->getColor();
	}

	//! Set flag for displaying Personal Line
	void setFlagPersonal(bool b) {
		personal->setFlagshow(b);
	}
	//! Get flag for displaying Personal Line
	bool getFlagPersonal(void) const {
		return personal->getFlagshow();
	}
	Vec3f getColorPersonal(void) const {
		return personal->getColor();
	}

	//! Set flag for displaying Personeq Line
	void setFlagPersoneq(bool b) {
		personeq->setFlagshow(b);
	}
	//! Get flag for displaying Personeq Line
	bool getFlagPersoneq(void) const {
		return personeq->getFlagshow();
	}
	Vec3f getColorPersoneq(void) const {
		return personeq->getColor();
	}

	void personal_load(string filename) const {
		personal->loadData(filename);
	}
	
	void personal_clear() const {
		personal->clear();
	};

	void personeq_load(string filename) const {
		personeq->loadData(filename);
	}
	
	void personeq_clear() const {
		personeq->clear();
	};



	//! Set flag for displaying Greenwich Line
	void setFlagGreenwichLine(bool b) {
		greenwich_line->setFlagshow(b);
	}
	//! Get flag for displaying Greenwich Line
	bool getFlagGreenwichLine(void) const {
		return greenwich_line->getFlagshow();
	}
	Vec3f getColorGreenwichLine(void) const {
		return greenwich_line->getColor();
	}

	//! Set flag for displaying Vertical Line
	void setFlagVerticalLine(bool b) {
		vertical_line->setFlagshow(b);
	}
	//! Get flag for displaying Vertical Line
	bool getFlagVerticalLine(void) const {
		return vertical_line->getFlagshow();
	}
	Vec3f getColorVerticalLine(void) const {
		return vertical_line->getColor();
	}
	
	//! Set flag for displaying Cardinals Points
	void setFlagCardinalsPoints(bool b) {
		cardinals_points->setFlagShow(b);
	}
	//! Get flag for displaying Cardinals Points
	bool getFlagCardinalsPoints(void) const {
		return cardinals_points->getFlagShow();
	}

	//! Set Cardinals Points color
	void setColorCardinalPoints(const Vec3f& v) {
		cardinals_points->setColor(v);
	}
	//! Get Cardinals Points color
	Vec3f getColorCardinalPoints(void) const {
		return cardinals_points->get_color();
	}

	///////////////////////////////////////////////////////////////////////////////////////
	// Colors

	void setColorConstellationBoundaries(const Vec3f& v) {
		asterisms->setBoundaryColor(v);
	}

	void setColorPlanetsNames(string englishName, const Vec3f& v) {
		ssystem->setPlanetLabelColor(englishName,v);
	}
	void setColorPlanetsTrails(string englishName, const Vec3f& v) {
		ssystem->setPlanetTrailColor(englishName, v);
	}
	void setColorAzimutalGrid(const Vec3f& v) {
		azi_grid->setColor(v);
	}
	void setColorEquatorGrid(const Vec3f& v) {
		equ_grid->setColor(v);
	}
	void setColorEquatorLine(const Vec3f& v) {
		equator_line->setColor(v);
		//~ equator_nav_line->setColor(v);
		tropic_line->setColor(v);
	}
	void setColorGalacticLine(const Vec3f& v) {
		galactic_line->setColor(v);
	}
	void setColorEclipticGrid(const Vec3f& v) {
		ecliptic_grid->setColor(v);
	}
	void setColorGalacticGrid(const Vec3f& v) {
		galactic_grid->setColor(v);
	}

	void setColorEclipticLine(const Vec3f& v) {
		ecliptic_line->setColor(v);
	}
	void setColorPrecessionCircle(const Vec3f& v) {
		precession_circle->setColor(v);
	}
	void setColorCircumpolarCircle(const Vec3f& v) {
		circumpolar_circle->setColor(v);
	}
	void setColorTropicLine(const Vec3f& v) {
		tropic_line->setColor(v);
	}
	void setColorMeridianLine(const Vec3f& v) {
		meridian_line->setColor(v);
	}
	void setColorZenithLine(const Vec3f& v) {
		zenith_line->setColor(v);
	}
	void setColorPolarCircle(const Vec3f& v) {
		polar_circle->setColor(v);
	}
    void setColorEclipticCenter(const Vec3f& v) {
		ecliptic_center->setColor(v);
	}
    void setColorGalacticPole(const Vec3f& v) {
		galactic_pole->setColor(v);
	}
  void setColorGalacticCenter(const Vec3f& v) {
		galactic_center->setColor(v);
	}
void setColorVernalPoints(const Vec3f& v) {
		vernal_points->setColor(v);
	}

	void setColorAnalemmaLine(const Vec3f& v) {
		analemma_line->setColor(v);
	}
	void setColorAnalemma(const Vec3f& v) {
		analemma->setColor(v);
	}
	void setColorAriesLine(const Vec3f& v) {
		aries_line->setColor(v);
	}
	void setColorZodiac(const Vec3f& v) {
		zodiac->setColor(v);
	}
	void setColorPersonal(const Vec3f& v) {
		personal->setColor(v);
	}
	void setColorPersoneq(const Vec3f& v) {
		personeq->setColor(v);
	}
	void setColorGreenwichLine(const Vec3f& v) {
		greenwich_line->setColor(v);
	}
	void setColorVerticalLine(const Vec3f& v) {
		vertical_line->setColor(v);
	}
	void setColorNebulaLabels(const Vec3f& v) {
		nebulas->setLabelColor(v);
	}
	void setColorNebulaCircle(const Vec3f& v) {
		nebulas->setCircleColor(v);
	}

	//! set flag to display generic Hint or specific DSO type
	void setDsoPictograms (bool value) {
		nebulas->setDisplaySpecificHint(value);
	}
	//! get flag to display generic Hint or specific DSO type
	bool getDsoPictograms () {
		return nebulas->getDisplaySpecificHint();
	}
	
	//!set Flag DSO Name who display DSO name 
	void setFlagNebulaNames (bool value) {
		nebulas->setNebulaNames(value);
	}
	
	//!get flag DSO Name who display DSO name
	bool getFlagNebulaNames () {
		return nebulas->getNebulaNames();
	}

	bool loadNebula(double ra, double de, double magnitude, double angular_size, double rotation,
					string name, string filename, string credit, double texture_luminance_adjust,
					double distance);

	//! remove one nebula added by user
	string removeNebula(const string& name);

	//! remove all user added nebulae
	string removeSupplementalNebulae();

	bool loadIlluminate(double ra, double de, double angular_size, string name, double r, double g, double b);

	string removeIlluminate(const string& name);
	string removeAllIlluminate();
	
	void setIlluminateSize (double value) {
		illuminate_size=value;
	}

	///////////////////////////////////////////////////////////////////////////////////////
	// Projection
	//! Set the horizontal viewport offset in pixels
	void setViewportHorizontalOffset(int hoff) const {
		projection->setViewportPosX(hoff);
	}
	//! Get the horizontal viewport offset in pixels
	int getViewportHorizontalOffset(void) const {
		return projection->getViewportPosX();
	}

	//! Set the vertical viewport offset in pixels
	void setViewportVerticalOffset(int voff) const {
		projection->setViewportPosY(voff);
	}
	//! Get the vertical viewport offset in pixels
	int getViewportVerticalOffset(void) const {
		return projection->getViewportPosY();
	}

	//! Maximize viewport according to passed screen values
	void setMaximizedViewport(int screenW, int screenH) {
		projection->setViewport(0, 0, screenW, screenH);
	}

	//! Set a centered squared viewport with passed vertical and horizontal offset
	void setSquareViewport(int screenW, int screenH, int hoffset, int voffset) {
		int m = MY_MIN(screenW, screenH);
		projection->setViewport((screenW-m)/2+hoffset, (screenH-m)/2+voffset, m, m);
	}

	//! Set whether a disk mask must be drawn over the viewport
	void setViewportMaskDisk(void) {
		projection->setMaskType(Projector::DISK);
	}

	//! also set disk viewport values at same time
	void setViewportMaskDisk(int cx, int cy, int screenW, int screenH, int radius) {
		projection->setMaskType(Projector::DISK);
		projection->setDiskViewport(cx, cy, screenW, screenH, radius);
	}

	void setCenterHorizontalOffset(int h) {
		projection->setCenterHorizontalOffset(h);
	}
	int getCenterHorizontalOffset() {
		return projection->getCenterHorizontalOffset();
	}

	//! Get whether a disk mask must be drawn over the viewport
	bool getViewportMaskDisk(void) const {
		return projection->getMaskType()==Projector::DISK;
	}

	//! Set whether no mask must be drawn over the viewport
	void setViewportMaskNone(void) {
		projection->setMaskType(Projector::NONE);
	}

// for gravity ui
	Vec3d getViewportCenter(void) {
		return projection->getViewportCenter();
	}

	//! Set the projection type
	void setProjectionType(const string& ptype);
	//! Get the projection type
	string getProjectionType(void) const {
		return Projector::typeToString(projection->getType());
	}

	void setProjectorConfiguration( int configuration ) {
		projection->setProjectorConfiguration(configuration);
	}

	//! can the projector be configured?
	bool projectorConfigurationSupported(void) {
		return projection->projectorConfigurationSupported();
	}

	int getProjectorConfiguration() {
		return projection->getProjectorConfiguration();
	}


	//! What lens is being used?
	int getLens() {
		return projection->getLens();
	}

	void setProjectorShearHorz( double shear ) {
		projection->setShearHorz(shear);
	}

	double getProjectorShearHorz() {
		return projection->getShearHorz();
	}

	void setProjectorOffset( double x, double y ) {
		projection->setProjectionOffset(x, y);
	}

	double getProjectorOffsetX() {
		return projection->getOffsetX();
	}

	double getProjectorOffsetY() {
		return projection->getOffsetY();
	}


	//! get/set horizontal/vertical image flip
	bool getFlipHorz(void) const {
		return projection->getFlipHorz();
	}
	bool getFlipVert(void) const {
		return projection->getFlipVert();
	}
	void setFlipHorz(bool flip);
	void setFlipVert(bool flip);

	//! Set flag for enabling gravity labels
	void setFlagGravityLabels(bool b) {
		projection->setFlagGravityLabels(b);
	}
	//! Get flag for enabling gravity labels
	bool getFlagGravityLabels(void) const {
		return projection->getFlagGravityLabels();
	}

	//! Get display width
	int getDisplayWidth(void) const {
		return projection->getDisplayWidth();
	}

	//! Get display height
	int getDisplayHeight(void) const {
		return projection->getDisplayHeight();
	}


	//! Get viewport width
	int getViewportWidth(void) const {
		return projection->getViewportWidth();
	}

	//! Get viewport height
	int getViewportHeight(void) const {
		return projection->getViewportHeight();
	}

	//! Get viewport X position
	int getViewportPosX(void) const {
		return projection->getViewportPosX();
	}

	//! Get viewport Y position
	int getViewportPosY(void) const {
		return projection->getViewportPosY();
	}

	//! Set the viewport width and height
	void setViewportSize(int w, int h);

	//! Print the passed string so that it is oriented in the drection of the gravity
	void printGravity(s_font* font, float altitude, float azimuth, const string& str, bool outline = 1,
					  int justify = 0, bool cache = 0) const {
		font->print_gravity(projection, altitude, azimuth, str, justify, cache, outline, 0, 0);

	}

	///////////////////////////////////////////////////////////////////////////////////////
	// Landscape
	//! Set flag for displaying Landscape
	void setFlagLandscape(bool b) {
		landscape->setFlagShow(b);
	}
	//! Get flag for displaying Landscape
	bool getFlagLandscape(void) const {
		return landscape->getFlagShow();
	}

	//! Set flag for displaying Fog
	void setFlagFog(bool b) {
		landscape->setFlagShowFog(b);
	}
	//! Get flag for displaying Fog
	bool getFlagFog(void) const {
		return landscape->getFlagShowFog();
	}

	void setInitialLandscapeName() {
		setLandscape(mBackup.initial_landscapeName);
	}

	///////////////////////////////////////////////////////////////////////////////////////
	// Atmosphere
	//! Set flag for displaying Atmosphere
	void setFlagAtmosphere(bool b) {
		atmosphere->setFlagShow(b);
	}
	//! Get flag for displaying Atmosphere
	bool getFlagAtmosphere(void) const {
		return atmosphere->getFlagShow();
	}
	
	//! Get flag for vp Optoma
	bool getFlagVpOptoma(void) const {
		return atmosphere->getFlagOptoma();
	}
		

	//! Set atmosphere fade duration in s
	void setAtmosphereFadeDuration(float f) {
		atmosphere->setFadeDuration(f);
	}
	//! Get atmosphere fade duration in s
	float getAtmosphereFadeDuration(void) const {
		return atmosphere->getFadeDuration();
	}

	//! Set light pollution limiting magnitude (naked eye)
	void setLightPollutionLimitingMagnitude(float mag) {
		lightPollutionLimitingMagnitude = mag;
		//~ SettingsState state;
		//~ state.m_state.light_pollution_luminance = mag;
		//~ SharedData::Instance()->Settings( state );
		float ln = log(mag);
		float lum = 30.0842967491175 -19.9408790405749*ln +2.12969160094949*ln*ln - .2206;
		atmosphere->setLightPollutionLuminance(lum);
	}
	//! Get light pollution limiting magnitude
	float getLightPollutionLimitingMagnitude(void) const {
		return lightPollutionLimitingMagnitude;
	}

	///////////////////////////////////////////////////////////////////////////////////////
	// Milky Way
	//! Set flag for displaying Milky Way
	void setFlagMilkyWay(bool b) {
		milky_way->setFlagShow(b);
	}
	//! Get flag for displaying Milky Way
	bool getFlagMilkyWay(void) const {
		return milky_way->getFlagShow();
	}

	//! Set Milky Way intensity
	void setMilkyWayIntensity(float f) {
		milky_way->set_intensity(f);
	}
	//! Get Milky Way intensity
	float getMilkyWayIntensity(void) const {
		return milky_way->get_intensity();
	}

	//! Set Milky Way initial intensity
	void setMilkyWayInitialIntensity(float f) {
		milky_way->set_initial_intensity(f);
	}

	void setMilkyWayDuration(float f) {
		milky_way->setFaderDuration(f);
	}

	void restoreMilkyWayIntensity() {
		milky_way->restore_intensity();
	}

	//! Change Milkyway texture
	void milkyswap(string mdir) {
		milky_way->set_texture(mdir);
	}

	///////////////////////////////////////////////////////////////////////////////////////
	// Nebulae
	//! Set flag for displaying Nebulae
	void setFlagNebula(bool b) {
		nebulas->setFlagShow(b);
	}
	//! Get flag for displaying Nebulae
	bool getFlagNebula(void) const {
		return nebulas->getFlagShow();
	}

	//! Set flag for displaying Nebulae Hints
	void setFlagNebulaHints(bool b) {
		nebulas->setFlagHints(b);
	}
	//! Get flag for displaying Nebulae Hints
	bool getFlagNebulaHints(void) const {
		return nebulas->getFlagHints();
	}

	//! Set Nebulae Hints circle scale
	void setNebulaCircleScale(float f) {
		nebulas->setNebulaCircleScale(f);
	}
	//! Get Nebulae Hints circle scale
	float getNebulaCircleScale(void) const {
		return nebulas->getNebulaCircleScale();
	}

	//! Set flag for displaying Nebulae as bright
	void setFlagBrightNebulae(bool b) {
		nebulas->setFlagBright(b);
	}
	//! Get flag for displaying Nebulae as brigth
	bool getFlagBrightNebulae(void) const {
		return nebulas->getFlagBright();
	}

	//! Set maximum magnitude at which nebulae hints are displayed
	void setNebulaMaxMagHints(float f) {
		nebulas->setMaxMagHints(f);
	}
	//! Get maximum magnitude at which nebulae hints are displayed
	float getNebulaMaxMagHints(void) const {
		return nebulas->getMaxMagHints();
	}

	//! hide a particular DSO
	void DSOselectName(bool hide, string DSOName) const {
		return nebulas->selectName(hide, DSOName);
	}

	//! hide all DSO
	void DSOhideAll() const {
		cout << "all hide" << endl;
		nebulas->hideAll();
	}
	
	//! show (unhide) all DSO
	void DSOshowAll() const {
		cout << "all show" << endl;
		nebulas->showAll();
	}

	//! select all DSO in constellationName to be hidden or showed
	void DSOselectConstellation(bool hide, string constellationName) const {
		cout << constellationName<< " ";
		if (hide)
			cout << "hide" << endl;
		else
			cout << "show" << endl;
		nebulas->selectConstellation(hide, constellationName);
	}

	//! select all DSO with typeName to be hidden or showed
	void DSOselectType(bool hide, string typeName) const {
		cout << typeName<< " ";
		if (hide)
			cout << "hide" << endl;
		else
			cout << "show" << endl;
		nebulas->selectType(hide, typeName);
	}

	//~ //! Set flag for displaying Nebulae even without textures
	//~ void setFlagNebulaDisplayNoTexture(bool b) {
		//~ nebulas->setFlagDisplayNoTexture(b);
	//~ }
	//~ //! Get flag for displaying Nebulae without textures
	//~ bool getFlagNebulaDisplayNoTexture(void) const {
		//~ return nebulas->getFlagDisplayNoTexture();
	//~ }

	//! return the color for the DSO object
	Vec3f getColorNebulaLabels(void) const {
		return nebulas->getLabelColor();
	}
	
	//! return the color of the DSO circle 
	Vec3f getColorNebulaCircle(void) const {
		return nebulas->getCircleColor();
	}

	void setNebulaPictoSize(int value) const {
		nebulas->setPictoSize(value);
	}

	///////////////////////////////////////////////////////////////////////////////////////
	// Observer
	
	//! Return the current observatory (as a const object)
	// made non const so can track when save data!  Hmmm. 20070215
	// TODO resolve issue	
	Observer* getObservatory(void) {
		return observatory;
	}

	//! Move to a new latitude and longitude on home planet
	void moveObserver(double lat, double lon, double alt, int delay, const string& name) {
		observatory->move_to(lat, lon, alt, delay, name);
	}
	
	//! Move to relative latitude where home planet is fixed.
	void moveRelLatObserver(double lat, int delay)
	{
	  double latimem=observatory->get_latitude()+lat;
	  if (latimem>90) latimem=90;
	  if (latimem<-90) latimem=-90;
	  moveObserver(latimem,observatory->get_longitude(),observatory->get_altitude(),delay,observatory->get_name());
	}

	//! Move to relative longitude where home planet is fixed.
	void moveRelLonObserver(double lon, int delay)
	{
		moveObserver(observatory->get_latitude(),observatory->get_longitude()+lon,observatory->get_altitude(),delay,observatory->get_name());
	}

	//! change the Heading value
	void moveHeadingRelative(float f){
		navigation->set_heading(navigation->get_heading() + f);
	}

	//! Set Meteor Rate in number per hour
	void setMeteorsRate(int f) {
		meteors->set_ZHR(f);
	}
	
	//! flag value for videoprojector model optoma.
	void setCorOptoma(bool b) {
		atmosphere->setFlagOptoma(b);
	}
	
	//! Get Meteor Rate in number per hour
	int getMeteorsRate(void) const {
		return meteors->get_ZHR();
	}


	///////////////////////////////////////////////////////////////////////////////////////
	//text

	bool text_add(std::string name, std::string text, int altitude, int azimuth, float r, float g, float b) {
		return text_usr->add(name, text,altitude, azimuth, r,g,b);
	}
	
	void text_del(std::string name) {
		text_usr->del(name);
	}

	void text_clear(){
		text_usr->clear();
	}
	
	void text_name_update(std::string name, std::string text) {
		text_usr->name_update(name, text);
	}

	void text_display(string name , bool displ) {
		text_usr->text_display(name, displ);
	}

	void text_fading_duration(int a) {
		text_usr->setFadingDuration(a);
	}

	///////////////////////////////////////////////////////////////////////////////////////
	// SunTrace

	//! Set flag for displaying SunTrace
	void setFlagSunTrace(bool b) const {
		suntrace->setFlagShow(b);
	}
	//! Get flag for displaying SunTrace
	bool getFlagSunTrace(void) const {
		return suntrace->getFlagShow();
	}
	
	void setSunTracePen(bool b) const {
		suntrace->setPen(b);
	}

	bool getSunTracePen() const {
		return suntrace->getPen();
	}

	void sunTraceClear () const {
		suntrace->clear();
	}

	void sunTraceHide(string value) const;

	///////////////////////////////////////////////////////////////////////////////////////
	// Others
	//! Load color scheme from the given ini file and section name
	void setColorScheme(const string& skinFile, const string& section);

	double getZoomSpeed() {
		return zoom_speed;
	}
	float getAutoMoveDuration() {
		return auto_move_duration;
	}

	void getFontForLocale(const string &_locale, string &_fontFile, float &_fontScale,
	                      string &_fixedFontFile, float &_fixedFontScale);

	string getLandscapeName(void) {
		return landscape->getName();
	}
	string getLandscapeAuthorName(void) {
		return landscape->getAuthorName();
	}
	string getLandscapeDescription(void) {
		return landscape->getDescription();
	}

	//! Get the current navigation (manages frame transformation) used in the core
	Navigator* getNavigation() {
		return navigation;
	}

	bool observerAboveHomePlanet() {
		return aboveHomePlanet;
	}

	void setFlagNav(bool a) {
		//~ if (a==true)
			//~ printf("nav on\n");
		//~ else
			//~ printf("nav off\n");
		flagNav=a;
		cardinals_points->setInternalNav(a);
		equ_grid->setInternalNav(a);
		//~ equator_nav_line->setInternalNav(a);
		equator_line->setInternalNav(a);
		galactic_line->setInternalNav(a);
		meridian_line->setInternalNav(a);
		ecliptic_line->setInternalNav(a);
	}
	
	bool getFlagNav() {
		return flagNav;
	}

	//! Get the shared instance of GeodesicGrid
	GeodesicGrid* getGeodesicGrid() {return geodesic_grid;}
	//! Get the shared instance of GeodesicGrid
	const GeodesicGrid* getGeodesicGrid() const {return geodesic_grid;}

	mCity_Mgr	*mCity;					//!for using this class MUST BE PRIVATE	

//work with the sun
	double dateSunSet();				//! return the JD time when the sun set up
	double dateSunRise();				//! return the JD time when the sun go down
	double dateSunMeridian();				//! return the JD time when the sun cross the meridian
	
	void getmBackup();					//! get the current variables to struct Backup
	void setmBackup();					//! set a previous Backup  directly in use

private:

	//! display splash screen at the begining
	void init_splash(string splash_text, float font_size, const string& font_name, int ViewportWidth, int Viewportheight);
	
	//! Callback to record actions
	mBoost::callback<void, string> recordActionCallback;

	//! Select passed object
	//! @return true if the object was selected (false if the same was already selected)
	bool selectObject(const Object &obj);

	//! Find any kind of object by the name
	Object searchByNameI18n(const string &name) const;

	//! Find in a "clever" way an object from its equatorial position
	Object clever_find(const Vec3d& pos) const;

	//! Find in a "clever" way an object from its screen position
	Object clever_find(int x, int y) const;

	string FontFileNameGeneral;				//! The font file used by default during initialization
	string FontFileNamePlanet;				//! The font for the planet system
	string FontFileNameConstellation;		//! The font for all asterims
	string FontFileNameMenu;
	string FontFileNameText;
	double FontSizeText;
	double FontSizeGeneral;
	double FontSizePlanet;
	double FontSizeConstellation;
	double FontSizeCardinalPoints;
	string dataRoot;					// The root directory where the data is
	string localeDir;					// The directory containing the translation .mo file
	string skyCultureDir;				// The directory containing data for the culture used for constellations, etc..
	Translator skyTranslator;			// The translator used for astronomical object naming

	// Main elements of the program
	Navigator * navigation;				// Manage all navigation parameters, coordinate transformations etc..
	Observer * observatory;			// Manage observer position
	Projector * projection;				// Manage the projection mode and matrix
	Object selected_object;			// The selected object
	class HipStarMgr * hip_stars;		// Manage the hipparcos stars
	ConstellationMgr * asterisms;		// Manage constellations (boundaries, names etc..)
	NebulaMgr * nebulas;				// Manage the nebulas
	IlluminateMgr * illuminates;		// Manage the illuminations
	Text_Mgr * text_usr;				// manage all user text in dome
	SolarSystem* ssystem;				// Manage the solar system
	Atmosphere * atmosphere;			// Atmosphere
	SkyGrid * equ_grid;					// Equatorial grid
	SkyGrid * ecliptic_grid;			// ecliptical grid
	SkyGrid * galactic_grid;			// ecliptical grid
	SkyGrid * azi_grid;					// Azimutal grid
	SkyLine * equator_line;				// Celestial Equator line
	//~ SkyLine * equator_nav_line;			// Celestial Equator line NAV version
	SkyLine * galactic_line;			// Celestial Equator line
	SkyLine * ecliptic_line;			// Ecliptic line
	SkyLine * precession_circle;		// Precession circle
	SkyLine * circumpolar_circle;		// Circumpolar circle
	SkyLine * tropic_line;				// Tropic line
	SkyLine * meridian_line;			// Meridian line
	SkyLine * zenith_line;				// zenit
	SkyLine * polar_circle;				// polar circle
	SkyLine * ecliptic_center;			// ecliptic_center
	SkyLine * galactic_pole;			// galactic_pole
	SkyLine * galactic_center;			// galactic_center
	SkyLine * vernal_points;			// vernal_points
	SkyLine * analemma_line;			// Analemma line
	SkyLine * analemma;					// Analemma
	SkyLine * aries_line;				// Meridian line
	SkyLine * zodiac;					// Zodiac zone
	SkyPerson * personal;				// Personal drawing
	SkyPerson * personeq;				// Personal equatorial drawing
	SkyLine * greenwich_line;			// Greenwich line
	SkyLine * vertical_line;			// Meridian line
	Cardinals * cardinals_points;		// Cardinals points
	MilkyWay * milky_way;				// Our galaxy
	MeteorMgr * meteors;				// Manage meteor showers
	Landscape * landscape;				// The landscape ie the fog, the ground and "decor"
	ToneReproductor * tone_converter;	// Tones conversion between simulation world and display device
	SkyLocalizer *skyloc;				// for sky cultures and locales
	class TelescopeMgr *telescope_mgr;
	SunTrace * suntrace;				// the pen suntrace 
	model3D_mgr *model3DObj;			//for commun 3D obj
	
	float sky_brightness;				// Current sky Brightness in ?
	bool object_pointer_visibility;		// Should selected object pointer be drawn
	void drawChartBackground(void);
	string get_cursor_pos(int x, int y);

	// Increment/decrement smoothly the vision field and position
	void updateMove(int delta_time);
	int FlagEnableZoomKeys;
	int FlagEnableMoveKeys;

	double deltaAlt, deltaAz, deltaFov, deltaHeight;	// View movement
	double move_speed, zoom_speed;		// Speed of movement and zooming
	bool firstTime;                     // For init to track if reload or first time setup
	float InitFov;						// Default viewing FOV
	bool flagNav; 						// define the NAV version eddition

	Vec3d InitViewPos;					// Default viewing direction
	int FlagManualZoom;					// Define whether auto zoom can go further
	Vec3f chartColor;					// ?
	float auto_move_duration;			// Duration of movement for the auto move to a selected objectin seconds
	float flight_duration;			    // Duration of flight between bodies in seconds


//	float constellationFontSize;        // size for constellation labels
	bool aboveHomePlanet;               // true if looking down at home planet (high altitude)

	bool FlagAntialiasLines;            // whether to antialias all line drawing
	float m_lineWidth;                  // width to use when drawing any line

	bool FlagShaders;                   // whether to render using shaders

	//! Manage geodesic grid
	GeodesicGrid* geodesic_grid;

	//! size of Illuminate star
	double illuminate_size;

	//! Backup Manage
	backupWorkspace mBackup;			// variable used to remember various indicators in use

	void inimBackup();					// init at NULL all var
	SDL_Window *window;					//for swap GL buffer
	float lightPollutionLimitingMagnitude;  // Defined naked eye limiting magnitude (due to light pollution)
};

#endif // _CORE_H_
