/*
 * Stellarium360 astronomy simulation and visualization
 *
 * Copyright (C) 2002 Fabien Chereau
 * Copyright (C) 2009 Digitalis Education Solutions, Inc.
 * Copyright (C) 2013 of the LSS team
 * Copyright (C) 2014 of the LSS Team & Association Sirius
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Stellarium360 is a free open project of the LSS team
 * See the TRADEMARKS file for free open project usage requirements.
 *
 */

#include "milkyway.hpp"
#include "s_texture.hpp"
#include "utility.hpp"
#include <string>

using namespace std;


	//! Class which manages the displaying of the Milky Way
	MilkyWay::MilkyWay(float _radius) : radius(_radius), color(1.f, 1.f, 1.f)
	{
		tex = NULL;
		default_tex = NULL;
		fader.set_duration(4000);
	}

	MilkyWay::~MilkyWay()
	{
		if(default_tex == tex) default_tex = NULL;
		else delete default_tex;

		if (tex) delete tex;
	}

	//! If tex_file is empty or "default" revert to default texture
	//! If make_default is true, this new texture will also replace default texture
	void MilkyWay::set_texture(const string& tex_file, bool blend, bool make_default)
	{
		if (tex && tex != default_tex) delete tex;

		if(default_tex && (tex_file == "" || tex_file == "default")) {

			tex = default_tex;
		} else {

			tex = new s_texture(1,tex_file, !blend ? TEX_LOAD_TYPE_PNG_SOLID_REPEAT : TEX_LOAD_TYPE_PNG_BLEND3);

			if(make_default || default_tex == NULL) default_tex = tex;
		}
	}

	void MilkyWay::draw(ToneReproductor * eye, const Projector* prj, const Navigator* nav) const
	{
		assert(tex);	// A texture must be loaded before calling this

		// .045 chosen so that ad_lum = 1 at standard NELM of 6.5
		float ad_lum=eye->adapt_luminance(.045);

		// NB The tone reproducer code is simply incorrect, so this function 
		// fades out the Milky Way by the time eye limiting mag gets to ~5
		if(ad_lum < .9987 ) 
			ad_lum = -ad_lum*3.6168 +ad_lum*ad_lum*9.6253 -ad_lum*ad_lum*ad_lum*5.0121;

		if(ad_lum < 0) ad_lum = 0;

		// Special case, default texture is drawn dimmer by default so can
		// increase brightness through milky way intensity setting.
		// With user supplied images the expectation is full intensity.
		if(tex == default_tex) ad_lum /= 2.5;

		float cmag = ad_lum * intensityActual * fader.getInterstate();
		glColor4f(cmag, cmag, cmag, 1);


		glEnable(GL_CULL_FACE);
		glEnable(GL_TEXTURE_2D);
		glDisable(GL_BLEND);
		glBindTexture(GL_TEXTURE_2D, tex->getID());

		int divisions = 40;

		// Most textures are in Alpha/Delta and not galactic coordinates
		// To fit perfectly, it must be a fine sphere...
		prj->sSphere(radius,1.0,divisions,divisions,
				nav->get_j2000_to_eye_mat()*
				Mat4d::xrotation(M_PI)*
				Mat4d::yrotation(M_PI)*
				Mat4d::zrotation(M_PI/180*270), 1);

		glDisable(GL_CULL_FACE);
	}
