/*
 * Stellarium360 astronomy simulation and visualization
 *
 * Copyright (C) 2002 Fabien Chereau
 * Copyright (C) 2009 Digitalis Education Solutions, Inc.
 * Copyright (C) 2013 of the LSS team
 * Copyright (C) 2014 of the LSS Team & Association Sirius
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Stellarium360 is a free open project of the LSS team
 * See the TRADEMARKS file for free open project usage requirements.
 *
 */

#include "entity.hpp"
#include "skygrid.hpp"
#include "s_texture.hpp"
#include "utility.hpp"
#include <string>
using namespace std;

//! rms added color as parameter
SkyGrid::SkyGrid(SKY_GRID_TYPE grid_type, unsigned int _nb_meridian, unsigned int _nb_parallel, double _radius,
        unsigned int _nb_alt_segment, unsigned int _nb_azi_segment) :
    nb_meridian(_nb_meridian), nb_parallel(_nb_parallel), 	radius(_radius),
    nb_alt_segment(_nb_alt_segment), nb_azi_segment(_nb_azi_segment), color(0.2,0.2,0.2)
{
    transparent_top = true;
    gtype = grid_type;
    switch (grid_type) {
        case ALTAZIMUTAL :
            proj_func = &Projector::project_local;
            break;
        case EQUATORIAL :
            proj_func = &Projector::project_earth_equ;
            break;
        case ECLIPTIC:
            proj_func = &Projector::project_earth_ecliptic;
            break;
        case GALACTIC:
            proj_func = &Projector::project_earth_galactic;
            break;
        default :
            proj_func = &Projector::project_earth_equ;
    }

    // Alt points are the points to draw along the meridian
    alt_points = new Vec3f*[nb_meridian];
    for (unsigned int nm=0; nm<nb_meridian; ++nm) {
        alt_points[nm] = new Vec3f[nb_alt_segment+1];
        for (unsigned int i=0; i<nb_alt_segment+1; ++i) {
            sphe_to_rect((float)nm/(nb_meridian)*2.f*M_PI,(float)i/nb_alt_segment*M_PI-M_PI_2, alt_points[nm][i]);
            alt_points[nm][i] *= radius;
        }
    }

    // Alt points are the points to draw along the meridian
    azi_points = new Vec3f*[nb_parallel];
    for (unsigned int np=0; np<nb_parallel; ++np) {
        azi_points[np] = new Vec3f[nb_azi_segment+1];
        for (unsigned int i=0; i<nb_azi_segment+1; ++i) {
            sphe_to_rect((float)i/(nb_azi_segment)*2.f*M_PI,
                    (float)(np+1)/(nb_parallel+1)*M_PI-M_PI_2, azi_points[np][i]);
            azi_points[np][i] *= radius;
        }
    }
}

SkyGrid::~SkyGrid()
{
    for (unsigned int nm=0; nm<nb_meridian; ++nm) {
        delete [] alt_points[nm];
    }
    delete [] alt_points;

    for (unsigned int np=0; np<nb_parallel; ++np) {
        delete [] azi_points[np];
    }
    delete [] azi_points;

    if (font) delete font;
    font = NULL;

}

void SkyGrid::set_font(float font_size, const string& font_name)
{
    font = new s_font(font_size, font_name);
    assert(font);
}

void SkyGrid::draw(const Projector* prj) const
{
    if (!fader.getInterstate()) return;

    glDisable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); // Normal transparency mode

    Vec3d pt1;
    Vec3d pt2;

    prj->set_orthographic_projection();	// set 2D coordinate

    // Draw meridians
    for (unsigned int nm=0; nm<nb_meridian; ++nm) {
        if (transparent_top) {	// Transparency for the first and last points
            if ((prj->*proj_func)(alt_points[nm][0], pt1) && (prj->*proj_func)(alt_points[nm][1], pt2) ) {
                glColor4f(color[0],color[1],color[2],0.f);

                glBegin (GL_LINES);
                glVertex2f(pt1[0],pt1[1]);
                glColor4f(color[0],color[1],color[2],fader.getInterstate());
                glVertex2f(pt2[0],pt2[1]);
                glEnd();
            }

            glColor4f(color[0],color[1],color[2],fader.getInterstate());

            for (unsigned int i=1; i<nb_alt_segment-1; ++i) {
                if ((prj->*proj_func)(alt_points[nm][i], pt1) && (prj->*proj_func)(alt_points[nm][i+1], pt2) ) {
                    glBegin(GL_LINES);
                    glVertex2f(pt1[0],pt1[1]);
                    glVertex2f(pt2[0],pt2[1]);
                    glEnd();

                    static char str[255];	// TODO use c++ string

                    glEnable(GL_TEXTURE_2D);

                    double angle;

                    const double dx = pt1[0]-pt2[0];
                    const double dy = pt1[1]-pt2[1];
                    const double dq = dx*dx+dy*dy;

                    // TODO: allow for other numbers of meridians and parallels without
                    // screwing up labels?
                    //if( gtype == EQUATORIAL && i == 8 ) {
                    if ( i == 8 ) {
                        // draw labels along equator for RA
                        const double d = sqrt(dq);

                        angle = acos((pt1[1]-pt2[1])/d);
                        if ( pt1[0] < pt2[0] ) {
                            angle *= -1;
                        }

                        if ( gtype == EQUATORIAL ) {
							if (internalNav)
								sprintf( str, "%d°E", (24-nm)*15);
							else
								sprintf( str, "%dh", nm);
                        } else {
                            sprintf( str, "%d", nm<=12 ? (12-nm)*15 : (36-nm)*15 );
                        }

                        prj->set_orthographic_projection();

                        glTranslatef(pt2[0],pt2[1],0);
                        glRotatef(90+angle*180./M_PI,0,0,-1);

                        if ( gtype == EQUATORIAL ) {
                            font->print(2,-2,str,1,1);
                        } else {
                            font->print(6,-2,str,1,1);
                        }

                        prj->reset_perspective_projection();


                    } else if (nm % 8 == 0 && i != 16) {

                        const double d = sqrt(dq);

                        angle = acos((pt1[1]-pt2[1])/d);
                        if ( pt1[0] < pt2[0] ) {
                            angle *= -1;
                        }

                        sprintf( str, "%d", (i-8)*10);

                        if ( gtype == ALTAZIMUTAL || (gtype == EQUATORIAL && i > 8)) {
                            angle += M_PI;
                        }

                        prj->set_orthographic_projection();

                        glTranslatef(pt2[0],pt2[1],0);
                        glRotatef(angle*180./M_PI,0,0,-1);
                        font->print(2,-2,str,1,1);
                        prj->reset_perspective_projection();

                    }
                    glDisable(GL_TEXTURE_2D);
                }


                }

                if ((prj->*proj_func)(alt_points[nm][nb_alt_segment-1], pt1) && (prj->*proj_func)(alt_points[nm][nb_alt_segment], pt2) ) {
                    glColor4f(color[0],color[1],color[2],fader.getInterstate());
                    glBegin (GL_LINES);
                    glVertex2f(pt1[0],pt1[1]);
                    glColor4f(color[0],color[1],color[2],0.f);
                    glVertex2f(pt2[0],pt2[1]);
                    glEnd();
                }

            } else {
                glColor4f(color[0],color[1],color[2],fader.getInterstate());
                for (unsigned int i=0; i<nb_alt_segment; ++i) {
                    if ((prj->*proj_func)(alt_points[nm][i], pt1) && (prj->*proj_func)(alt_points[nm][i+1], pt2) ) {
                        glBegin (GL_LINES);
                        glVertex2f(pt1[0],pt1[1]);
                        glVertex2f(pt2[0],pt2[1]);
                        glEnd();
                    }
                }
            }
        }

        // Draw parallels
        glColor4f(color[0],color[1],color[2],fader.getInterstate());
        for (unsigned int np=0; np<nb_parallel; ++np) {
            for (unsigned int i=0; i<nb_azi_segment; ++i) {
                if ((prj->*proj_func)(azi_points[np][i], pt1) && (prj->*proj_func)(azi_points[np][i+1], pt2) ) {
                    glBegin (GL_LINES);
                    glVertex2f(pt1[0],pt1[1]);
                    glVertex2f(pt2[0],pt2[1]);
                    glEnd();
                }
            }
		if (internalNav) 
			np++;
        }

        prj->reset_perspective_projection();
    }

