/*
 * Stellarium360 astronomy simulation and visualization
 *
 * Copyright (C) 2002 Fabien Chereau
 * Copyright (C) 2009-2010 Digitalis Education Solutions, Inc.
 * Copyright (C) 2014 of the LSS Team & Association Sirius
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Stellarium360 is a free open project of of LSS team
 * See the TRADEMARKS file for free open project usage requirements.
 *
 */

#include <iostream>
#include <iomanip>

#include "night_shader.hpp"
#include "bump_shader.hpp"
#include "ring_shader.hpp"
#include "ringed_shader.hpp"
#include "entity.hpp"
#include "navigator.hpp"
#include "projector.hpp"
#include "s_font.hpp"
//#include "s_gui.hpp"
#include "stellastro.h" // just for get_apparent_sidereal_time
#include "log.hpp"

s_font* Entity::planet_name_font = NULL;
float Entity::object_scale = 1.f;
float Entity::object_size_limit = 9;
LinearFader Entity::flagShow;
LinearFader Entity::flagClouds;
s_texture *Entity::defaultTexMap = NULL;
s_texture *Entity::tex_eclipse_map = NULL;

void glCircle(const Vec3d& pos, float radius, float line_width = 1)
{
	float angle, facets;
	bool lastState = glIsEnabled(GL_TEXTURE_2D);

	glDisable(GL_TEXTURE_2D);
	glLineWidth(line_width);
	glBegin(GL_LINE_LOOP);

	if (radius < 2) facets = 6;
	else facets = (int)(radius*3);

	for (int i = 0; i < facets; i++) {
		angle = 2.0f*M_PI*i/facets;
		glVertex3f(pos[0] + radius * sin(angle), pos[1] + radius * cos(angle), 0.0f);
	}
	glEnd();

	if (lastState) glEnable(GL_TEXTURE_2D);
	glLineWidth(1.0f);
}


Entity::Entity(Entity *parent,
               const string& englishName,
               ENTITY_TYPE _typePlanet,
               int flagHalo,
               int flag_lighting,
               double radius,
               double oblateness,
               Vec3f color,
               Vec3f label_color,
               Vec3f orbit_color,
               Vec3f trail_color,
               float _sol_local_day,
               float albedo,
               const string& tex_map_path,
               const string& tex_map_name,
               const string& tex_halo_name,
               Orbit *orbit,
               bool close_orbit,
               const string& model_name,
               model3D_mgr * _myObjMgr,
               bool _hidden,
               bool _deleteable,
               double orbit_bounding_radius,
			   const string& tex_norm_name,
			   const string& tex_night_name,
			   const string& tex_specular_name,
			   const string& tex_cloud_name,
			   const string& tex_norm_cloud_name) :

		englishName(englishName), flagHalo(flagHalo),
		flag_lighting(flag_lighting),
		radius(radius), one_minus_oblateness(1.0-oblateness),
		interpolating(false), interpolationDirection(0),
		color(color), label_color(label_color), orbit_color(orbit_color), trail_color(trail_color), albedo(albedo), axis_rotation(0.),
		tex_map(NULL), tex_halo(NULL), tex_big_halo(NULL), 
		tex_norm(NULL), tex_night(NULL), tex_specular(NULL), tex_cloud(NULL),
		tex_shadow_cloud(NULL), tex_norm_cloud(NULL), eye_sun(0.0f, 0.0f, 0.0f),
		bumpEnable(tex_norm_name != ""), nightEnable(tex_night_name != ""), 
		ringedEnable(0), rings(NULL), sphere_scale(1.f),
		lastJD(J2000), last_orbitJD(0), deltaJD(JD_SECOND), orbit_cached(0),
		orbit(orbit), 
		parent(parent), close_orbit(close_orbit), hidden(_hidden), deleteable(_deleteable),
		is_satellite(0), orbit_bounding_radius(orbit_bounding_radius),
		boundingRadius(-1), ring_shadow(false), sun_half_angle(0.0)
{
	hiddenInitial = hidden;
	sol_local_day = _sol_local_day;
	typePlanet = _typePlanet;
	if (parent) {
		parent->satellites.push_back(this);
		if (parent->getEnglishName() != "Sun") is_satellite = 1; // quicker lookup
	}
	ecliptic_pos=Vec3d(0.,0.,0.);
	rot_local_to_parent = Mat4d::identity();
	rot_local_to_parent_unprecessed = Mat4d::identity();
	mat_local_to_parent = Mat4d::identity();

	if(tex_map_name == "") {
		// NOTE, default body texture initialization would ideally be done elsewhere
		if(defaultTexMap == NULL) 
			defaultTexMap = new s_texture("bodies/nomap.png", TEX_LOAD_TYPE_PNG_SOLID_REPEAT, 1);
		tex_map = defaultTexMap;
	} else {
		// planets are mipmapped	
		if ( tex_map_path == "" ) {
			tex_map = new s_texture(tex_map_name, TEX_LOAD_TYPE_PNG_SOLID_REPEAT, 1);
			if (flagHalo) tex_halo = new s_texture(tex_halo_name);
		} else {
			// TODO: CAN NOT TELL if textures are loaded or not!!!
			tex_map = new s_texture(1, tex_map_path + tex_map_name, TEX_LOAD_TYPE_PNG_SOLID_REPEAT, 1);
			if (flagHalo) tex_halo = new s_texture(1, tex_map_path + tex_halo_name, TEX_LOAD_TYPE_PNG_BLEND1);
		}
	}

	if(flagHalo) {
		if ( tex_map_path == "" ) tex_halo = new s_texture(tex_halo_name);
		else tex_halo = new s_texture(1, tex_map_path + tex_halo_name, TEX_LOAD_TYPE_PNG_BLEND1);
	}

//	if(flagShaders) {
		if (nightEnable) {
			if ( tex_map_path == "" ) tex_night = new s_texture(tex_night_name, TEX_LOAD_TYPE_PNG_SOLID_REPEAT, 1);
			else tex_night = new s_texture(1, tex_map_path + tex_night_name, TEX_LOAD_TYPE_PNG_SOLID_REPEAT, 1);
			if(tex_map_path=="") tex_specular = new s_texture(tex_specular_name, TEX_LOAD_TYPE_PNG_SOLID_REPEAT);
			else tex_specular = new s_texture(1, tex_map_path + tex_specular_name, TEX_LOAD_TYPE_PNG_SOLID_REPEAT);
			if(!NightShader::instance()->isSupported()) {
				Log.write("Unable to use night shader for " + englishName, cLog::eLOG_TYPE_ERROR);
				nightEnable = false;
			} else programObject.attachShader(NightShader::instance());
			bumpEnable = false;
		} 
		else 
		if (bumpEnable) {
			// Currently night and bump shader are exclusive
			if ( tex_map_path == "" )
				tex_norm = new s_texture(tex_norm_name, TEX_LOAD_TYPE_PNG_SOLID_REPEAT);
			else tex_norm = new s_texture(1, tex_map_path + tex_norm_name, TEX_LOAD_TYPE_PNG_SOLID_REPEAT);
			if(!tex_norm->getID() || !BumpShader::instance()->isSupported()) {
				Log.write("Unable to use bump shader for " + englishName , cLog::eLOG_TYPE_ERROR);
				bumpEnable = false;
			} else programObject.attachShader(BumpShader::instance());
		} 
		else 
		if(englishName != "Sun") {
			// Default Entity shader
			ringedEnable = true;
			if(!RingedShader::instance()->isSupported()) {
				Log.write( "Unable to use ringed shader for " + englishName, cLog::eLOG_TYPE_ERROR);
				ringedEnable = false;
			} else programObject.attachShader(RingedShader::instance());
			nightEnable = bumpEnable = false;
		} 
		else {
			ringedEnable = nightEnable = bumpEnable = false;
		}
//	}

	if(tex_cloud_name != "") {
		// Try to use cloud texture in any event, even if can not use shader
		if(tex_map_path == "") tex_cloud = new s_texture(tex_cloud_name, TEX_LOAD_TYPE_PNG_ALPHA, true);
		else tex_cloud = new s_texture(1, tex_map_path + tex_cloud_name, TEX_LOAD_TYPE_PNG_ALPHA, true);

		if(tex_norm_cloud_name=="") {
			Log.write("No cloud normal texture defined for " + englishName, cLog::eLOG_TYPE_ERROR);
		} else {
			if( tex_map_path == "") tex_norm_cloud = new s_texture(tex_norm_cloud_name, TEX_LOAD_TYPE_PNG_SOLID_REPEAT, true);
			else tex_norm_cloud = new s_texture(tex_norm_cloud_name, TEX_LOAD_TYPE_PNG_SOLID_REPEAT, true);
		}
	}

	// Body trails
	DeltaTrail = 1;
	MaxTrail = 1460;
	last_trailJD = 0; // for now
	trail_on = 0;
	first_point = 1;

	nameI18 = englishName;

	visibilityFader = true;
	drawaxis=false;
	
	useModel3D= true;
	myObjMgr = _myObjMgr;
	obj= new obj3D();

	if (myObjMgr == NULL) //all artificial planet
		useModel3D = false;
	else
		if (model_name =="") {
			obj= myObjMgr->addStellObj("Sphere");
			useModel3D = false; }
		else
			obj= myObjMgr->addStellObj(model_name);

	if (obj==NULL)
		useModel3D = false;
		
	//~ if (useModel3D)
		//~ cout << englishName << " " << "oki" << endl;
}

Entity::~Entity()
{
	if (tex_map && tex_map != defaultTexMap) delete tex_map;
	tex_map = NULL;
	if (tex_halo) delete tex_halo;
	tex_halo = NULL;
	if (rings) delete rings;
	rings = NULL;
	if (tex_big_halo) delete tex_big_halo;
	tex_big_halo = NULL;
	if (tex_norm) delete tex_norm;
	tex_norm = NULL;
	if (tex_night) delete tex_night;
	tex_night = NULL;
	if (tex_specular) delete tex_specular;
	tex_specular = NULL;
	if (tex_cloud) delete tex_cloud;
	tex_cloud = NULL;
	if (tex_norm_cloud) delete tex_norm_cloud;
	tex_norm_cloud = NULL;
	if(orbit) delete orbit;
	orbit = NULL;
}

// Return the information string "ready to print" :)
string Entity::getInfoString(const Navigator * nav) const
{
	double tempDE, tempRA;
	ostringstream oss;

	oss << _(englishName);  // UI translation can differ from sky translation
	oss.setf(ios::fixed);
	oss.precision(1);
	oss << endl;

	oss.precision(2);
	oss << _("Magnitude: ") << compute_magnitude(nav->get_observer_helio_pos()) << endl;

	Vec3d equPos = get_earth_equ_pos(nav);
	rect_to_sphe(&tempRA,&tempDE,equPos);

	oss << _("RA/DE: ") << Utility::printAngleHMS(tempRA) << " / " << Utility::printAngleDMS(tempDE) << endl;

	// calculate alt az position
	Vec3d localPos = nav->earth_equ_to_local(equPos);
	rect_to_sphe(&tempRA,&tempDE,localPos);
	tempRA = 3*M_PI - tempRA;  // N is zero, E is 90 degrees
	if (tempRA > M_PI*2) tempRA -= M_PI*2;

	oss << _("Alt/Az: ") << Utility::printAngleDMS(tempDE) << " / " << Utility::printAngleDMS(tempRA) << endl;

	oss.precision(8);
	oss << _("Distance: ") << equPos.length() << " " << _("AU") << endl;

	return oss.str();
}

//! Get sky label (sky translation)
string Entity::getSkyLabel(const Navigator * nav) const
{
	return nameI18;
}


// Return the information string "ready to print" :)
string Entity::getShortInfoString(const Navigator * nav) const
{
	ostringstream oss;
	oss << _(englishName);  // UI translation can differ from sky translation
	oss << ", " << getTypePlanet(typePlanet) << ":"; 
	oss.setf(ios::fixed);
	oss.precision(2);
	oss << "  " << _("Magnitude: ") << compute_magnitude(nav->get_observer_helio_pos());

	Vec3d equPos = get_earth_equ_pos(nav);
	oss.precision(4);
	oss << "  " <<  _("Distance: ") << equPos.length() << " " << _("AU");
	return oss.str();
}

void Entity::getAltAz(const Navigator * nav, double *alt, double *az) const
{
	double _alt, _az;
	// calculate alt az
	Vec3d equPos = get_earth_equ_pos(nav);
	Vec3d local_pos = nav->earth_equ_to_local(equPos);
	rect_to_sphe(&_az,&_alt,local_pos);
	_az = 3*M_PI - _az; 
	if (_az > M_PI*2) _az -= M_PI*2;
	*az=_az;
	*alt= _alt;
}

string Entity::getShortInfoNavString(const Navigator * nav) const
{
	ostringstream oss;	
	double tempDE, tempRA;
	Vec3d equPos = get_earth_equ_pos(nav);
	rect_to_sphe(&tempRA,&tempDE,equPos);
	oss << _("RA/DE: ") << Utility::printAngleHMS(tempRA) << "/" << Utility::printAngleDMS(tempDE);

    double jd=nav->get_Julian();
    double sidereal;
    double T;
    double Le;
    double HA;
    double GHA;
    double PA;
    
    T = (jd - 2451545.0) / 36525.0;
    Le = nav->get_longitude();    
    /* calc mean angle */
    sidereal = 280.46061837 + (360.98564736629 * (jd - 2451545.0)) + (0.000387933 * T * T) - (T * T * T / 38710000.0);
    HA=sidereal+Le-tempRA*180.0/M_PI;
    GHA=sidereal-tempRA*180.0/M_PI;
    while (HA>=360) HA-=360;
    while (HA<0)    HA+=360;
    if (HA<180) PA=HA; else PA=360-HA;
    while (GHA>=360) GHA-=360;
    while (GHA<0)    GHA+=360;
    while (tempRA>=2*M_PI) tempRA-=2*M_PI;
    while (tempRA<0)    tempRA+=2*M_PI;

	oss << _("SA ") << Utility::printAngleDMS(2*M_PI-tempRA) 
		<< " GHA " << Utility::printAngleDMS(GHA*M_PI/180.0)
		<< " LHA " << Utility::printAngleDMS(HA*M_PI/180.0); 

		// calculate alt az
	Vec3d local_pos = nav->earth_equ_to_local(equPos);
	rect_to_sphe(&tempRA,&tempDE,local_pos);
	tempRA = 3*M_PI - tempRA;  // N is zero, E is 90 degrees
	if (tempRA > M_PI*2) tempRA -= M_PI*2;
	oss << "@" << _(" Az/Alt/coA: ") << Utility::printAngleDMS(tempRA) << "/" << Utility::printAngleDMS(tempDE) << "/" << Utility::printAngleDMS(((M_PI/180.f)*90.)-tempDE) << " LPA " << Utility::printAngleDMS(PA*M_PI/180.0);
	return oss.str();
}

double Entity::get_close_fov(const Navigator* nav) const
{
	return atanf(radius*sphere_scale*2.f/get_earth_equ_pos(nav).length())*180./M_PI * 4;
}

double Entity::get_satellites_fov(const Navigator * nav) const
{

	if ( !satellites.empty() && englishName != "Sun") {
		double rad = getBoundingRadius();
		if ( rad > 0 ) return atanf(rad/get_earth_equ_pos(nav).length()) *180./M_PI * 4;
	}

	return -1.;

}

double Entity::get_parent_satellites_fov(const Navigator *nav) const
{
	if (parent && parent->parent) return parent->get_satellites_fov(nav);
	return -1.0;
}

// Set the orbital elements
void Entity::set_rotation_elements(float _period, float _offset, double _epoch, float _obliquity, 
								   float _ascendingNode, float _precessionRate, double _sidereal_period, float _axial_tilt)
{
	re.period = _period;
	re.offset = _offset;
	re.epoch = _epoch;
	re.obliquity = _obliquity;
	re.ascendingNode = _ascendingNode;
	re.precessionRate = _precessionRate;
	re.sidereal_period = _sidereal_period;  // used for drawing orbit lines
	re.axialTilt = _axial_tilt; // Used for drawing tropic lines

	delta_orbitJD = re.sidereal_period/ORBIT_SEGMENTS;
}


// Return the Entity position in rectangular earth equatorial coordinate
Vec3d Entity::get_earth_equ_pos(const Navigator * nav) const
{
	Vec3d v = get_heliocentric_ecliptic_pos();
	return nav->helio_to_earth_pos_equ(v);		// this is earth equatorial but centered
	// on observer's position (latitude, longitude)
	//return navigation.helio_to_earth_equ(&v); this is the real equatorial centered on earth center
}

Vec3d Entity::getObsJ2000Pos(const Navigator *nav) const
{
	return mat_vsop87_to_j2000.multiplyWithoutTranslation(
	           get_heliocentric_ecliptic_pos()
	           - nav->get_observer_helio_pos());
}


// Compute the position in the parent Entity coordinate system
// Actually call the provided function to compute the ecliptical position
void Entity::computePositionWithoutOrbits(const double date)
{
	if (fabs(lastJD-date)>deltaJD) {
		if(orbit) orbit->positionAtTimevInVSOP87Coordinates(date, ecliptic_pos);
		lastJD = date;
	}
}

void Entity::compute_position(const double date)
{

	if(!orbit) return;

	// Large performance advantage from avoiding object overhead
	OsculatingFunctionType *oscFunc = orbit->getOsculatingFunction();

	// for performance only update orbit points if visible
	if (orbit_fader.getInterstate()*visibilityFader.getInterstate()>0.000001 
		&& delta_orbitJD > 0 
		&& (fabs(last_orbitJD-date)>delta_orbitJD || !orbit_cached)) {

		// calculate orbit first (for line drawing)
		double date_increment = re.sidereal_period/ORBIT_SEGMENTS;
		double calc_date;
		//	  int delta_points = (int)(0.5 + (date - last_orbitJD)/date_increment);
		int delta_points;

		if ( date > last_orbitJD ) {
			delta_points = (int)(0.5 + (date - last_orbitJD)/date_increment);
		} else {
			delta_points = (int)(-0.5 + (date - last_orbitJD)/date_increment);
		}
		double new_date = last_orbitJD + delta_points*date_increment;

		//printf( "Updating orbit coordinates for %s (delta %f) (%d points)\n", name.c_str(), delta_orbitJD, delta_points);
		//cout << englishName << ": " << delta_points << "  " << orbit_cached << endl;

		if ( delta_points > 0 && delta_points < ORBIT_SEGMENTS && orbit_cached) {

			for ( int d=0; d<ORBIT_SEGMENTS; d++ ) {
				if (d + delta_points >= ORBIT_SEGMENTS ) {
					// calculate new points
					calc_date = new_date + (d-ORBIT_SEGMENTS/2)*date_increment;
					// date increments between points will not be completely constant though

					if(oscFunc) (*oscFunc)(date,calc_date,orbitPoint[d]);
					else orbit->fastPositionAtTimevInVSOP87Coordinates(date,calc_date,orbitPoint[d]);
				} else {
					orbitPoint[d] = orbitPoint[d+delta_points];
				}
			}

			last_orbitJD = new_date;

		} else if ( delta_points < 0 && abs(delta_points) < ORBIT_SEGMENTS  && orbit_cached) {

			for ( int d=ORBIT_SEGMENTS-1; d>=0; d-- ) {
				if (d + delta_points < 0 ) {
					// calculate new points
					calc_date = new_date + (d-ORBIT_SEGMENTS/2)*date_increment;

					if(oscFunc) (*oscFunc)(date,calc_date,orbitPoint[d]);
					else orbit->fastPositionAtTimevInVSOP87Coordinates(date,calc_date,orbitPoint[d]);
				} else {
					orbitPoint[d] = orbitPoint[d+delta_points];
				}
			}

			last_orbitJD = new_date;

		} else if ( delta_points || !orbit_cached) {

			// update all points (less efficient)
			for ( int d=0; d<ORBIT_SEGMENTS; d++ ) {
				calc_date = date + (d-ORBIT_SEGMENTS/2)*date_increment;

				if(oscFunc) (*oscFunc)(date,calc_date,orbitPoint[d]);
				else orbit->fastPositionAtTimevInVSOP87Coordinates(date,calc_date,orbitPoint[d]);
			}

			last_orbitJD = date;

			// \todo remove this for efficiency?  Can cause rendering issues near body though
			// If orbit is largely constant through time cache it
			if (orbit->isStable(date)) orbit_cached = 1;
		}

		// calculate actual Entity position
		if(oscFunc) (*oscFunc)(date,date,ecliptic_pos);
		else orbit->positionAtTimevInVSOP87Coordinates(date,date,ecliptic_pos);

		lastJD = date;

	} else {

		double delta = date-lastJD;
		delta = fabs(delta);

		if(delta >= deltaJD ) {
			if(oscFunc) (*oscFunc)(date,date,ecliptic_pos);
			else orbit->positionAtTimevInVSOP87Coordinates(date,date,ecliptic_pos);

			lastJD = date;
		}
	}
}

// Compute the transformation matrix from the local Entity coordinate to the parent Entity coordinate
void Entity::compute_trans_matrix(double jd)
{
	axis_rotation = getSiderealTime(jd);

	// Special case - heliocentric coordinates are on ecliptic,
	// not solar equator...
	if (parent) {
		rot_local_to_parent = Mat4d::zrotation(re.ascendingNode
		                                       -re.precessionRate*(jd-re.epoch))
		                      * Mat4d::xrotation(re.obliquity);
		rot_local_to_parent_unprecessed = Mat4d::zrotation(re.ascendingNode)
		                      * Mat4d::xrotation(re.obliquity);

	}

	mat_local_to_parent = Mat4d::translation(ecliptic_pos)
	                      * rot_local_to_parent;

}

Mat4d Entity::getRotEquatorialToVsop87(void) const
{
	Mat4d rval = rot_local_to_parent;
	if (parent) for (const Entity *p=parent; p->parent; p=p->parent) {
			rval = p->rot_local_to_parent * rval;
		}
	return rval;
}

void Entity::setRotEquatorialToVsop87(const Mat4d &m) {
  Mat4d a = Mat4d::identity();
  if (parent) for (const Entity *p=parent;p->parent;p=p->parent) {
    a = p->rot_local_to_parent * a;
  }
  rot_local_to_parent = a.transpose() * m;
}

string Entity::getTypePlanet(const ENTITY_TYPE str)  const {
	switch (str) {
	case 0 : return "sun";
	case 1 : return "planet";
	case 2 : return "moon";
	case 3 : return "dwarf planet";
	case 4 : return "asteroid";
	case 5 : return "plutoid";
	default : return "unknown";
	}
	return "";
}

// Compute the z rotation to use from equatorial to geographic coordinates
double Entity::getSiderealTime(double jd) const
{
	if (englishName=="Earth") return get_apparent_sidereal_time(jd);

	double t = jd - re.epoch;
	double rotations = t / (double) re.period;
	double wholeRotations = floor(rotations);
	double remainder = rotations - wholeRotations;

	return remainder * 360. + re.offset;
}

// Get the Entity position in the parent Entity ecliptic coordinate
Vec3d Entity::get_ecliptic_pos() const
{
	return ecliptic_pos;
}

// Return the heliocentric ecliptical position
// used only for earth shadow, lunar eclipse -- DIGITALIS: this statement is not true!
Vec3d Entity::get_heliocentric_ecliptic_pos() const
{
	Vec3d pos = ecliptic_pos;
	const Entity *p = parent;

	while (p && p->parent) {
		pos += p->ecliptic_pos;
		p = p->parent;
	}

	return pos;
}


void Entity::set_heliocentric_ecliptic_pos(const Vec3d &pos)
{
	ecliptic_pos = pos;
	const Entity *p = parent;
	if (p) while (p->parent)
	{
		ecliptic_pos -= p->ecliptic_pos;
		p = p->parent;
	}
}


// Compute the distance to the given position in heliocentric coordinate (in AU)
double Entity::compute_distance(const Vec3d& obs_helio_pos)
{
	distance = (obs_helio_pos-get_heliocentric_ecliptic_pos()).length();
	return distance;
}

// Get the phase angle for an observer at pos obs_pos in the heliocentric coordinate (dist in AU)
double Entity::get_phase(Vec3d obs_pos) const
{
	const double sq = obs_pos.lengthSquared();
	const Vec3d heliopos = get_heliocentric_ecliptic_pos();
	const double Rq = heliopos.lengthSquared();
	const double pq = (obs_pos - heliopos).lengthSquared();
	const double cos_chi = (pq + Rq - sq)/(2.0*sqrt(pq*Rq));
	return (1.0 - acos(cos_chi)/M_PI) * cos_chi
	       + sqrt(1.0 - cos_chi*cos_chi) / M_PI;
}

float Entity::compute_magnitude(Vec3d obs_pos) const
{
	float rval = 0;
	const double sq = obs_pos.lengthSquared();
	if (parent == 0) {
		// sun
		rval = -26.73f + 2.5f*log10f(sq);
	}
	else {
		const Vec3d heliopos = get_heliocentric_ecliptic_pos();
		const double Rq = heliopos.lengthSquared();
		const double pq = (obs_pos - heliopos).lengthSquared();
		const double cos_chi = (pq + Rq - sq)/(2.0*sqrt(pq*Rq));
		const double phase = (1.0 - acos(cos_chi)/M_PI) * cos_chi
							 + sqrt(1.0 - cos_chi*cos_chi) / M_PI;
		const float F = 2.0 * albedo * radius * radius * phase / (3.0*pq*Rq);
		rval = -26.73f - 2.5f * log10f(F);
	}
	return rval;
}

float Entity::compute_magnitude(const Navigator * nav) const
{
	return compute_magnitude(nav->get_observer_helio_pos());
}

void Entity::set_big_halo(const string& halotexfile, const string &path)
{
	tex_big_halo = new s_texture((path!=""), path + halotexfile, TEX_LOAD_TYPE_PNG_SOLID);
}

// Return the radius of a circle containing the object on screen
float Entity::get_on_screen_size(const Projector* prj, const Navigator * nav, bool orb_only)
{
	double rad;
	if (rings && !orb_only) rad = rings->get_size();
	else rad = radius;

	return atanf(rad*sphere_scale*2.f/get_earth_equ_pos(nav).length())*180./M_PI/prj->get_fov()*prj->getViewportHeight();
}

// Return the angle (degrees) of the Entity orb
float Entity::get_angular_size(const Projector* prj, const Navigator * nav)
{

	return atanf(radius*sphere_scale*2.f/get_earth_equ_pos(nav).length())*180./M_PI;
}


// Return the radius of a circle containing the object and its satellites on screen
float Entity::get_on_screen_bounding_size(const Projector* prj, const Navigator * nav)
{
	double rad = getBoundingRadius();

	return atanf(rad*sphere_scale*2.f/get_earth_equ_pos(nav).length())*180./M_PI/prj->get_fov()*prj->getViewportHeight();
}


// Draw the Entity and all the related infos : name, circle etc..
double Entity::draw(Projector* prj, const Navigator * nav, const ToneReproductor* eye, int flag_point, bool stencil, bool depthTest, bool drawHomePlanet,bool selected)
{
	if (hidden) return 0;

	eye_sun = nav->get_helio_to_eye_mat() * Vec3f(0.0f, 0.0f, 0.0f);

	Mat4d mat = mat_local_to_parent;
	Mat4d parent_mat = Mat4d::identity();

	// \todo account for moon orbit precession (independent of parent)
	// also does not allow for multiple levels of precession

	const Entity *p = parent;

	bool myParent = true;
	while (p && p->parent) {   //cette bocule ne sert que pour les lunes des planetes

		// Some orbits are already precessed, namely elp82
		if(myParent && !useParentPrecession(lastJD)) {
			mat = Mat4d::translation(p->ecliptic_pos)
				* mat
				* p->rot_local_to_parent_unprecessed; 
						//printf("unprecessed name:  %s\n", englishName.c_str());
		} else {
			mat = Mat4d::translation(p->ecliptic_pos)
				* mat
				* p->rot_local_to_parent; 
						//printf("precessed name:  %s\n", englishName.c_str());
		}
		
		parent_mat = Mat4d::translation(p->ecliptic_pos)
			* parent_mat;

		p = p->parent;

		myParent = false;
	}


	// This removed totally the Entity shaking bug!!!
	mat = nav->get_helio_to_eye_mat() * mat;
	parent_mat = nav->get_helio_to_eye_mat() * parent_mat;

	eye_planet = mat * Vec3f(0.0f, 0.0f, 0.0f);

	Vec3f lightDirection = eye_sun - eye_planet;
	sun_half_angle = atan(696000.0/AU/lightDirection.length());  // hard coded Sun radius!

	//	cout << sun_half_angle << " sun angle on " << englishName << endl;
	lightDirection.normalize();

	if (!drawHomePlanet && this == nav->getHomePlanet()) {
		if (rings) rings->draw(prj,mat,1000.0,lightDirection,eye_planet,radius);
		return 0;
	}
	
	// Compute the 2D position and check if in the screen
	float screen_sz = get_on_screen_size(prj, nav);

	// special case if big halo (Sun)
	float screen_size_with_halo = screen_sz;
	if (tex_big_halo && (big_halo_size > screen_sz))
		screen_size_with_halo = big_halo_size;

	// Check Entity visible in viewport
	bool isVisible = prj->project_custom_check(Vec3f(0,0,0), screenPos, mat, (int)(screen_size_with_halo/2));

	bool visible = isVisible;
	if(selected) visible = true;  // always draw orbit if selected, even if not on screen

	visible = true;

	// fade orbits as enter/leave view
	if (is_satellite && prj->get_fov()>30) {

		// If not in the parent Entity system
		// OR one of the other sister moons do not draw
		if (parent != nav->getHomePlanet() &&
			parent != nav->getHomePlanet()->get_parent() &&
			get_earth_equ_pos(nav).length() > 1) {  // If further than 1 AU of object if flying
			visibilityFader = false; // orbits will fade in and out now
		} else visibilityFader = visible;
	} else visibilityFader = visible;

	if ( is_satellite) 
		draw_orbit_3d(nav, prj, parent_mat);
	else {
		draw_orbit_2d(nav, prj, parent_mat);
		//			cout << "2d orbit: " << englishName << endl;
	}

	// Do not draw anything else if was not visible
	// Draw the name, and the circle if it's not too close from the body it's turning around
	// this prevents name overlaping (ie for jupiter satellites)
	float ang_dist = 300.f*atan(get_ecliptic_pos().length()/get_earth_equ_pos(nav).length())/prj->get_fov();
	if (ang_dist==0.f)
		ang_dist = 1.f; // if ang_dist == 0, the Entity is sun..

	draw_trail(nav, prj);


	if (ang_dist>0.25) {
		if (ang_dist>1.f) ang_dist = 1.f;
		draw_hints(nav, prj);
	}
	if (depthTest) glEnable(GL_DEPTH_TEST);

	if (screen_sz > 1 && isVisible) {  // huge improvement in performance

		if (rings) {
			glEnable(GL_DEPTH_TEST);

			if (drawaxis) 
				draw_axis(prj, mat, screen_sz);
				
			draw_sphere(prj, nav, mat, screen_sz);  // SL O O W
			rings->draw(prj,mat,screen_sz,lightDirection,eye_planet,radius);
		} else {
			if (stencil) glEnable(GL_STENCIL_TEST);

			if (drawaxis) 
				draw_axis(prj, mat, screen_sz);
				
			draw_sphere(prj, nav, mat, screen_sz);  // SL O O W
			if (stencil) glDisable(GL_STENCIL_TEST);
			}
	}

	if (tex_halo) {
		// Workaround for depth buffer precision and near planets
		if(get_earth_equ_pos(nav).length() > 0.0001) {
			glDisable(GL_DEPTH_TEST);
			if (flag_point) draw_point_halo(nav, prj, eye);
			else draw_halo(nav, prj, eye);
		} else {
			if (flag_point) draw_point_halo(nav, prj, eye, mat);
			else draw_halo(nav, prj, eye, mat);
		}
	}

	glDisable(GL_DEPTH_TEST);

	if (tex_big_halo) draw_big_halo(nav, prj, eye);	


	double distanceSquared =
	    (screenPos[0] - previousScreenPos[0]) *
	    (screenPos[0] - previousScreenPos[0]) +
	    (screenPos[1] - previousScreenPos[1]) *
	    (screenPos[1] - previousScreenPos[1]);
	previousScreenPos = screenPos;
	return distanceSquared;
}

void Entity::draw_hints(const Navigator* nav, const Projector* prj)
{

	//	printf("Out level %f\n", hint_fader.getInterstate());
	if (!hint_fader.getInterstate()) return;

	prj->set_orthographic_projection();    // 2D coordinate

	glEnable(GL_BLEND);
	glDisable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);

	// Draw nameI18 + scaling if it's not == 1.
	float tmp = 10.f + get_on_screen_size(prj, nav)/sphere_scale/2.f; // Shift for nameI18 printing

	glColor4f(label_color[0], label_color[1], label_color[2],hint_fader.getInterstate());

	prj->getFlagGravityLabels() ?
	prj->print_gravity180(planet_name_font, screenPos[0],screenPos[1], getSkyLabel(nav), 1, tmp, tmp) :
		planet_name_font->print(screenPos[0]+tmp,screenPos[1]+tmp, getSkyLabel(nav), 1);

	// hint disapears smoothly on close view
	tmp -= 10.f;
	if (tmp<1) tmp=1;
	glColor4f(label_color[0], label_color[1], label_color[2],hint_fader.getInterstate()/tmp);

	// Draw the 2D small circle
	glCircle(screenPos, 8);
	prj->reset_perspective_projection();		// Restore the other coordinate
}

void Entity::draw_sphere(const Projector* prj, const Navigator * nav, const Mat4d& mat, float screen_sz)
{
	// Adapt the number of facets according with the size of the sphere for optimization
	int nb_facet;
	//~ int nb_facet = (int)(screen_sz*80/50);
	//~ if (nb_facet % 2) nb_facet--;
	//~ if (nb_facet<10) nb_facet = 10;
	//~ if (nb_facet>80) nb_facet = 80;
	
	//cout << "Sphere for " << englishName << " with " << nb_facet << " facets. " << screen_sz << " \n";

	//~ if (useModel3D) {
		if (screen_sz < 40){
			nb_facet = 10;
			currentModel3D = obj->low;
			//~ cout << "Planet " << englishName << " with low_poly size " << screen_sz << " \n";
		}
		else
			if (screen_sz >200) {
				nb_facet = 80;
				currentModel3D = obj->high;
				//~ cout << "Planet " << englishName << " with high_poly size " << screen_sz << " \n";
				//~ if (screen_sz >400)
					//~ nb_facet = 80;
			}
			else {
				nb_facet=60;
				//~ cout << "Planet " << englishName << " with medium_poly size " << screen_sz << " \n";
				currentModel3D = obj->medium;
			}
	//~ }

	glEnable(GL_TEXTURE_2D);
	glEnable(GL_CULL_FACE);
	glDisable(GL_BLEND);

	if (flag_lighting) glEnable(GL_LIGHTING);
	else {
		glDisable(GL_LIGHTING);
		glColor3fv(color);
	}
	glBindTexture(GL_TEXTURE_2D, tex_map->getID());

	// Rotate and add an extra quarter rotation so that the Entity texture map
	// fits to the observers position. No idea why this is necessary,
	// perhaps some openGl strangeness, or confusing sin/cos.
	//prj->sSphere(radius*sphere_scale, one_minus_oblateness, nb_facet, nb_facet,
	//             mat * Mat4d::zrotation(M_PI/180*(axis_rotation + 90.)));


	// TODO Shaders need to be consolidated and rationalized in a better way
	bool useShader = /*Entity::flagShaders && */(bumpEnable || nightEnable || ringedEnable);
	if(ringedEnable && !isSatellite() && !hasSatellite() && !rings) useShader = false;

	if(useShader) {

		programObject.enable();

		if (bumpEnable)
		{
			programObject.setParam("TextureMap", tex_map);
			programObject.setParam("NormalTexture", tex_norm);
		}


		if (nightEnable) {
			programObject.setParam("DayTexture", tex_map);
			programObject.setParam("NightTexture", tex_night);
			programObject.setParam("SpecularTexture", tex_specular);

			if(tex_cloud) {
				programObject.setParam("CloudTexture", tex_cloud);
				programObject.setParam("CloudNormalTexture", tex_norm_cloud);
			}
			programObject.setParam("Clouds", Entity::flagClouds && tex_norm_cloud);
		}

		if (ringedEnable) {
			// Handle generic or ringed planets
			programObject.setParam("Texture", tex_map);

			if(rings && ring_shadow) {
				programObject.setParam("RingTexture", rings->get_tex());
				programObject.setParam("RingInnerRadius", rings->get_inner_radius());
				programObject.setParam("RingOuterRadius", rings->get_size());
				programObject.setParam("Ambient", 0.04);
			} else {
				programObject.setParam("RingOuterRadius", 0.0);  // No ring shadow to render
				programObject.setParam("Ambient", 0.0);
			}

			if(!flag_lighting) programObject.setParam("Ambient", 1.0);

		}

		if( bumpEnable || nightEnable || ringedEnable ) {
			// Handle moon and parent shadowing
			int index=1;

			if(tex_eclipse_map->getID()) {
				programObject.setParam("LightPosition", eye_sun);
				programObject.setParam("ShadowTexture", tex_eclipse_map);
				programObject.setParam("SunHalfAngle", sun_half_angle);
				
				double length;
				double moonDotLight;
				Vec3d tmp(0,0,0);
				Vec3d tmp2(0.4, 0.12, 0.0);
				
				if(getEnglishName() == "Moon") programObject.setParam("UmbraColor", tmp2);
				else programObject.setParam("UmbraColor", tmp);
				
				Vec3d planet_helio = get_heliocentric_ecliptic_pos();
				Vec3d light = -planet_helio;
				light.normalize();
				
				list<Entity*>::iterator iter;
				for(iter=satellites.begin(); iter!=satellites.end() && index <= 4; iter++) {
					tmp2 = (*iter)->get_heliocentric_ecliptic_pos() - planet_helio;
					length = tmp2.length();
					tmp2.normalize();
					moonDotLight = tmp2.dot(light);
					if(moonDotLight > 0 && length*sin(acos(moonDotLight)) <= radius + 2*(*iter)->getRadius()) {
						//cout << "Adding satellite: " << (*iter)->getEnglishName() << endl;
						//cout << (*iter)->getEnglishName() << " had dot of " << tmp2.dot(light) << endl;
						tmp = nav->get_helio_to_eye_mat() * (*iter)->get_heliocentric_ecliptic_pos();
						
						ostringstream oss;		
						oss << index;
						
						programObject.setParam(string("MoonPosition") + oss.str(), tmp);
						programObject.setParam(string("MoonRadius") + oss.str(), (*iter)->getRadius());
						index++;
					}
				}
				
				if(isSatellite() && parent->getEnglishName() != "Sun" && index <= 4) {
					// Parent may shadow this satellite
					ostringstream oss;		
					oss << index;
					
					tmp = nav->get_helio_to_eye_mat() * parent->get_heliocentric_ecliptic_pos();
					programObject.setParam(string("MoonPosition") + oss.str(), tmp);
					programObject.setParam(string("MoonRadius") + oss.str(), parent->getRadius());
					index++;
				}
				
			}

			// clear any leftover values
			for(; index<=4; index++) {
				ostringstream oss;		
				oss << index;
				programObject.setParam(string("MoonRadius") + oss.str(), 0.0); // No moon data
			}
		}
	}


	if (useModel3D)
		prj->sModel3D(currentModel3D, 1, one_minus_oblateness, radius, sphere_scale, mat * Mat4d::zrotation(M_PI/180*(axis_rotation + 0.)), 0, useShader);
	else
		prj->sSphere(radius, sphere_scale, one_minus_oblateness, nb_facet, nb_facet, mat * Mat4d::zrotation(M_PI/180*(axis_rotation + 90.)), 0, useShader);


	if(useShader) programObject.disable();
	else if(Entity::flagClouds && tex_cloud && tex_cloud->getID()) {
		// Non shader cloud fallback
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glBindTexture(GL_TEXTURE_2D, tex_cloud->getID());
		
		
		if (useModel3D)
			prj->sModel3D(currentModel3D,1,one_minus_oblateness,(radius * 1.002)*sphere_scale,1.0,mat * Mat4d::zrotation(M_PI/180*(axis_rotation + 0.)),/*tex_map->getID(),*/0,false);
		else
			prj->sSphere((radius * 1.002)*sphere_scale, one_minus_oblateness, nb_facet, nb_facet, mat * Mat4d::zrotation(M_PI/180*(axis_rotation + 90.)), 0, false);
	}

	glDisable(GL_CULL_FACE);
	glDisable(GL_LIGHTING);

}

// Orthographic version
void Entity::draw_halo(const Navigator* nav, const Projector* prj, const ToneReproductor* eye)
{
	float cmag;

	// johannes 20070216 mag bug fix
	float fov_q = prj->get_fov();
	if (fov_q > 60) fov_q = 60;
	fov_q = 1.f/(fov_q*fov_q);

	float rmag = sqrtf(eye->adapt_luminance((expf(-0.92103f*(compute_magnitude(nav->get_observer_helio_pos()) + 12.12331f)) * 108064.73f) * fov_q)) * 30.f * Entity::object_scale;

	cmag = 1.f;

	// if size of star is too small (blink) we put its size to 1.2 --> no more blink
	// And we compensate the difference of brighteness with cmag
	if (rmag<1.2f) {
		if (rmag<0.3f) return;
		cmag=rmag*rmag/1.44f;
		rmag=1.2f;
	} else {

		float limit = Entity::object_size_limit/1.8;
		if (rmag>limit) {
			rmag = limit + sqrt(rmag-limit)/(limit + 1);

			if (rmag > Entity::object_size_limit) {
				rmag = Entity::object_size_limit;
			}
		}
	}

	glBlendFunc(GL_ONE, GL_ONE);
	float screen_r = get_on_screen_size(prj, nav);
	cmag *= 0.5*rmag/screen_r;
	if (cmag>1.f) cmag = 1.f;

	if (rmag<screen_r) {
		cmag*=rmag/screen_r;
		rmag = screen_r;
	}

	prj->set_orthographic_projection();    	// 2D coordinate

	glBindTexture(GL_TEXTURE_2D, tex_halo->getID());
	glEnable(GL_BLEND);
	glDisable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);
	glColor3f(color[0]*cmag, color[1]*cmag, color[2]*cmag);
	glTranslatef(screenPos[0], screenPos[1], 0.f);
	glBegin(GL_QUADS);
	glTexCoord2i(0,0);
	glVertex3f(-rmag, rmag,0.f);	// Bottom Left
	glTexCoord2i(1,0);
	glVertex3f( rmag, rmag,0.f);	// Bottom Right
	glTexCoord2i(1,1);
	glVertex3f( rmag,-rmag,0.f);	// Top Right
	glTexCoord2i(0,1);
	glVertex3f(-rmag,-rmag,0.f);	// Top Left
	glEnd();

	prj->reset_perspective_projection();		// Restore the other coordinate
}


// Depth buffer version
// NOTE: may flash if draw near center of Entity while tracking
void Entity::draw_halo(const Navigator* nav, const Projector* prj, const ToneReproductor* eye, const Mat4d& mat)
{
	float cmag;

	// johannes 20070216 mag bug fix
	float fov_q = prj->get_fov();
	if (fov_q > 60) fov_q = 60;
	fov_q = 1.f/(fov_q*fov_q);

	float rmag = sqrtf(eye->adapt_luminance((expf(-0.92103f*(compute_magnitude(nav->get_observer_helio_pos()) + 12.12331f)) * 108064.73f) * fov_q)) * 30.f * Entity::object_scale;

	cmag = 1.f;

	// if size of star is too small (blink) we put its size to 1.2 --> no more blink
	// And we compensate the difference of brighteness with cmag
	if (rmag<1.2f) {
		if (rmag<0.3f) return;
		cmag=rmag*rmag/1.44f;
		rmag=1.2f;
	} else {

		float limit = Entity::object_size_limit/1.8;
		if (rmag>limit) {
			rmag = limit + sqrt(rmag-limit)/(limit + 1);

			if (rmag > Entity::object_size_limit) {
				rmag = Entity::object_size_limit;
			}
		}
	}

	float screen_r = get_on_screen_size(prj, nav);
	cmag *= 0.5*rmag/screen_r;
	if (cmag>1.f) cmag = 1.f;

	if (rmag<screen_r) {

		if(rmag < 10*screen_r ) return;  // No need to draw if body is drawn fully (interferes with orbit lines)

		cmag*=rmag/screen_r;
		rmag = screen_r;
	}

	glPushMatrix();
	glLoadMatrixd(mat);

	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE);

	glBindTexture(GL_TEXTURE_2D, tex_halo->getID());
	glEnable(GL_TEXTURE_2D);
	glColor3f(color[0]*cmag, color[1]*cmag, color[2]*cmag);

	glBegin(GL_QUADS);

	glTexCoord2i(0,0);  // Bottom Left
	prj->oVertex3(screenPos[0]-rmag, screenPos[1]-rmag, screenPos[2]-5*radius, mat);
	glTexCoord2i(1,0);
	prj->oVertex3(screenPos[0]+rmag, screenPos[1]-rmag, screenPos[2]-5*radius, mat);
	glTexCoord2i(1,1);
	prj->oVertex3(screenPos[0]+rmag, screenPos[1]+rmag, screenPos[2]-5*radius, mat);
	glTexCoord2i(0,1);
	prj->oVertex3(screenPos[0]-rmag, screenPos[1]+rmag, screenPos[2]-5*radius, mat);

	glEnd();

	glPopMatrix();

}


// Orthographic version
void Entity::draw_point_halo(const Navigator* nav, const Projector* prj, const ToneReproductor* eye)
{
	float cmag;
	
	// already in svn from johannes 20070216 mag bug fix
	float fov_q = prj->get_fov();
	if (fov_q > 60) fov_q = 60;
	fov_q = 1.f/(fov_q*fov_q);

	// float rmag = sqrtf(eye->adapt_luminance(term1 * fov_q)) * 30.f;
	float rmag = sqrtf(eye->adapt_luminance((expf(-0.92103f*(compute_magnitude(nav->get_observer_helio_pos()) +
	                                        12.12331f)) * 108064.73f) * fov_q)) *
											30.f * Entity::object_scale;
	cmag = 1.f;

	// if size of star is too small (blink) we put its size to 1.2 --> no more blink
	// And we compensate the difference of brighteness with cmag
	if (rmag<0.3f) return;
	cmag=rmag*rmag/(1.4f*1.4f);
	rmag=1.4f;

	glBlendFunc(GL_ONE, GL_ONE);
	float screen_r = get_on_screen_size(prj, nav);
	cmag *= rmag/screen_r;
	if (cmag>1.f) cmag = 1.f;

	if (rmag<screen_r) {
		cmag*=rmag/screen_r;
		rmag = screen_r;
	}
	prj->set_orthographic_projection();    	// 2D coordinate

	glBindTexture(GL_TEXTURE_2D, tex_halo->getID());
	glEnable(GL_BLEND);
	glDisable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);
	glColor3f(color[0]*cmag, color[1]*cmag, color[2]*cmag);
	glTranslatef(screenPos[0], screenPos[1], 0.f);
	glBegin(GL_QUADS);
	glTexCoord2i(0,0);
	glVertex3f(-rmag, rmag,0.f);	// Bottom Left
	glTexCoord2i(1,0);
	glVertex3f( rmag, rmag,0.f);	// Bottom Right
	glTexCoord2i(1,1);
	glVertex3f( rmag,-rmag,0.f);	// Top Right
	glTexCoord2i(0,1);
	glVertex3f(-rmag,-rmag,0.f);	// Top Left
	glEnd();

	prj->reset_perspective_projection();		// Restore the other coordinate
}


// Depth buffer version
void Entity::draw_point_halo(const Navigator* nav, const Projector* prj, const ToneReproductor* eye, const Mat4d& mat)
{
	float cmag;

	float fov_q = prj->get_fov();
	if (fov_q > 60) fov_q = 60;
	fov_q = 1.f/(fov_q*fov_q);

	// float rmag = sqrtf(eye->adapt_luminance(term1 * fov_q)) * 30.f;
	float rmag = sqrtf(eye->adapt_luminance((expf(-0.92103f*(compute_magnitude(nav->get_observer_helio_pos()) +
	                                        12.12331f)) * 108064.73f) * fov_q)) *
	             30.f * Entity::object_scale;


	cmag = 1.f;

	// if size of star is too small (blink) we put its size to 1.2 --> no more blink
	// And we compensate the difference of brighteness with cmag
	if (rmag<0.3f) return;
	cmag=rmag*rmag/(1.4f*1.4f);
	rmag=1.4f;

	glBlendFunc(GL_ONE, GL_ONE);
	float screen_r = get_on_screen_size(prj, nav);
	cmag *= rmag/screen_r;
	if (cmag>1.f) cmag = 1.f;

	if (rmag<screen_r) {
		cmag*=rmag/screen_r;
		rmag = screen_r;
	}


	glPushMatrix();
	glLoadMatrixd(mat);

	glEnable(GL_BLEND);
	glDisable(GL_LIGHTING);

	glBlendFunc(GL_ONE, GL_ONE);

	glBindTexture(GL_TEXTURE_2D, tex_halo->getID());
	glEnable(GL_TEXTURE_2D);
	glColor3f(color[0]*cmag, color[1]*cmag, color[2]*cmag);

	glBegin(GL_QUADS);

	glTexCoord2i(0,0);  // Bottom Left
	prj->oVertex3(screenPos[0]-rmag, screenPos[1]-rmag, screenPos[2], mat);
	glTexCoord2i(1,0);
	prj->oVertex3(screenPos[0]+rmag, screenPos[1]-rmag, screenPos[2], mat);
	glTexCoord2i(1,1);
	prj->oVertex3(screenPos[0]+rmag, screenPos[1]+rmag, screenPos[2], mat);
	glTexCoord2i(0,1);
	prj->oVertex3(screenPos[0]-rmag, screenPos[1]+rmag, screenPos[2], mat);

	glEnd();
	glPopMatrix();

}


//just for Sun
void Entity::draw_big_halo(const Navigator* nav, const Projector* prj, const ToneReproductor* eye)
{
	glBlendFunc(GL_ONE, GL_ONE);
	float screen_r = get_on_screen_size(prj, nav);

	float rmag = big_halo_size/2;

	float cmag = rmag/screen_r;
	if (cmag>1.f) cmag = 1.f;

	if (rmag<screen_r*2) {
		cmag*=rmag/(screen_r*2);
		rmag = screen_r*2;
	}

	prj->set_orthographic_projection();    	// 2D coordinate

	glBindTexture(GL_TEXTURE_2D, tex_big_halo->getID());
	glEnable(GL_BLEND);
	glDisable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);
	glColor3f(color[0]*cmag, color[1]*cmag, color[2]*cmag);
	glTranslatef(screenPos[0], screenPos[1], 0.f);
	glBegin(GL_QUADS);
	glTexCoord2i(0,0);
	glVertex3f(-rmag, rmag,0.f);	// Bottom Left
	glTexCoord2i(1,0);
	glVertex3f( rmag, rmag,0.f);	// Bottom Right
	glTexCoord2i(1,1);
	glVertex3f( rmag,-rmag,0.f);	// Top Right
	glTexCoord2i(0,1);
	glVertex3f(-rmag,-rmag,0.f);	// Top Left
	glEnd();

	prj->reset_perspective_projection();		// Restore the other coordinate
}


// draw orbital path of Planet
void Entity::draw_orbit_3d(const Navigator * nav, const Projector* prj, const Mat4d &mat)
{

	if (!orbit_fader.getInterstate()) return;
	if (!visibilityFader.getInterstate()) return;

	// only draw moon orbits as zoom in
	// TODO: could fade in

	Vec3d onscreen;

	if (!re.sidereal_period) return; // TODO change name to visualization_period

	glPushMatrix();
	glLoadMatrixd(mat);

	glEnable(GL_DEPTH_TEST);

	// Normal transparency mode
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_CULL_FACE);
	glColor4f(orbit_color[0], orbit_color[1], orbit_color[2], orbit_fader.getInterstate()*visibilityFader.getInterstate());

	Vec3d p[4];
	Vec3d pos = get_ecliptic_pos();
	int d, offset;
	bool on = false;
	bool spline = true;

	for ( int n=0; n<=ORBIT_SEGMENTS; n++) {
		
		d = n;
		offset = abs(ORBIT_SEGMENTS/2 - d);

		if(!close_orbit && (n < 2 || ORBIT_SEGMENTS - n < 3)) {
			// Not enough points at ends to create a spline
			if(n == ORBIT_SEGMENTS) break;
			p[2] = orbitPoint[d];
			spline = false;			
		} else {
			spline = true;
			// Draw interpolated spline
			for(int i=0; i<4; i++) {
				if((d + i-1) == ORBIT_SEGMENTS/2) {
					// special case - use current Entity position as center vertex so that draws
					// on it's orbit all the time (since segmented rather than fully smooth curve)
					p[i] = pos;
				} else if((d + i-1) < 0)
					p[i] = orbitPoint[ORBIT_SEGMENTS + (d + i-1)];
				else if((d + i-1) >= ORBIT_SEGMENTS)
					p[i] = orbitPoint[(d + i-1) - ORBIT_SEGMENTS];
				else
					p[i] = orbitPoint[d + i-1];
			}
		}

		if (prj->project_custom(p[2], onscreen, mat)) {
			if (!on) glBegin(GL_LINE_STRIP);
			
			if(spline) drawSpline3d(prj, mat, (offset<2 ? 1 : 3), p[0], p[1], p[2], p[3]);
			else prj->sVertex3( orbitPoint[d][0], orbitPoint[d][1], orbitPoint[d][2], mat);
			
			on = true;
		} else if ( on ) {
			glEnd();
			on = false;
		}
	}
	
	if (on) glEnd();

	glPopMatrix();
	glDisable(GL_DEPTH_TEST);
}

// draw orbital path of Entity in 2d projection
// for when clipping volume is too small for 3d drawing
void Entity::draw_orbit_2d(const Navigator * nav, const Projector* prj, const Mat4d &mat)
{

	if (!orbit_fader.getInterstate()) return;
	if (!visibilityFader.getInterstate()) return;

	Vec3d onscreen;

	if (!re.sidereal_period) return;

	prj->set_orthographic_projection();    // 2D coordinate

	// Normal transparency mode
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glEnable(GL_BLEND);
	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glColor4f(orbit_color[0], orbit_color[1], orbit_color[2], orbit_fader.getInterstate()*visibilityFader.getInterstate());

	Vec3d p[4];
	Vec3d pos = get_ecliptic_pos();
	int d, offset;
	bool on = false;
	bool spline = true;

	for ( int n=0; n<=ORBIT_SEGMENTS; n++) {
		
		d = n;
		offset = abs(ORBIT_SEGMENTS/2 - d);

		if(!close_orbit && (n < 2 || ORBIT_SEGMENTS - n < 3)) {
			// Not enough points at ends to create a spline
			if(n == ORBIT_SEGMENTS) break;
			p[2] = orbitPoint[d];
			spline = false;			
		} else {
			spline = true;
			// Draw interpolated spline
			for(int i=0; i<4; i++) {
				if((d + i-1) == ORBIT_SEGMENTS/2) {
					// special case - use current Entity position as center vertex so that draws
					// on it's orbit all the time (since segmented rather than fully smooth curve)
					p[i] = pos;
				} else if((d + i-1) < 0)
					p[i] = orbitPoint[ORBIT_SEGMENTS + (d + i-1)];
				else if((d + i-1) >= ORBIT_SEGMENTS)
					p[i] = orbitPoint[(d + i-1) - ORBIT_SEGMENTS];
				else
					p[i] = orbitPoint[d + i-1];
			}
		}

		if (prj->project_custom(p[2], onscreen, mat)) {
			if (!on) 
				glBegin(GL_LINE_STRIP);
			if(spline)
				drawSpline2d(prj, mat, (offset < 2 ? 1 : 15), p[0], p[1], p[2], p[3]);
			else 
				glVertex3d(onscreen[0], onscreen[1], 0);
			
			on = true;
		} 
		else if ( on ) {
			glEnd();
			on = false;
		}
	}

	if (on) glEnd();

	prj->reset_perspective_projection();		// Restore the other coordinate

	glEnable(GL_BLEND);
	glDisable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);
}


// draw trail of Entity as seen from earth
void Entity::draw_trail(const Navigator * nav, const Projector* prj)
{
	float fade = trail_fader.getInterstate();

	if (!fade) return;

	if (trail.empty()) return;

	//  if(!re.sidereal_period) return;   // limits to planets

	Vec3d onscreen1;
	Vec3d onscreen2;

	prj->set_orthographic_projection();    // 2D coordinate

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glEnable(GL_BLEND);
	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);

	list<TrailPoint>::iterator iter;
	list<TrailPoint>::iterator begin = trail.begin();
	begin++;

	float segment = 0;
	bool line_on = 0;

	// TODO: take out viewport checks so doesn't flicker on/off when zoomed in?
	// draw final segment to finish at current Entity position
	if ( !first_point && prj->project_earth_equ_line_check( (*trail.begin()).point, onscreen2, get_earth_equ_pos(nav), onscreen1) ) {

		if (!line_on) {
			glBegin(GL_LINE_STRIP);
			line_on = 1;
		}

		glColor3fv(trail_color*fade);
		glVertex3d(onscreen1[0], onscreen1[1], 0);

		//    glColor3fv(trail_color*fade* (1 - .9*segment/MaxTrail) );
		glColor3fv(trail_color*fade* (1 - .9*segment/trail.size()) );
		glVertex3d(onscreen2[0], onscreen2[1], 0);
	}

	if (trail.size() > 1) {

		for (iter=begin; iter != trail.end(); iter++) {

			segment++;

			if ( prj->project_earth_equ_check( (*iter).point, onscreen1 )) {
				if (!line_on) {
					glBegin(GL_LINE_STRIP);
					line_on = 1;
				}

				//	glColor3fv(trail_color*fade* (1 - .9*segment/MaxTrail));
				glColor3fv(trail_color*fade* (1 - .9*segment/trail.size()));
				glVertex3d(onscreen1[0], onscreen1[1], 0);

			} else {
				line_on = 0;
				glEnd();
			}

		}
	}

	if (line_on) glEnd();

	prj->reset_perspective_projection();		// Restore the other coordinate

	glEnable(GL_BLEND);
	glDisable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);

}

// update trail points as needed
void Entity::update_trail(const Navigator* nav)
{
	if (!trail_on) return;

	double date = nav->get_JDay();

	int dt=0;
	if (first_point || (dt=abs(int((date-last_trailJD)/DeltaTrail))) > MaxTrail) {
		dt=1;
		// clear old trail
		trail.clear();
		first_point = 0;
	}

	// Note that when jump by a week or day at a time, loose detail on trails
	// particularly for moon (if decide to show moon trail)
	// add only one point at a time, using current position only
	if (dt) {
		last_trailJD = date;
		TrailPoint tp;
		Vec3d v = get_heliocentric_ecliptic_pos();
		//      trail.push_front( nav->helio_to_earth_equ(v) );  // centered on earth
		tp.point = nav->helio_to_earth_pos_equ(v);
		tp.date = date;
		trail.push_front( tp );

		if ( trail.size() > (unsigned int)MaxTrail ) {
			trail.pop_back();
		}
	}

	// because sampling depends on speed and frame rate, need to clear out
	// points if trail gets longer than desired
	list<TrailPoint>::iterator iter;
	list<TrailPoint>::iterator end = trail.end();

	for ( iter=trail.begin(); iter != end; iter++) {
		if ( fabs((*iter).date - date)/DeltaTrail > MaxTrail ) {
			trail.erase(iter, end);
			break;
		}
	}
}

// Start/stop accumulating new trail data (clear old data)
void Entity::startTrail(bool b)
{
	if (b) {
		first_point = 1;
		//  printf("trail for %s: %f\n", name.c_str(), re.sidereal_period);
		// only interested in trails for planets
		// if(parent && parent->getEnglishName() == "Sun") trail_on = 1;
		if (!is_satellite && parent) trail_on = 1; // No trail for Sun or moons
		//	if(re.sidereal_period > 0) trail_on = 1;
	} else {
		trail_on = 0;
	}
}

void Entity::translateName(Translator& trans) {
	nameI18 = trans.translateUTF8(englishName);
}

void Entity::update(int delta_time)
{
	visibilityFader.update(delta_time);
	hint_fader.update(delta_time);
	orbit_fader.update(delta_time);
	trail_fader.update(delta_time);
}

void Entity::removeSatellite(Entity *planet)
{
	list<Entity *>::iterator iter;
	for (iter=satellites.begin(); iter != satellites.end(); iter++) {
		//      if( (*iter)->getEnglishName() == planet->getEnglishName() ) {
		if ( (*iter) == planet ) {
			satellites.erase(iter);
			break;
		}
	}
	/*
	// DEBUG
	cout << "Satellites for " << englishName << endl;
	for(iter=satellites.begin(); iter != satellites.end(); iter++)
	  cout << "\t" << (*iter)->getEnglishName() << endl;
	*/
}

// Update bounding radii from child up to parent(s)
void Entity::updateBoundingRadii()
{
	calculateBoundingRadius();
	Entity *p;
	p = parent;
	while (p) {
		p->calculateBoundingRadius();
		p = p->parent;
	}
}

// Calculate a bounding radius in AU
// For a Entity with satellites, this bounds the most distant satellite orbit
// for planets with no elliptical satellites, this is the ring (if one) 
// or Entity radius
// Caches result until next call, can retrieve with getBoundingRadius
double Entity::calculateBoundingRadius()
{
	// TODO: only have boundary function for elliptical orbits right now

	if ( hidden && satellites.empty() ) {
		boundingRadius = -1;
		return -1;  // special case, not visible
	}

	double d = radius;

	if (rings) d = rings->get_size();

	list<Entity *>::const_iterator iter;
	list<Entity *>::const_iterator end = satellites.end();

	//	cout << englishName << endl;

	double r;
	for ( iter=satellites.begin(); iter != end; iter++) {

		r = (*iter)->getBoundingRadius();
		//		cout << "\t" << (*iter)->getEnglishName() << " : " << r << endl;
		if ( r > d ) d = r;
	}

	// if we are a planet, we want the boundary radius including all satellites
	// if we are a satellite, we want the full orbit radius as well
	if ( is_satellite ) {
		//		cout << "Sat: " << d << " bnd rad: " << orbit_bounding_radius << endl;
		if ( d > orbit_bounding_radius + radius ) boundingRadius = d;
		else boundingRadius = orbit_bounding_radius + radius;
	} else boundingRadius = d;

	//  cout << "Bounding radius for " << englishName << " = " << boundingRadius << endl;

	return boundingRadius;
}

// See if simple Entity not needing depth buffer
// (no rings or satellites)
// 20080805 or it has satellites but orbit lines are not on
bool Entity::isSimplePlanet()
{

	if (rings) return false;

	if ( satellites.empty() ) return true;
	return false;
}

// draw Entity axis
void Entity::draw_axis(const Projector* prj, const Mat4d& mat, float screen_sz)
{
	glEnable(GL_BLEND);
	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_LINE_SMOOTH);

	glColor3f(1,0,0); 
	glPushMatrix();
	glLoadMatrixd(mat);

	//axis
	glBegin(GL_LINE_STRIP);
	prj->sVertex3(0, 0,  1.4 * radius * sphere_scale, mat);
	prj->sVertex3(0, 0, -1.4 * radius * sphere_scale, mat);
	glEnd();
	
	//equatorial circle
	glBegin(GL_LINE_LOOP);
	for(int i=0; i<180; i++) {
		prj->sVertex3(cos(i*2*M_PI/180)* radius * sphere_scale*1.03, sin(i*2*M_PI/180)* radius * sphere_scale*1.03, 0,  mat);
	}
	glEnd();
	glDisable(GL_LINE_SMOOTH);
	glPopMatrix();
}


// Catmull-Rom Spline
// 0 >= t <= 1, returns point on curve between points v1 and v2

Vec3d Entity::calculateSplinePoint(double t, Vec3d &v0, Vec3d &v1, Vec3d &v2, Vec3d &v3)
{
	Vec3d c2, c3, c4;

	c2 = -.5*v0 + .5*v2;
	c3 = v0 -2.5*v1 + 2.*v2 -0.5*v3;
	c4 = -0.5*v0 + 1.5*v1 -1.5*v2 + 0.5*v3;

	return(((c4*t + c3)*t +c2)*t + v1);
}

// Draw line segments approximating an interpolated spline
void Entity::drawSpline3d(const Projector *prj, const Mat4d &mat, int segments, Vec3d &v0, Vec3d &v1, Vec3d &v2, Vec3d &v3)
{
	Vec3d point;

	for(int n=1; n<=segments; n++) {
		point = calculateSplinePoint(double(n)/segments, v0, v1, v2, v3);
		prj->sVertex3( point[0], point[1], point[2], mat);
	}
}

// Draw line segments approximating an interpolated spline
void Entity::drawSpline2d(const Projector *prj, const Mat4d &mat, int segments, Vec3d &v0, Vec3d &v1, Vec3d &v2, Vec3d &v3)
{
	Vec3d point, onscreen;

	for(int n=1; n<=segments; n++) {
		point = calculateSplinePoint((double)n/segments, v0, v1, v2, v3);
		if(prj->project_custom(point, onscreen, mat))
			glVertex3d( onscreen[0], onscreen[1], 0);
	}
}
