/*
 * Stellarium360 astronomy simulation and visualization
 *
 * Copyright (C) 2002 Fabien Chereau
 * Copyright (C) 2009 Digitalis Education Solutions, Inc.
 * Copyright (C) 2014 of the LSS Team & Association Sirius
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Stellarium360 is a free open project of of LSS team
 * See the TRADEMARKS file for free open project usage requirements.
 *
 */

#ifndef _ENTITY_HPP_
#define _ENTITY_HPP_

#include "program_object.hpp"
#include "object_base.hpp"
#include "stellarium360.hpp"
#include "utility.hpp"
#include "s_font.hpp"
#include "tone_reproductor.hpp"
#include "vecmath.hpp"
#include "callbacks.hpp"
#include "fader.hpp"
#include "translator.hpp"
#include "orbit.hpp"
#include "model3D.hpp"
#include "ring.hpp"
#include <list>
#include <string>


// epoch J2000: 12 UT on 1 Jan 2000
#define J2000 2451545.0

#define ORBIT_SEGMENTS 180

using namespace std;

struct TrailPoint {
	Vec3d point;
	double date;
};

enum ENTITY_TYPE {SUN = 0, PLANET = 1 , MOON = 2, DWARF = 3, ASTEROID = 4, PLUTOID = 5, UNKNOWN = 10};

// Class used to store orbital elements
class RotationElements
{
public:
	RotationElements(void)
	  : period(1.), offset(0.), epoch(J2000),
	    obliquity(0.), ascendingNode(0.), precessionRate(0.), 
		sidereal_period(0.), axialTilt(0.) {}
	float period;        // rotation period
	float offset;        // rotation at epoch
	double epoch;
	float obliquity;     // tilt of rotation axis w.r.t. ecliptic
	float ascendingNode; // long. of ascending node of equator on the ecliptic
	float precessionRate; // rate of precession of rotation axis in rads/day
	double sidereal_period; // sidereal period (Entity year in earth days)
	float axialTilt; // Only used for tropic lines on planets
};



class Entity : public ObjectBase
{
public:

	// TODO, shader texture flags are redundant
	Entity(Entity *parent,
	       const string& englishName,
	       ENTITY_TYPE _typePlanet,
	       int flagHalo,
	       int flag_lighting,
	       double radius,
	       double oblateness,
           Vec3f color,
           Vec3f label_color,
           Vec3f orbit_color,
           Vec3f trail_color,
           float _sol_local_day,
	       float albedo,
	       const string& tex_map_path,
	       const string& tex_map_name,
	       const string& tex_halo_name,
	       Orbit *orbit,
	       bool close_orbit,
	       const string& model_name,
	       model3D_mgr * _myObjMgr,
	       bool hidden,
	       bool _deleteable,
	       double orbit_bounding_radius,
		   const string& tex_norm_name,
		   const string& tex_night_name,
		   const string& tex_specular_name,
		   const string& tex_cloud_name,
		   const string& tex_norm_cloud_name);
	~Entity();

	double getRadius(void) const {
		return radius;
	}

	double getOblateness(void) const {
		return 1-one_minus_oblateness;
	}

	// Return the information string "ready to print" :)
	string getSkyLabel(const Navigator * nav) const;
	string getInfoString(const Navigator * nav) const;
	 string getTypePlanet(const ENTITY_TYPE str) const;
	
	//is still use ?
	void getRaDeValue(const Navigator *nav ,double *ra, double *de) const{
	}
	void getAltAz(const Navigator * nav, double *alt, double *az) const;
	
	string getShortInfoString(const Navigator * nav) const;
	string getShortInfoNavString(const Navigator * nav) const;
	double get_close_fov(const Navigator * nav) const;
	double get_satellites_fov(const Navigator * nav) const;
	double get_parent_satellites_fov(const Navigator * nav) const;
	float get_mag(const Navigator * nav) const {
		return compute_magnitude(nav);
	}
	const s_texture *getMapTexture(void) const {
		return tex_map;
	}

	Orbit *getOrbit() { return orbit; }

	/** Translate Entity name using the passed translator */
	void translateName(Translator& trans);

	// Compute the z rotation to use from equatorial to geographic coordinates
	double getSiderealTime(double jd) const;
	Mat4d getRotEquatorialToVsop87(void) const;
	void setRotEquatorialToVsop87(const Mat4d &m);

	const RotationElements &getRotationElements(void) const {return re;}

	// Compute the position in the parent Entity coordinate system
	void computePositionWithoutOrbits(double date);
	void compute_position(double date);

	// Compute the transformation matrix from the local Entity coordinate to the parent Entity coordinate
	void compute_trans_matrix(double date);

	// Get the phase angle for an observer at pos obs_pos in the heliocentric coordinate (in AU)
	double get_phase(Vec3d obs_pos) const;

	// Get the magnitude for an observer at pos obs_pos in the heliocentric coordinate (in AU)
	 float compute_magnitude(const Vec3d obs_pos) const;
	float compute_magnitude(const Navigator * nav) const;

	// Draw the Planet, if hint_ON is != 0 draw a circle and the name as well
	// Return the squared distance in pixels between the current and the
	// previous position this Entity was drawn at.
	 double draw(Projector* prj, const Navigator* nav, const ToneReproductor* eye,
	            int flag_point, bool stencil, bool depthTest, bool drawHomePlanet,
				bool selected);

	// Set the orbital elements
	void set_rotation_elements(float _period, float _offset, double _epoch,
	                           float _obliquity, float _ascendingNode, float _precessionRate, double _sidereal_period, float _axial_tilt);
	double getRotAscendingnode(void) const {
		return re.ascendingNode;
	}
	double getRotObliquity(void) const {
		return re.obliquity;
	}
	double getAxialTilt(void) const {
		return re.axialTilt;
	}

	double getSiderealDay(void) const {
		return re.period;
	}

	// Get the Entity position in the parent Entity ecliptic coordinate
	Vec3d get_ecliptic_pos() const;

	// Return the heliocentric ecliptical position
	Vec3d get_heliocentric_ecliptic_pos() const;
	void set_heliocentric_ecliptic_pos(const Vec3d &pos);

	// Compute the distance to the given position in heliocentric coordinate (in AU)
	double compute_distance(const Vec3d& obs_helio_pos);
	double get_distance(void) const {
		return distance;
	}

	ObjectRecord::OBJECT_TYPE get_type(void) const {
		return ObjectRecord::OBJECT_ENTITY;
	}

	// Return the Entity position in rectangular earth equatorial coordinate
	Vec3d get_earth_equ_pos(const Navigator *nav) const;
	// observer centered J2000 coordinates
	Vec3d getObsJ2000Pos(const Navigator *nav) const;

	string getEnglishName(void) const {
		return englishName;
	}
	string getNameI18n(void) const {
		return nameI18;
	}

	void set_rings(Ring* r) {
		rings = r;
	}

	void set_ring_shadow(bool shadow) {
		ring_shadow = shadow;
	}

	void set_sphere_scale(float s) {
		sphere_scale = s;
	}
	float get_sphere_scale(void) const {
		return sphere_scale;
	}

	const Entity *get_parent(void) const {
		return parent;
	}

	// modifiable
	Entity *getParent(void) {
		return parent;
	}

	 void set_big_halo(const string& halotexfile, const string& path);
	
	void set_halo_size(float s) {
		big_halo_size = s;
	}

	static void set_font(s_font* f) {
		planet_name_font = f;
	}

	//cela parle de magnitude avec les étoiles
	static void setScale(float s) {
		object_scale = s;
	}
	//cela parle de magnitude avec les étoiles
	static float getScale(void) {
		return object_scale;
	}

	void setSizeScale(float s) {
		sphere_scale = s;
	}
	float getSizeScale(void) {
		return sphere_scale;
	}

	static void setSizeLimit(float s) {
		object_size_limit = s;
	}
	static float getSizeLimit(void) {
		return object_size_limit;
	}

	void setLabelColor(const Vec3f& lc) {
		label_color = lc;
	}
	const Vec3f& getLabelColor(void) {
		return label_color;
	}

	void setOrbitColor(const Vec3f& oc) {
		orbit_color = oc;
	}
	const Vec3f& getOrbitColor() {
		return orbit_color;
	}

	// draw orbital path of body
	void draw_orbit_2d(const Navigator * nav, const Projector* prj, const Mat4d &mat);
	void draw_orbit_3d(const Navigator * nav, const Projector* prj, const Mat4d &mat);

	void update_trail(const Navigator* nav);
	void draw_trail(const Navigator * nav, const Projector* prj);
	
	void setTrailColor(const Vec3f& c) {
		trail_color = c;
	}
	
	const Vec3f& getTrailColor() {
		return trail_color;
	}

	//! Start/stop accumulating new trail data (clear old data)
	void startTrail(bool b);

	void update(int delta_time);

	void setFlagHints(bool b) {
		hint_fader = b;
	}
	bool getFlagHints(void) const {
		return hint_fader;
	}

	void setFlagOrbit(bool b) {
		orbit_fader = b;
	}

	bool getFlagAxis(void) const {
		return drawaxis;
	}

	void setFlagAxis(bool b) {
		drawaxis = b;
	}
	bool getFlagOrbit(void) const {
		return orbit_fader;
	}

	void setFlagTrail(bool b) {
		if (b == trail_fader) return;
		trail_fader = b;
		startTrail(b);
	}
	bool getFlagTrail(void) const {
		return trail_fader;
	}

	static void setFlagShow(bool b) {
		Entity::flagShow = b;
	}

	static bool getFlagShow(void) {
		return Entity::flagShow;
	}

	static void setFlagClouds(bool b) {
		Entity::flagClouds = b;
	}

	static bool getFlagClouds(void) {
		return Entity::flagClouds;
	}


	void setVisibility(bool b) {
		visibilityFader = b;
	}

	bool getVisibility(void) {
		return visibilityFader;
	}


	bool isDeleteable() const {
		return deleteable;    // If allowed to delete from script
	}
	bool isSatellite() const {
		return is_satellite;
	}

	// remove from parent satellite list
	void removeSatellite(Entity *planet);

	// for depth buffering of orbits
	void updateBoundingRadii();
	double calculateBoundingRadius();

	double getBoundingRadius() const {
		return boundingRadius;
	}

	// Return the radius of a circle containing the object on screen
	float get_on_screen_size(const Projector* prj, const Navigator * nav, bool orb_only = false);

	// Return the radius of a circle containing the object and satellites on screen
	float get_on_screen_bounding_size(const Projector* prj, const Navigator * nav);


	// Return the angle of the Entity orb
	float get_angular_size(const Projector* prj, const Navigator * nav);

	// See if Entity might be visible
	//	bool testVisibility(const Projector* prj, const Navigator * nav);

	// See if simple Entity not needing depth buffer
	bool isSimplePlanet();
	bool hasSatellite() {
		return !satellites.empty();
	}

	double getLastJD() const { return lastJD; }

	// TODO: this should be protected or moved to solarsystem.cpp
	static s_texture *tex_eclipse_map;  // for moon shadow lookups

//	static bool flagShaders;  // whether to use shaders when rendering

	//set flag Entity to be displayed or not
	void set_planet_hidden(bool planethidden) {
		hidden=planethidden;
	}



	bool get_planet_hidden(void){
		return hidden;
	};
	
	//return the duration of a day in the Entity 
	const float getSolLocalDay() const {
		return sol_local_day;
	}
	 
	void reinitParam() {
		hidden= hiddenInitial;
		 
	}
	
protected:
	// Draw the 3D sphere
	 void draw_sphere(const Projector* prj, const Navigator * nav, const Mat4d& mat, float screen_sz);

	// draw Entity axis
	void draw_axis(const Projector* prj, const Mat4d& mat, float screen_sz);

	// Draw the small star-like 2D halo
	void draw_halo(const Navigator* nav, const Projector* prj, const ToneReproductor* eye);
	void draw_halo(const Navigator* nav, const Projector* prj, const ToneReproductor* eye, const Mat4d& mat);

	// Draw the small star-like point
	void draw_point_halo(const Navigator* nav, const Projector* prj, const ToneReproductor* eye);
	void draw_point_halo(const Navigator* nav, const Projector* prj, const ToneReproductor* eye, const Mat4d& mat);

	// Draw the circle and name of the Planet
	void draw_hints(const Navigator* nav, const Projector* prj);

	// Draw the big halo (for sun or moon)
	 void draw_big_halo(const Navigator* nav, const Projector* prj, const ToneReproductor* eye);

	bool useParentPrecession(double jd) {
		return getOrbit()->useParentPrecession(jd);
	}

	// Interpolate a curve for orbits
	static Vec3d calculateSplinePoint(double, Vec3d &v0, Vec3d &v1, Vec3d &v2, Vec3d &v3);

	static void drawSpline2d(const Projector *prj, const Mat4d &mat, int segments, Vec3d &v0, Vec3d &v1, Vec3d &v2, Vec3d &v3);
	static void drawSpline3d(const Projector *prj, const Mat4d &mat, int segments, Vec3d &v0, Vec3d &v1, Vec3d &v2, Vec3d &v3);

	string englishName; 			// english Entity name
	string nameI18;					// International translated name
	int flagHalo;					// Set wether a little "star like" halo will be drawn
	int flag_lighting;				// Set wether light computation has to be proceed
	RotationElements re;			// Rotation param
	double radius;					// Entity radius in UA
	double one_minus_oblateness;    // (polar radius)/(equatorial radius)
	Vec3d orbitPoint[ORBIT_SEGMENTS];    // store heliocentric coordinates for drawing the orbit
	Vec3d ecliptic_pos; 			// Position in UA in the rectangular ecliptic coordinate system
	// centered on the parent Planet
	Vec3d last_pos;                 // Used when interpolating
	Vec3d next_pos;                 // Used when interpolating
	bool interpolating;             // Whether to do a linear interpolation when delta_time < deltaJD
	int interpolationDirection;     // is time going forward (+) or backward (-)?
	Vec3d screenPos;			// Used to store temporarily the 2D position on screen
	Vec3d previousScreenPos;		// The position of this Entity in the previous frame.
	ENTITY_TYPE typePlanet;			//get the type of Entity in univers: real planet, moon, dwarf ...
	
	//gestion des couleurs de l'objet
	Vec3f color;		//la couleur du halo de l'objet
	Vec3f label_color;
	Vec3f orbit_color;
	Vec3f trail_color;
	
	float sol_local_day;			//time of a sideral day in this planet
	float albedo;					// Entity albedo
	Mat4d rot_local_to_parent;
	Mat4d rot_local_to_parent_unprecessed;  // currently used for moons (Moon elliptical orbit required this)
	Mat4d mat_local_to_parent;		// Transfo matrix from local ecliptique to parent ecliptic
	float axis_rotation;			// Rotation angle of the Entity on it's axis
	s_texture * tex_map;			// Entity map texture
	s_texture * tex_halo;			// Little halo texture
	s_texture * tex_big_halo;		// Big halo texture
	s_texture * tex_norm;
	s_texture * tex_night;
	s_texture * tex_specular;
	s_texture * tex_cloud;
	s_texture * tex_shadow_cloud;
	s_texture * tex_norm_cloud;

	Vec3f eye_sun;
	Vec3f eye_planet;
	ProgramObject programObject;    // Shaders
	//ProgramObject cloudProgramObject; inutile	
	bool bumpEnable;
	bool nightEnable;
	bool ringedEnable;
	bool useModel3D;	//temporaire: transition model3D ou construction à la main.
	obj3D *obj;		//unique representation du model3D
	model3D *currentModel3D;
	model3D_mgr * myObjMgr;
	float big_halo_size;				// Halo size on screen

	Ring* rings;					// Entity rings

	double distance;				// Temporary variable used to store the distance to a given point
	// it is used for sorting while drawing

	float sphere_scale;				// Artificial scaling for better viewing

	double lastJD;
	double last_orbitJD;
	double deltaJD;
	double delta_orbitJD;
	bool orbit_cached;       // whether orbit calculations are cached for drawing orbit yet

	Orbit *orbit;            // orbit object for this body

// made non-const so can actually use this for things!
	Entity *parent;				// Entity parent i.e. sun for earth

	list<Entity *> satellites;		// satellites of the Planet

	static s_font* planet_name_font;// Font for names
	static float object_scale;
	static float object_size_limit;  // in pixels

	list<TrailPoint>trail;
	bool trail_on;  // accumulate trail data if true
	double DeltaTrail;
	int MaxTrail;
	double last_trailJD;
	bool first_point;  // if need to take first point of trail still

	static LinearFader flagShow;
	static LinearFader flagClouds;

	LinearFader hint_fader;
	LinearFader orbit_fader;
	LinearFader trail_fader;
	LinearFader visibilityFader;  // allows related lines and labels to fade in/out

	bool close_orbit; // whether to close orbit loop
	bool hidden;  // useful for fake planets used as observation positions - not drawn or labeled
	bool hiddenInitial; // sert à conserver la valeur du flag hidden apres chargement de l'objet
	bool deleteable;  // whether allowed to delete from scripts
	bool is_satellite;  // whether has a Entity as a parent
	bool drawaxis;  // display or not Entity axis

	double orbit_bounding_radius; // AU calculated at load time for elliptical orbits at least DIGITALIS
	double boundingRadius;  // Cached AU value for use with depth buffer
	static s_texture *defaultTexMap;  // Default texture map for bodies if none supplied

	bool ring_shadow; 
	double sun_half_angle; // for moon shadow calcs

};

#endif // _ENTITY_HPP_



