/*
 * Stellarium360 astronomy simulation and visualization
 *
 * Copyright (C) 2002 Fabien Chereau
 * Copyright (C) 2009-2010 Digitalis Education Solutions, Inc.
 * Copyright (C) 2014 of the LSS Team & Association Sirius
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Stellarium360 is a free open project of of LSS team
 * See the TRADEMARKS file for free open project usage requirements.
 *
 */

#include <iostream>
#include <iomanip>

#include "night_shader.hpp"
#include "bump_shader.hpp"
#include "ring_shader.hpp"
#include "ringed_shader.hpp"
#include "entity_sun.hpp"
#include "entity.hpp"
#include "navigator.hpp"
#include "projector.hpp"
#include "s_font.hpp"
//#include "s_gui.hpp"
#include "stellastro.h" // just for get_apparent_sidereal_time
#include "log.hpp"



EntitySun::EntitySun(	int flagHalo,
	       int flag_lighting,
	       double radius,
	       double oblateness,
           Vec3f color,
           Vec3f label_color,
           Vec3f orbit_color,
           float _sol_local_day,
	       float _albedo,
	       const string& tex_map_path,
	       const string& tex_map_name,
	       const string& tex_halo_name,
	       const string& model_name,
	       model3D_mgr * _myObjMgr
	): Entity(NULL,"Sun",
               ENTITY_TYPE::SUN,
               flagHalo,
               flag_lighting,
               radius,
               oblateness,
               color,
               label_color,
               orbit_color,
               trail_color,
               _sol_local_day,
               _albedo,
               tex_map_path,
               tex_map_name,
               tex_halo_name,
               orbit,
               close_orbit,
               model_name,
               _myObjMgr,
               false,
               false,
               orbit_bounding_radius,
			   "",
			   "",
			   "",
			   "",
			   "")
{
	one_minus_oblateness=(1.0-oblateness);
	interpolating=false;
	interpolationDirection=0;
	albedo=_albedo;
	axis_rotation=0.;
	tex_halo=NULL;
	tex_big_halo=NULL;
	tex_norm=NULL;
	tex_night=NULL;
	tex_specular=NULL;
	tex_cloud=NULL;
	tex_shadow_cloud=NULL;
	tex_norm_cloud=NULL;
	eye_sun.set(0.0f, 0.0f, 0.0f);
	bumpEnable=false;
	nightEnable=false;
	ringedEnable=false;
	rings=NULL;
	sphere_scale=1.f;
	orbit=orbit;

	sol_local_day = _sol_local_day;
	typePlanet = SUN;
	is_satellite = 0; // quicker lookup
	ecliptic_pos=Vec3d(0.,0.,0.);
	rot_local_to_parent = Mat4d::identity();
	rot_local_to_parent_unprecessed = Mat4d::identity();
	mat_local_to_parent = Mat4d::identity();

	if ( tex_map_path == "" ) {
		tex_map = new s_texture(tex_map_name, TEX_LOAD_TYPE_PNG_SOLID_REPEAT, 1);
		} else {
			tex_map = new s_texture(1, tex_map_path + tex_map_name, TEX_LOAD_TYPE_PNG_SOLID_REPEAT, 1);
		}

	flagHalo=false;

	ringedEnable = nightEnable = bumpEnable = false;

	DeltaTrail = 1;
	MaxTrail = 1460;
	last_trailJD = 0; // for now
	trail_on = 0;
	first_point = 1;

	nameI18 = englishName;

	visibilityFader = true;
	drawaxis=false;
	
	useModel3D= true;
	myObjMgr = _myObjMgr;
	obj= new obj3D();
	obj= myObjMgr->addStellObj("Sphere");
	if (obj==NULL)
		useModel3D = false;
}

EntitySun::~EntitySun()
{
	if (tex_map && tex_map != defaultTexMap) delete tex_map;
	tex_map = NULL;
	if (tex_halo) delete tex_halo;
	tex_halo = NULL;
	if (rings) delete rings;
	rings = NULL;
	if (tex_big_halo) delete tex_big_halo;
	tex_big_halo = NULL;
	if (tex_norm) delete tex_norm;
	tex_norm = NULL;
	if (tex_night) delete tex_night;
	tex_night = NULL;
	if (tex_specular) delete tex_specular;
	tex_specular = NULL;
	if (tex_cloud) delete tex_cloud;
	tex_cloud = NULL;
	if (tex_norm_cloud) delete tex_norm_cloud;
	tex_norm_cloud = NULL;
	if(orbit) delete orbit;
	orbit = NULL;
}

float EntitySun::compute_magnitude(Vec3d obs_pos) const
{
	float rval;
	const double sq = obs_pos.lengthSquared();
		rval = -26.73f + 2.5f*log10f(sq);
	return rval;
}

void EntitySun::set_big_halo(const string& halotexfile, const string &path)
{
	tex_big_halo = new s_texture((path!=""), path + halotexfile, TEX_LOAD_TYPE_PNG_SOLID);
}


// Draw the Entity and all the related infos : name, circle etc..
double EntitySun::draw(Projector* prj, const Navigator * nav, const ToneReproductor* eye, int flag_point, bool stencil, bool depthTest, bool drawHomePlanet,bool selected)
{
	if (hidden) return 0;

	eye_sun = nav->get_helio_to_eye_mat() * Vec3f(0.0f, 0.0f, 0.0f);

	Mat4d mat = mat_local_to_parent;
	Mat4d parent_mat = Mat4d::identity();

	// \todo account for moon orbit precession (independent of parent)
	// also does not allow for multiple levels of precession

	const Entity *p = parent;

	bool myParent = true;
	while (p && p->parent) {

		// Some orbits are already precessed, namely elp82
		if(myParent && !useParentPrecession(lastJD)) {
			mat = Mat4d::translation(p->ecliptic_pos)
				* mat
				* p->rot_local_to_parent_unprecessed; 
		} else {
			mat = Mat4d::translation(p->ecliptic_pos)
				* mat
				* p->rot_local_to_parent; 
		}
		
		parent_mat = Mat4d::translation(p->ecliptic_pos)
			* parent_mat;

		p = p->parent;

		myParent = false;
	}


	// This removed totally the Entity shaking bug!!!
	mat = nav->get_helio_to_eye_mat() * mat;
	parent_mat = nav->get_helio_to_eye_mat() * parent_mat;

	eye_planet = mat * Vec3f(0.0f, 0.0f, 0.0f);

	Vec3f lightDirection = eye_sun - eye_planet;
	sun_half_angle = atan(696000.0/AU/lightDirection.length());  // hard coded Sun radius!

	//	cout << sun_half_angle << " sun angle on " << englishName << endl;
	lightDirection.normalize();

	if (!drawHomePlanet && this == nav->getHomePlanet()) {
		if (rings) rings->draw(prj,mat,1000.0,lightDirection,eye_planet,radius);
		return 0;
	}
	
	// Compute the 2D position and check if in the screen
	float screen_sz = get_on_screen_size(prj, nav);

	// special case if big halo (Sun)
	float screen_size_with_halo = screen_sz;
	if (tex_big_halo && (big_halo_size > screen_sz))
		screen_size_with_halo = big_halo_size;

	// Check Entity visible in viewport
	bool isVisible = prj->project_custom_check(Vec3f(0,0,0), screenPos, mat, (int)(screen_size_with_halo/2));

	bool visible = isVisible;
	if(selected) visible = true;  // always draw orbit if selected, even if not on screen

	visible = true;

	// fade orbits as enter/leave view
	if (is_satellite && prj->get_fov()>30) {

		// If not in the parent Entity system
		// OR one of the other sister moons do not draw
		if (parent != nav->getHomePlanet() &&
			parent != nav->getHomePlanet()->get_parent() &&
			get_earth_equ_pos(nav).length() > 1) {  // If further than 1 AU of object if flying
			visibilityFader = false; // orbits will fade in and out now
		} else visibilityFader = visible;
	} else visibilityFader = visible;

	if ( is_satellite) 
		draw_orbit_3d(nav, prj, parent_mat);
	else {
		draw_orbit_2d(nav, prj, parent_mat);
		//			cout << "2d orbit: " << englishName << endl;
	}

	// Do not draw anything else if was not visible
	// Draw the name, and the circle if it's not too close from the body it's turning around
	// this prevents name overlaping (ie for jupiter satellites)
	float ang_dist = 300.f*atan(get_ecliptic_pos().length()/get_earth_equ_pos(nav).length())/prj->get_fov();
	if (ang_dist==0.f)
		ang_dist = 1.f; // if ang_dist == 0, the Entity is sun..

	draw_trail(nav, prj);


	if (ang_dist>0.25) {
		if (ang_dist>1.f) ang_dist = 1.f;
		draw_hints(nav, prj);
	}
	if (depthTest) glEnable(GL_DEPTH_TEST);

	if (screen_sz > 1 && isVisible) {  // huge improvement in performance
			if (stencil) glEnable(GL_STENCIL_TEST);
			if (drawaxis) 
				draw_axis(prj, mat, screen_sz);
			draw_sphere(prj, nav, mat, screen_sz);  // SL O O W
			if (stencil) glDisable(GL_STENCIL_TEST);
	}

	if (tex_halo) {
		// Workaround for depth buffer precision and near planets
		if(get_earth_equ_pos(nav).length() > 0.0001) {
			glDisable(GL_DEPTH_TEST);
			if (flag_point) draw_point_halo(nav, prj, eye);
			else draw_halo(nav, prj, eye);
		} else {
			if (flag_point) draw_point_halo(nav, prj, eye, mat);
			else draw_halo(nav, prj, eye, mat);
		}
	}

	glDisable(GL_DEPTH_TEST);

	draw_big_halo(nav, prj, eye);	

	double distanceSquared =
	    (screenPos[0] - previousScreenPos[0]) *
	    (screenPos[0] - previousScreenPos[0]) +
	    (screenPos[1] - previousScreenPos[1]) *
	    (screenPos[1] - previousScreenPos[1]);
	previousScreenPos = screenPos;
	return distanceSquared;
}

void EntitySun::draw_sphere(const Projector* prj, const Navigator * nav, const Mat4d& mat, float screen_sz)
{
		if (screen_sz < 40){
			currentModel3D = obj->low;
			//~ cout << "Planet " << englishName << " with low_poly size " << screen_sz << " \n";
		}
		else
			if (screen_sz >200) {
				currentModel3D = obj->high;
				//~ cout << "Planet " << englishName << " with high_poly size " << screen_sz << " \n";
			}
			else {
				//~ cout << "Planet " << englishName << " with medium_poly size " << screen_sz << " \n";
				currentModel3D = obj->medium;
			}

	glEnable(GL_TEXTURE_2D);
	glEnable(GL_CULL_FACE);
	glDisable(GL_BLEND);

	glDisable(GL_LIGHTING);
	glColor3fv(color);

	glBindTexture(GL_TEXTURE_2D, tex_map->getID());

	// Rotate and add an extra quarter rotation so that the Entity texture map
	// fits to the observers position. No idea why this is necessary,
	// perhaps some openGl strangeness, or confusing sin/cos.
	//prj->sSphere(radius*sphere_scale, one_minus_oblateness, nb_facet, nb_facet,
	//             mat * Mat4d::zrotation(M_PI/180*(axis_rotation + 90.)));


		prj->sModel3D(currentModel3D, 1, one_minus_oblateness, radius, sphere_scale, mat * Mat4d::zrotation(M_PI/180*(axis_rotation + 0.)), 0, false);

	glDisable(GL_CULL_FACE);
	glDisable(GL_LIGHTING);
}

//just for Sun
void EntitySun::draw_big_halo(const Navigator* nav, const Projector* prj, const ToneReproductor* eye)
{
	glBlendFunc(GL_ONE, GL_ONE);
	float screen_r = get_on_screen_size(prj, nav);

	float rmag = big_halo_size/2;

	float cmag = rmag/screen_r;
	if (cmag>1.f) cmag = 1.f;

	if (rmag<screen_r*2) {
		cmag*=rmag/(screen_r*2);
		rmag = screen_r*2;
	}

	prj->set_orthographic_projection();    	// 2D coordinate

	glBindTexture(GL_TEXTURE_2D, tex_big_halo->getID());
	glEnable(GL_BLEND);
	glDisable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);
	glColor3f(color[0]*cmag, color[1]*cmag, color[2]*cmag);
	glTranslatef(screenPos[0], screenPos[1], 0.f);
	glBegin(GL_QUADS);
	glTexCoord2i(0,0);
	glVertex3f(-rmag, rmag,0.f);	// Bottom Left
	glTexCoord2i(1,0);
	glVertex3f( rmag, rmag,0.f);	// Bottom Right
	glTexCoord2i(1,1);
	glVertex3f( rmag,-rmag,0.f);	// Top Right
	glTexCoord2i(0,1);
	glVertex3f(-rmag,-rmag,0.f);	// Top Left
	glEnd();

	prj->reset_perspective_projection();		// Restore the other coordinate
}
