/*
 * Stellarium360 astronomy simulation and visualization
 *
 * Copyright (C) 2002 Fabien Chereau
 * Copyright (C) 2009-2011 Digitalis Education Solutions, Inc.
 * Copyright (C) 2013 of the LSS team
 * Copyright (C) 2014_2015 Association Sirius & LSS team 
 * Stellarium360 astronomy simulation and visualization
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Stellarium360 is a free open project of of LSS team
 * See the TRADEMARKS file for free open project usage requirements.
 *
 */

#include <string>
#include <iostream>
#include <cstdlib>

#include "SDL2/SDL.h"

#include "stellarium360.hpp"
#include "app.hpp"
#include "utility.hpp"
#include "app_settings.hpp"
#include "sdl_facade.hpp"
#include "signals.hpp"
#include <dirent.h>
#include "system.hpp"
#include <sys/stat.h>
#include "checkkeys.hpp"

#define __main__
#include "log.hpp"

#ifdef LINUX
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <sys/utsname.h>
#endif
       
#ifdef WIN32
#include <Shlobj.h>
#endif

using namespace std;

//~ string CDIR;		// Config Directory
//~ string LDIR;		// Locale dir for PO translation files
//~ string DATA_ROOT;	// Data Root Directory

void drawIntro(void)
{
	Log.mark();
	Log.write("Welcome to stellarium360",  cLog::eLOG_TYPE_INFO);
	string strMsg="Version : " + string(APP_NAME) + " - " + string(EDITION) + " Edition";
	Log.write(strMsg, cLog::eLOG_TYPE_INFO);
	strMsg.clear();
	strMsg +=" Copyright (c) 2015 Association Sirius, LSS team et al.\n";
	strMsg +=" Copyright (c) 2012-2014 LSS team et al.\n";
	strMsg +=" Copyright (c) 2003-2011 Digitalis Education Solutions, Inc. et al.\n";
	strMsg +=" Copyright (c) 2000-2008 Fabien Chereau et al.\n";
	strMsg +=" http://lss-planetariums.info";
	Log.write(strMsg, cLog::eLOG_TYPE_INFO);
	Log.mark();
}

void getUnameInfo()
{
#ifdef LINUX
	//get uname's info
	struct utsname buffer;
   errno = 0;
   if (uname(&buffer) != 0) {
      switch (errno) {
         case EFAULT:
            Log.write("call to uname() with an invalid pointer", cLog::eLOG_TYPE_ERROR);
            break;
         default:
			Log.write("error calling uname()", cLog::eLOG_TYPE_ERROR);
      }
      exit(EXIT_FAILURE);
   }
	string strMsg="system name = "+string(buffer.sysname);
	strMsg+="\nnode name   = "+string(buffer.nodename);
    strMsg+="\nrelease     = "+string(buffer.release);
    strMsg+="\nversion     = "+string(buffer.version);
    strMsg+="\nmachine     = "+string(buffer.machine);
	Log.write(strMsg, cLog::eLOG_TYPE_INFO);
	Log.mark();
#endif
}

// Display usage in the console
void usage(char **argv)
{
	cout << APP_NAME << endl;
	cout << _("Usage: %s [OPTION] ...\n -v, --version          Output version information and exit.\n -h, --help             Display this help and exit.\n");
}

// Check command line arguments
void check_command_line(int argc, char **argv)
{
	if (argc == 2) {
		if (!(strcmp(argv[1],"--version") && strcmp(argv[1],"-v"))) {
			cout << APP_NAME << endl;
			exit(0);
		}
		if (!(strcmp(argv[1],"--help") && strcmp(argv[1],"-h"))) {
			usage(argv);
			exit(0);
		}
	}

	if (argc > 1) {
		cout << APP_NAME << endl;
		cout << _("%s: Bad command line argument(s)\n")<< endl;
		cout << _("Try `%s --help' for more information.\n");
		exit(1);
	}
}

#ifdef LINUX
void create_lock_file(string lock_file)
{
	ofstream file(lock_file.c_str(), ios::out);
	file << getpid() << endl;
	file.close();
}

bool is_lock_file(string lock_file)
{
	ifstream file(lock_file.c_str(), ios::in);
    if (!file ) {
        return false;
    }
    else {
			string ch;
			file >> ch;
			file.close();
			Log.write("Saved getpid is : "+ ch, cLog::eLOG_TYPE_INFO);
			
			string test="kill -0 "+ch +"\n";
			//printf("%s\n",test.c_str());
			
			if (system(test.c_str())!=0)
				return false; //lock_file exist & it's an old session
            else
                return true;  //lock_file exist & pid is the same
    }
    return false; //for compiler
}
#endif

void checkUserDirectory(const string CDIR, string & dirResult)
{
	if (dir_exist( CDIR ) ) {
		dirResult = "Check home directory ok\n";
	}
	else {
		if ( mkdir(CDIR.c_str(), S_IRWXU | S_IRWXG) == 0)  {
			dirResult = "Creating home directory succesfull\n";
		}
		else {
			fprintf(stderr, "Unable to create local home directory\nProgram stopped\n");
			exit(2);
		}
	}
}

void checkUserSubDirectory(const string CDIR, string& dirResult)
{
		const int NBR_REP=9;
		string rep[NBR_REP]={"vframes", "videos", "audio", "ftp", "fonts", "internal", "scripts", "fscripts", "log" };
		string DATADIR=string(CONFIG_DATA_DIR)+"/data/";
		
		for (int i=0;i<NBR_REP; i++) 
		{
			string subDir= CDIR + rep[i];
			if ( ! dir_exist( subDir.c_str()) ) {
				if ( mkdir( subDir.c_str(), S_IRWXU | S_IRWXG) == 0)  {
					dirResult += "Creating home subdirectory "+rep[i]+" succesfull\n";
					//todo copier le contenu du répertoire en question sur /usr/local/share/...
					
					printf("cp  %s%s %s -r\n", DATADIR.c_str(), rep[i].c_str(), CDIR.c_str());
					if (system( (string("cp ") + DATADIR + rep[i]+ " " + CDIR + " -r").c_str() )) {
						fprintf(stderr, "I can't copy %s in %s directory\n",rep[i].c_str() ,CDIR.c_str() );
						exit(-1); 
					}
					else
					dirResult += "Copy " +rep[i] +" in "+ CDIR + " ok\n";
				}
				else {
					string err = "Unable to create local home subdirectory "+ rep[i] +"\nProgram stopped\n";
					fprintf(stderr, "%s\n",err.c_str() );
					exit(3);
				}
			}
			else
				dirResult +="Check " + rep[i] + " subdirectory ok\n";
		}
}

void checkIniFiles(const string CDIR, const string DATA_ROOT)
{
	string DATADIR=string(CONFIG_DATA_DIR)+"/data/";
	//config.ini
	if ( ! file_exist(CDIR+"/config.ini")) {
		printf("a\n");
		file_copy(DATADIR +"default_config.ini", CDIR +"config.ini");
		Log.write("No config.ini file, i will try to copy default_config.ini", cLog::eLOG_TYPE_INFO);
	}

	//allocate.ini
	if ( ! file_exist(CDIR+"/allocate.ini")) {
		printf("a\n");
		file_copy(DATADIR +"default_allocate.ini", CDIR +"allocate.ini");
		Log.write("No allocate.ini file, i will try to copy default_allocate.ini", cLog::eLOG_TYPE_INFO);
	}
	
	//landscapes.ini
	if ( ! file_exist(CDIR+"/landscapes.ini")) {
		printf("a\n");
		file_copy(DATADIR +"default_landscapes.ini", CDIR +"landscapes.ini");
		Log.write("No landscapes.ini file, i will try to copy default_landscapes.ini", cLog::eLOG_TYPE_INFO);
	}

	//ssystem.ini
	if ( ! file_exist(CDIR+"/ssystem.ini")) {
		file_copy(DATADIR +"default_ssystem.ini", CDIR +"ssystem.ini");
		Log.write("No ssystem.ini file, i will try to copy default_ssystem.ini", cLog::eLOG_TYPE_INFO);
	}
	
	//stars.ini
	if ( ! file_exist(CDIR+"/stars.ini")) {
		file_copy(DATADIR +"default_stars.ini", CDIR +"stars.ini");
		Log.write("No stars.ini file, i will try to copy default_stars.ini", cLog::eLOG_TYPE_INFO);
	}
}

// Main procedure
int main(int argc, char **argv)
{
	string homeDir = getenv("HOME");
	string CDIR = homeDir + "/." + APP_LOWER_NAME+"/";
	string DATA_ROOT = string(CONFIG_DATA_DIR);
	
	string dirResult;

	// Check the command line
	check_command_line(argc, argv);

	//check if home Directory exist and if not try to create it.
	checkUserDirectory(CDIR, dirResult);
	checkUserSubDirectory(CDIR, dirResult);
	
	// Write the console logo & Uname Information...
	drawIntro();
	getUnameInfo();
	
	// write in logfile the result of repertory checking
	Log.write(dirResult, cLog::eLOG_TYPE_INFO);
	dirResult.clear();
	checkIniFiles(CDIR, DATA_ROOT);

	
#ifdef LINUX
	string lock_file = PATH_FILE_LOCK;
	Log.write("Lock file is "+ lock_file,cLog::eLOG_TYPE_INFO);
	Log.write("My getpid() is "+ Utility::doubleToString(getpid()), cLog::eLOG_TYPE_INFO);

	if (is_lock_file(lock_file)) {
		Log.write("There already has an instance of the running program. New instance aborded", cLog::eLOG_TYPE_WARNING);
		return 0;
	}
	create_lock_file(lock_file);
#endif

	// Used for getting system date formatting
	setlocale(LC_TIME, "");

	//save the environnement variable 
	Log.write("CONFIG DIR: " + string(CDIR), cLog::eLOG_TYPE_INFO);
	Log.write("ROOT   DIR: " + string(DATA_ROOT), cLog::eLOG_TYPE_INFO);
	Log.write("LOCALE DIR: " + string(LOCALEDIR), cLog::eLOG_TYPE_INFO);

	
	//test if config.ini is not to old.
	checkConfigIni(CDIR+"config.ini", string(VERSION));
	
	AppSettings::Init(CDIR, DATA_ROOT, LOCALEDIR);
	InitParser conf;
	AppSettings* ini = AppSettings::Instance();
	ini->loadAppSettings( &conf );

	SDLFacade* sdl = new SDLFacade();
	sdl->initSDL();

	Uint16 curW, curH;
	bool fullscreen;
	int antialiasing;

	curW = conf.get_int("video:screen_w");
	curH = conf.get_int("video:screen_h");
	fullscreen = conf.get_boolean("video:fullscreen");
	antialiasing = conf.get_int("rendering:antialiasing");
	
	sdl->createWindow(curW, curH, conf.get_int("video:bbp_mode"), antialiasing, fullscreen, DATA_ROOT + "/data/icon.bmp");

	App* app = new App( sdl );

	// Register custom suspend and term signal handers
	ISignals* signalObj = ISignals::Create(app);
	signalObj->Register( SIGTSTP, ISignals::NSSigTSTP );
	signalObj->Register( SIGTERM, ISignals::NSSigTERM );
	signalObj->Register( SIGINT, ISignals::NSSigTERM );
	signalObj->Register( SIGQUIT, ISignals::NSSigTERM );

	app->init();

	app->startMainLoop();

#ifdef LINUX
	if( unlink(lock_file.c_str()) != 0 )
		Log.write("Error deleting file.lock",  cLog::eLOG_TYPE_ERROR);
	else
		Log.write("File file.lock successfully deleted",  cLog::eLOG_TYPE_INFO);
#endif

	delete app;
	delete sdl;
	delete signalObj;

	Log.write("EOF", cLog::eLOG_TYPE_INFO);
	Log.write("EOF", cLog::eLOG_TYPE_SCRIPT_INFO);
	return 0;
}

