/*
 * Stellarium360 astronomy simulation and visualization
 *
 * Copyright (C) 2002 Fabien Chereau
 * Copyright (C) 2009 Digitalis Education Solutions, Inc.
 * Copyright (C) 2014 of the LSS Team & Association Sirius
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Stellarium360 is a free open project of of LSS team
 * See the TRADEMARKS file for free open project usage requirements.
 *
 */

#include <iostream>
#include "nebula.hpp"
#include "s_texture.hpp"
#include "s_font.hpp"
#include "navigator.hpp"
#include "utility.hpp"
#include "log.hpp"

s_texture * Nebula::tex_GALXY = NULL;
s_texture * Nebula::tex_OPNCL = NULL;
s_texture * Nebula::tex_GLOCL = NULL;
s_texture * Nebula::tex_BRTNB = NULL;
s_texture * Nebula::tex_PLNNB = NULL;
s_texture * Nebula::tex_DRKNB = NULL;
s_texture * Nebula::tex_CLNEB = NULL;
s_texture * Nebula::tex_STARS = NULL;
s_texture * Nebula::tex_GALCL = NULL;
s_texture * Nebula::tex_QUASR = NULL;
s_texture * Nebula::tex_SNREM = NULL;
s_texture * Nebula::tex_GENRC = NULL;

s_font* Nebula::nebulaFont = NULL;

float Nebula::circleScale = 1.f;
float Nebula::hintsBrightness = 0;
float Nebula::nebulaBrightness = 1;
Vec3f Nebula::labelColor = Vec3f(0.,0.,0.);
Vec3f Nebula::circleColor = Vec3f(0.,0.,0.);
bool Nebula::flagBright = false;
bool Nebula::displaySpecificHint = false;
float Nebula::dsoRadius = 1.f;
int Nebula::dsoPictoSize = 6;

Nebula::Nebula(string _englishName, string _mtype, string _constellation, float _ra, float _de, float _mag, float _size, string _classe,
			float _distance, string tex_name, bool path, float tex_angular_size, float tex_rotation, string tex_credit, float _luminance, bool _deletable, bool _hidden) :
		neb_tex(NULL)
{
	neb_color=Vec3f(0.2,0.2,1.);
	tex_circle = NULL;
	englishName = _englishName;
	nameI18 = englishName;
	myDistance = _distance;
	DSOclass = _classe;
	DSOstringType = _mtype;
	m_deletable = _deletable;
	constellation= _constellation;
	m_hidden = _hidden;
	m_selected = true;
	texLuminanceAdjust= _luminance;
	if (tex_credit  == "none")
		credit = "";
	else
		credit = string("Credit: ") + tex_credit;
	for (string::size_type i=0; i<credit.length(); ++i) {
		if (credit[i]=='_') credit[i]=' ';
	}

	// - keep base info for drawing (in radians)
	myRA = _ra*M_PI/180.;
	myDe = _de*M_PI/180.;
	texAngularSize = tex_angular_size/60*M_PI/180;
	myRotation = tex_rotation*M_PI/180;

	// Calc the Cartesian coord with RA and DE
	sphe_to_rect(myRA,myDe,XYZ);
	XYZ*=dsoRadius;
	
	mag= _mag;
	if ((mag!=-30) && (mag < 1.0)) mag = 99.0;

	if (tex_angular_size < 0.0)
		tex_angular_size = 1.0;
	if (tex_angular_size > 150.0)
		tex_angular_size = 150.0;

	// Calc the angular size in radian
	m_angular_size = tex_angular_size/2/60*M_PI/180;
	
	neb_tex = new s_texture(path, tex_name, TEX_LOAD_TYPE_PNG_ALPHA, true);  // use mipmaps

	luminance = mag_to_luminance(mag, tex_angular_size*tex_angular_size*3600);
	//~ if (luminance < 0)
		//~ luminance = .0075;

	// this is a huge performance drag if called every frame, so cache here
	tex_avg_luminance = neb_tex->get_average_luminance();

	float tex_size = dsoRadius * sin(tex_angular_size/2/60*M_PI/180);
	
	// Precomputation of the rotation/translation matrix
	Mat4f mat_precomp = Mat4f::translation(XYZ) *
	                    Mat4f::zrotation(myRA) *
	                    Mat4f::yrotation(-myDe) *
	                    Mat4f::xrotation(tex_rotation*M_PI/180.);

	tex_quad_vertex[0] = mat_precomp * Vec3f(0.,-tex_size,-tex_size);
	tex_quad_vertex[1] = mat_precomp * Vec3f(0., tex_size,-tex_size);
	tex_quad_vertex[2] = mat_precomp * Vec3f(0.,-tex_size, tex_size);
	tex_quad_vertex[3] = mat_precomp * Vec3f(0., tex_size, tex_size);
	
	//what sort of circle should we draw ?
	DSOType = getDsoType(_mtype);

	switch (DSOType) {
	case GALXY:
		neb_color= Vec3f(0.3,0.3,1.);
		tex_circle = tex_GALXY;
		break;
	case OPNCL:
		neb_color= Vec3f(0.,1.0,1.0);
		tex_circle = tex_OPNCL;
		break;
	case GLOCL:
		neb_color= Vec3f(1.,1.0,0.0);
		tex_circle = tex_GLOCL;
		break;
	case BRTNB:
		neb_color= Vec3f(1.,0.2,0.2);
		tex_circle = tex_BRTNB;
		break;
	case PLNNB:
		neb_color= Vec3f(0.0,1.0,0.0);
		tex_circle = tex_PLNNB;
		break;
	case DRKNB:
		neb_color= Vec3f(0.4,0.4,0.4);
		tex_circle = tex_DRKNB;
		break;
	case CLNEB:
		neb_color= Vec3f(1.0,0.5,0.5);
		tex_circle = tex_CLNEB;
		break;
	case STARS:
		neb_color= Vec3f(1.0,1.0,0.5);
		tex_circle = tex_STARS;
		break;
	case GALCL:
		neb_color= Vec3f(0.5,0.5,1.0);
		tex_circle = tex_GALCL;
		break;
	case QUASR:
		neb_color= Vec3f(0.5,0.,0.8);
		tex_circle = tex_QUASR;
		break;
	case SNREM:
		neb_color= Vec3f(0.8,0.,0.5);
		tex_circle = tex_SNREM;
		break;
	default: //case GENRC
		neb_color = getCircleColor();
		tex_circle = tex_GENRC;
		break;
	}
}

Nebula::~Nebula()
{
	delete neb_tex;
	neb_tex = NULL;
}

nebula_type Nebula::getDsoType( string type)
{
	if (type == "GALXY")
		return GALXY;
	else
	if (type == "OPNCL")
		return OPNCL;
	else
	if (type == "GLOCL")
		return GLOCL;
	else
	if (type == "BRTNB")
		return BRTNB;
	else
	if (type == "PLNNB")
		return PLNNB;
	else
	if (type == "DRKNB")
		return DRKNB;
	else
	if (type == "CLNEB")
		return CLNEB;
	else
	if (type == "STARS")
		return STARS;
	else
	if (type == "GALCL")
		return GALCL;
	else
	if (type == "QUASR")
		return QUASR;
	else
	if (type == "SNREM")
		return SNREM;
	else
		return GENRC;
}

string Nebula::getInfoString(const Navigator* nav) const
{
	float tempDE, tempRA;

	Vec3d equPos = nav->j2000_to_earth_equ(XYZ);
	rect_to_sphe(&tempRA,&tempDE,equPos);

	ostringstream oss;
	
	if (nameI18!="") {
		oss << nameI18 << endl;
	}
	else
		oss << englishName << endl;

	oss.setf(ios::fixed);
	oss.precision(2);

	if(mag < 99) oss << _("Magnitude: ") << mag << endl;

	oss << _("RA/DE: ") << Utility::printAngleHMS(tempRA) << " / " << Utility::printAngleDMS(tempDE) << endl;

	// calculate alt az
	Vec3d localPos = nav->earth_equ_to_local(equPos);
	rect_to_sphe(&tempRA,&tempDE,localPos);
	tempRA = 3*M_PI - tempRA;  // N is zero, E is 90 degrees
	if (tempRA > M_PI*2) tempRA -= M_PI*2;

	oss << _("Alt/Az: ") << Utility::printAngleDMS(tempDE) << " / " << Utility::printAngleDMS(tempRA) << endl;

	oss << _("Type: ") << getTypeToString() << endl;

	oss << _("Size: ") << Utility::printAngleDMS(m_angular_size*M_PI/180.) << endl;

	return oss.str();
}

string Nebula::getShortInfoString(const Navigator*) const
{
	ostringstream oss;
	
	if (nameI18 != "")
		oss << nameI18 << "  ";
	else
		oss << englishName << "  ";
	
	if (mag < 99) oss << _("Magnitude: ") << mag;

	oss << " " <<  _("Type: ") << getTypeToString() << " " ;

	if( myDistance > 0 ) {
		string units = _("ly");
		double distance = myDistance;
	
		if(distance >= 1000) {
			distance /= 1000;
			if(distance < 1000) units = _("kly");
			else {
				distance /= 1000;
				units = _("Mly");
			}
		}
		oss.precision(4);
		oss << "  " <<  _("Distance: ") << distance << " " << _(units);
	}

	return oss.str();
}


string Nebula::getShortInfoNavString(const Navigator*) const
{
	return "";
}

double Nebula::get_close_fov(const Navigator*) const
{
	return m_angular_size * 180./M_PI * 4;
}

void Nebula::drawTex(const Projector* prj, const Navigator* nav, ToneReproductor* eye, double sky_brightness)
{
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	if (!neb_tex || m_hidden || !m_selected) return;

	// if start zooming in, turn up brightness to full for DSO images
	// gradual change might be better

	// daylight hackery
	float ad_lum=eye->adapt_luminance(luminance);

	float color = 1;
	if (flagBright && sky_brightness < 0.011 && (get_on_screen_size(prj, nav) > prj->getViewportHeight()/64.)){
		//cout << "Bright nebula drawn for" << getEnglishName() << endl;
		color *= nebulaBrightness;
	} else {
		// TODO this should be revisited to be less ad hoc
		// 3 is a fudge factor since only about 1/3 of a texture is not black background
		float cmag = 3 * ad_lum / tex_avg_luminance * texLuminanceAdjust;
		color = color * cmag * nebulaBrightness;
		//cout << "No bright nebula drawn for" << getEnglishName() << endl;
	}

	glColor4f(1, 1, 1, color);

	glBindTexture(GL_TEXTURE_2D, neb_tex->getID());


	/* DIGITALIS added distortion correction 20080312

	Vec3d v;

	glBegin(GL_TRIANGLE_STRIP);
	    glTexCoord2i(1,0);              // Bottom Right
		prj->project_j2000(tex_quad_vertex[0],v); glVertex3dv(v);
	    glTexCoord2i(0,0);              // Bottom Left
		prj->project_j2000(tex_quad_vertex[1],v); glVertex3dv(v);
	    glTexCoord2i(1,1);              // Top Right
		prj->project_j2000(tex_quad_vertex[2],v); glVertex3dv(v);
	    glTexCoord2i(0,1);              // Top Left
		prj->project_j2000(tex_quad_vertex[3],v); glVertex3dv(v);
		glEnd();
	*/

	Vec3d gridpt, onscreen;

	Vec3d imagev = Mat4d::zrotation(myRA-M_PI_2) * Mat4d::xrotation(myDe) * Vec3d(0,1,0);
	Vec3d ortho1 = Mat4d::zrotation(myRA-M_PI_2) * Vec3d(1,0,0);
	Vec3d ortho2 = imagev^ortho1;

	//	  cout << RA << " " << DE << " " << texAngularSize << " " << myRotation << endl;

	float image_scale = texAngularSize;

	int grid_size = int(image_scale/5.);  // divisions per row, column
	if (grid_size < 5) grid_size = 5;

	for (int i=0; i<grid_size; i++) {

		glBegin(GL_QUAD_STRIP);

		for (int j=0; j<=grid_size; j++) {

			for (int k=0; k<=1; k++) {

				// image height is maximum angular dimension
				gridpt = Mat4d::rotation( imagev, myRotation+M_PI) *
				         Mat4d::rotation( ortho1, image_scale*(j-grid_size/2.)/(float)grid_size) *
				         Mat4d::rotation( ortho2, image_scale*(i+k-grid_size/2.)/(float)grid_size) *
				         imagev;

				if ( prj->project_j2000(gridpt, onscreen)) {

					//		  cout << "on " << onscreen[0] << " " << onscreen[1] << " : "
					//   << (i+k)/(float)grid_size << " " << j/(float)grid_size << endl;
					glTexCoord2f((i+k)/(float)grid_size,j/(float)grid_size);

					glVertex3d(onscreen[0], onscreen[1], 0);

				}
			}

		}
		glEnd();

	}

}

void Nebula::drawHint(const Projector* prj, const Navigator * nav)
{
	if (tex_circle == NULL) {
		//printf("Texture error from %s\n", englishName.c_str());
		Log.write("Texture error from "+englishName , cLog::eLOG_TYPE_ERROR);
		return; }

	if (m_hidden || !m_selected) return;

	if (2.f/get_on_screen_size(prj, nav)<0.1) return;

	float r = dsoPictoSize; 

	if (displaySpecificHint) {
		glBindTexture (GL_TEXTURE_2D, Nebula::tex_circle->getID());
		glColor4f(neb_color[0], neb_color[1], neb_color[2], hintsBrightness);
	}
	else {
		glBindTexture (GL_TEXTURE_2D, Nebula::tex_GENRC->getID());
		glColor4f(circleColor[0], circleColor[1], circleColor[2] , hintsBrightness);
	}

	glBegin(GL_TRIANGLE_STRIP);
	glTexCoord2i(1,0);              // Bottom Right
	glVertex3f(XY[0] + r, XY[1] - r, 0.0f);
	glTexCoord2i(0,0);              // Bottom Left
	glVertex3f(XY[0] - r, XY[1] - r, 0.0f);
	glTexCoord2i(1,1);              // Top Right
	glVertex3f(XY[0] + r, XY[1] + r,0.0f);
	glTexCoord2i(0,1);              // Top Left
	glVertex3f(XY[0] - r, XY[1] + r,0.0f);
	glEnd ();
}

void Nebula::drawName(const Projector* prj)
{
	if (m_hidden || !m_selected) return;

	glColor4f(labelColor[0], labelColor[1], labelColor[2], hintsBrightness);
	float size = get_on_screen_size(prj);
	float shift = 8.f + size/2.f;

	string nebulaname = getNameI18n();

	if (prj->getFlagGravityLabels())
		prj->print_gravity180(nebulaFont, XY[0], XY[1], nebulaname, 1, shift, shift);
	else
		nebulaFont->print(XY[0]+shift, XY[1]+shift,nebulaname,1,1);

	// draw image credit, if it fits easily
	if (credit != "" && size > nebulaFont->getStrLen(credit, 1)) {
		if (prj->getFlagGravityLabels())
			prj->print_gravity180(nebulaFont, XY[0]-shift-40, XY[1]+-shift-40, credit, 1, 0, 0);
		else
			nebulaFont->print(XY[0]-shift, XY[1]-shift-60, credit, 1, 1);
	}
}


string Nebula::getTypeToString(void) const
{
	switch (DSOType) {
	case GALXY:
		return _("Galaxy");
		break;
	case OPNCL:
		return _("Open Cluster");
		break;
	case GLOCL:
		return _("Globular Cluster");
		break;
	case DRKNB:
		return _("Dark Nebula");
		break;
	case PLNNB:
		return _("Planetary Nebula");
		break;
	case BRTNB:
		return _("Bright Nebula");
		break;
	case CLNEB:
		return _("Cluster + Nebula");
		break;
	case STARS:
		return _("Stars");
		break;
	case GALCL:
		return _("Galaxies Cluster");
		break;
	case QUASR:
		return _("Quasar");
		break;
	case SNREM:
		return _("Supernova Remnant");
		break;
	default:
		return _("Unknown");
	}
	return "";
}

void Nebula::translateName(Translator& trans) {
	nameI18 = trans.translateUTF8(englishName);
}

