/*
 * Stellarium360 astronomy simulation and visualization
 *
 * Copyright (C) 2002 Fabien Chereau
 * Copyright (C) 2009-2010 Digitalis Education Solutions, Inc.
 * Copyright (C) 2014 of the LSS Team & Association Sirius
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Stellarium360 is a free open project of of LSS team
 * See the TRADEMARKS file for free open project usage requirements.
 *
 */

#include <iostream>
#include <iomanip>

#include "night_shader.hpp"
#include "bump_shader.hpp"
#include "ring_shader.hpp"
#include "ringed_shader.hpp"
#include "entity.hpp"
#include "navigator.hpp"
#include "projector.hpp"
#include "s_font.hpp"
//#include "s_gui.hpp"
#include "stellastro.h" // just for get_apparent_sidereal_time
#include "log.hpp"


Ring::Ring(double radius_min,double radius_max,const string &texname, const string &path)
		:radius_min(radius_min),radius_max(radius_max)
{
	tex = new s_texture((path!=""), path + texname,TEX_LOAD_TYPE_PNG_ALPHA,1);
	if(/* !Entity::flagShaders || */ !RingShader::instance()->isSupported()) {
		Log.write("Unable to use ring shader", cLog::eLOG_TYPE_ERROR);
		shaderEnable = false;
	} else {
		shaderEnable = true;
		programObject.attachShader(RingShader::instance());
	}
}

Ring::~Ring(void)
{
	if (tex) delete tex;
	tex = NULL;
}

void Ring::draw(const Projector* prj,const Mat4d& mat,double screen_sz, Vec3f& lightDirection, Vec3f& planetPosition, float planetRadius)
{

	screen_sz -= 50;
	screen_sz /= 350.0;

	if (screen_sz < 0.0) screen_sz = 0.0;
	else if (screen_sz > 1.0) screen_sz = 1.0;

	// perf testing
	const int slices = 64+(int)((256-64)*screen_sz);
	const int stacks = 4+(int)((16)*screen_sz);

	// Normal transparency mode
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	//glRotatef(axis_rotation + 180.,0.,0.,1.);

	glColor3f(1.0f, 1.0f, 1.0f); // For saturn only..  //TDO qu'est ce que c'est que ca ?
	//glColor3f(1.0f, 0.88f, 0.82f); // For saturn only..

	glEnable(GL_TEXTURE_2D);
	glEnable(GL_CULL_FACE);
	glEnable(GL_BLEND);

	glBindTexture (GL_TEXTURE_2D, tex->getID());

	// solve the ring wraparound by culling:
	// decide if we are above or below the ring plane
	const double h = mat.r[ 8]*mat.r[12]
	                 + mat.r[ 9]*mat.r[13]
	                 + mat.r[10]*mat.r[14];

	if (shaderEnable) {
		RingShader::instance()->setLightDirection(lightDirection);
		RingShader::instance()->setPlanetPosition(planetPosition);
		RingShader::instance()->setTexMap(tex->getID());
		RingShader::instance()->setPlanetRadius(planetRadius);
		RingShader::instance()->setSunnySideUp(h>0.0);
	}

	if(shaderEnable) programObject.enable();
	prj->sRing(radius_min,radius_max,(h<0.0)?slices:-slices,stacks, mat, 0, shaderEnable);
	if(shaderEnable) programObject.disable();

	glDisable(GL_CULL_FACE);
}
