#ifndef SHADER_H_INCLUDED
#define SHADER_H_INCLUDED

//#include "stellarium360.hpp"
#include <GL/glew.h>
#include <GL/glu.h>
#include <string>

/**
 * @file    shader.h shader.c
 * @author  Jérôme Lartillot modified by by olivier Nivoix
 * @version 1.0
 *
 * @section REQUIREMENTS
 * -stdlib.h
 * -stdio.h
 * -the GLEW library
 * -using namespace std;
 *
 * @section DESCRIPTION
 *
 *  This allow to manage shaders easily.
 *
 *
    Example:
    --------
    Program prog;
    prog.init(  makeShader("bump.vert",GL_VERTEX_SHADER),
                makeShader("bump.frag",GL_FRAGMENT_SHADER));

    prog.use();


    glUniform1i(glGetUniformLocation(prog.program,"texunit0"),0);
    glUniform1i(glGetUniformLocation(prog.program,"texunit1"),1);

    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D,texture);
    glActiveTexture(GL_TEXTURE1);
    glBindTexture(GL_TEXTURE_2D,ntexture);

    //... some drawings...

    glUseProgram(0);
 *
 */



/**
*   \class Program
*   Creates a program for the pipeline from a vertex shader and a fragment shader.
*   You must initialise this with Program(GLuint vs,GLuint fs)
*   or init(GLuint vs,GLuint fs) before calling use().
*
*   Use Program::program to set uniforms variables.
*/
class shaderProgram{
public:    
    shaderProgram(){}
    shaderProgram(GLuint vs,GLuint fs);
    ~shaderProgram();
    
    void init(GLuint vs,GLuint fs);

/**
*   \fn setShaderDir
*   set the Directory path to all shader source
*   \param shader: the directory path
*/
	static void setShaderDir(const std::string& _shaderDir) {
		shaderProgram::shaderDir = _shaderDir;
	}
	
	/**
*   \fn makeShader Creates shader from a file.
*   This function call debugShader.
*   \param str:        must contain a filename whichcontain the shader code or the shader code.
*   \param ShaderType: cf glCreateShader: GL_COMPUTE_SHADER, GL_VERTEX_SHADER,
*                      GL_TESS_CONTROL_SHADER, GL_TESS_EVALUATION_SHADER, GL_GEOMETRY_SHADER,
*                      or GL_FRAGMENT_SHADER
*   \return Return the shader id as GLuint. You must assign it (or be stupid).
*/   
    GLuint makeShader(std::string str, GLenum ShaderType); 
    
    void use();
    GLuint program;
    /**
*   \fn debugShader
*   Print compilation errors of a shader in cout(if they exists).
*   Example: debugShader(myShaderId);
*   \param shader: the shader to debug.
*/
    void debugShader(GLuint shader);
/**
*   \fn printInformations
*   Print informations about uniforms,
*   (ID, name, type)
*/    
    void printInformations(GLuint program);
private:
    GLuint vshader;
    GLuint fshader;
    static  std::string shaderDir;
    void LoadSource(const char *filename,GLint*numberOfLines,GLchar***code,GLint**linesLengths);
    
 /**
 *  \fn makeShader    Creates a shader from a source code in a two dimensions array.
 *  This may help: https://www.opengl.org/sdk/docs/man/html/glShaderSource.xhtml
 *  This function call debugShader.
 *  \param ShaderType: cf glCreateShader: GL_COMPUTE_SHADER, GL_VERTEX_SHADER,
 *                     GL_TESS_CONTROL_SHADER, GL_TESS_EVALUATION_SHADER, GL_GEOMETRY_SHADER,
 *                     or GL_FRAGMENT_SHADER
 *  \param numberOfLines: of the shader source code.
 *  \param code: the source code of the shader.
 *  \param linesLengths: explicit parameter.
 *  \return Return the shader id as GLuint. You must assign it (or be stupid).
 */
    GLuint makeShader(GLenum ShaderType, GLint numberOfLines, const GLchar**code,const GLint*linesLengths);
};

#endif // SHADER_H_INCLUDED
