/*
 * Stellarium360 astronomy simulation and visualization
 *
 * Copyright (C) 2002 Fabien Chereau
 * Copyright (C) 2009 Digitalis Education Solutions, Inc.
 * Copyright (C) 2014 of the LSS Team & Association Sirius
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Stellarium360 is a free open project of of LSS team
 * See the TRADEMARKS file for free open project usage requirements.
 *
 */

#ifndef _SOLARSYSTEM_H_
#define _SOLARSYSTEM_H_

#include <vector>
#include <functional>

#include "entity.hpp"
#include "stellarium360.hpp"
#include "object.hpp"
#include "orbit.hpp"
#include "translator.hpp"

#define SOLAR_MASS 1.989e30
#define EARTH_MASS 5.976e24
#define LUNAR_MASS 7.354e22
#define MARS_MASS  0.64185e24

using namespace std;

typedef std::map< std::string, Entity * > entityHash_t;
typedef entityHash_t::const_iterator entityHashIter_t;


class SolarSystem
{
public:
	SolarSystem();
	virtual ~SolarSystem();

	void update(int delta_time, Navigator* nav);

	//! Draw all the elements of the solar system
	//! \return get the maximum squared distance in pixels that any planet has travelled since the last update.
	double draw(Projector * du, const Navigator * nav,
	            const ToneReproductor* eye,
	            bool flag_point,
	            bool drawHomePlanet );

	//! Load the bodies data from a file
	void load(const string& planetfile);

	// load one object from a hash (returns error message if any)
	// this public method always adds bodies as deletable
	string addBody(stringHash_t & param) { return addBody(param, true); }

	// remove an object (returns error message if any)
	// only allows removal of non-ini file objects
	string removeBody(string name);

	//! remove all bodies added after ssystem.ini file was read
	string removeSupplementalBodies(const Entity *home_planet);

	//! @brief Update i18 names from english names according to passed translator
	//! The translation is done using gettext with translated strings defined in translations.h
	void translateNames(Translator& trans);

	string getPlanetHashString();  // locale and ssystem.ini names, newline delimiter, for tui

	void setFont(float font_size, const string& font_name);
	
	void setPlanetLabelColor(const string englishName, const Vec3f& c);
	
	const Vec3f getPlanetLabelColor(const string englishName) const;

	void setPlanetOrbitColor(const string englishName, const Vec3f& c);
	
	const Vec3f getPlanetOrbitColor(const string englishName) const;

	void setDefaultPlanetOrbitColor(const Vec3f& c) {
		defaultOrbitColor=c;
	}

	const Vec3f getDefaultPlanetOrbitColor() {
		return defaultOrbitColor;
	}

	void setDefaultPlanetTrailColor(const Vec3f& c) {
		defaultTrailColor=c;
	}

	const Vec3f getDefaultPlanetTrailColor() {
		return defaultTrailColor;
	}

	void setDefaultPlanetLabelColor(const Vec3f& c) {
		defaultLabelColor=c;
	}

	const Vec3f getDefaultPlanetLabelColor() {
		return defaultLabelColor;
	}

	//! Compute the position for every elements of the solar system.
	//! home_planet is needed for light travel time computation
	void computePositions(double date,const Entity *home_planet);

	//! Compute the transformation matrix for every elements of the solar system.
	//! home_planet is needed for light travel time computation
	void computeTransMatrices(double date,const Entity *home_planet);

	//! Search if any Planet is close to position given in earth equatorial position.
	Object search(Vec3d, const Navigator * nav, const Projector * prj) const;

	//! Return a stl vector containing the planets located inside the lim_fov circle around position v
	vector<Object> search_around(Vec3d v,
	                             double lim_fov,
	                             const Navigator * nav,
	                             const Projector * prj,
	                             bool *default_last_item,
	                             bool aboveHomePlanet ) const;

	//! Return the matching planet pointer if exists or NULL
	Entity* searchByEnglishName(string planetEnglishName) const;

	//! Return the matching planet pointer if exists or NULL
	//! @param planetNameI18n The case sensistive translated planet name
	Object searchByNamesI18(string planetNameI18n) const;

	
	//! get the position Alt Az for Sun
	void getSunTraceAltAz(const Navigator *nav, double *alt, double *az) const;
		
	//!return Earth planet object
	Entity* getEarth(void) const {
		return earth;
	}
	
	//! return Sun planet object
	Entity* getSun(void) const {
		return sun;
	}
	
	//! return Moon planet object
	Entity* getMoon(void) const {
		return moon;
	}

	//! set flag for Activate/Deactivate planets axis
	void setFlagAxis(bool b);

	//! get flag for Activate/Deactivate planets axis
	bool getFlagAxis(void) const{
		return flagAxis;
	}; 

	//! set flag for Activate/Deactivate planets display
	void setFlagPlanets(bool b) {
		Entity::setFlagShow(b);
	}

	//! get flag for Activate/Deactivate planets display
	bool getFlagPlanets(void) const {
		return Entity::getFlagShow();
	}

	//! set flag for Activate/Deactivate planets trails display
	void setFlagTrails(bool b);
	
	//! get flag for Activate/Deactivate planets trails display	
	bool getFlagTrails(void) const {
		return flagTrails;
	}
	
	//! set flag for Activate/Deactivate planets hints display
	void setFlagHints(bool b);
	
	//! get flag for Activate/Deactivate planets hints display
	bool getFlagHints(void) const {
		return flagHints;
	}

	//! Activate/Deactivate planet&&satellites orbits display
	void setFlagOrbits(bool b) {
		setFlagPlanetsOrbits(b);
		setFlagSatellitesOrbits(b);
	}
	
	bool getFlagOrbits(void) const {
		return (flagPlanetsOrbits||flagSatellitesOrbits);
	}
	
	//! Set flag for Activate/Deactivate planets orbits display
	void setFlagPlanetsOrbits(bool b);
	
	//! Get flag for Activate/Deactivate planets orbits display	
	bool getFlagPlanetsOrbits(void) const {
		return flagPlanetsOrbits;
	}
	
	//! Set flag for Activate/Deactivate satellites orbits display	
	void setFlagSatellitesOrbits(bool b);
	
	//! getflag for Activate/Deactivate satellites orbits display	
	bool getFlagSatellitesOrbits(void) const {
		return flagSatellitesOrbits;
	}
	
	void setFlagLightTravelTime(bool b) {
		flag_light_travel_time = b;
	}
	
	bool getFlagLightTravelTime(void) const {
		return flag_light_travel_time;
	}

	//! Start/stop accumulating new trail data (clear old data)
	void startTrails(bool b);

	void setPlanetTrailColor(const string englishName, const Vec3f& c) ;

	const Vec3f getPlanetTrailColor(const string englishName) const;

	void updateTrails(const Navigator* nav);

	//! Set base planets display scaling factor
	void setScale(float scale) {
		Entity::setScale(scale);
	}
	
	//! Get base planets display scaling factor
	float getScale(void) const {
		return Entity::getScale();
	}

	//! Set the scale factor s of the planet name 
	//! @param name the planet's name
	void setPlanetSizeScale(string name, float s);

	//! Get the scale factor of the planet name 
	float getPlanetSizeScale(string name);

	//! Set base planets display limit in pixels
	void setSizeLimit(float scale) {
		Entity::setSizeLimit(scale);
	}
	
	//! Get base planets display limit in pixels	
	float getSizeLimit(void) const {
		return Entity::getSizeLimit();
	}


	//! Set if Moon display is scaled
	void setFlagMoonScale(bool b) {
		if (!b) getMoon()->set_sphere_scale(1);
		else getMoon()->set_sphere_scale(moonScale);
		flagMoonScale = b;
    }

	//! Get if Moon display is scaled
    bool getFlagMoonScale(void) const {
        return flagMoonScale;
    }
    
  	//! Set if Sun display is scaled
    void setFlagSunScale(bool b) {
        if (!b) {
            getSun()->set_sphere_scale(1);
            getSun()->set_halo_size(200);
        }
        else {
            getSun()->set_sphere_scale(SunScale);
            getSun()->set_halo_size(SunScale*200);
        }

        flagSunScale = b;
    }

  	//! Get if Sun display is scaled
    bool getFlagSunScale(void) const {
        return flagSunScale;
    }
    
    //! Set Moon display scaling factor
    void setMoonScale(float f) {
        moonScale = f;
        if (flagMoonScale) getMoon()->set_sphere_scale(moonScale);
    }
    
    //! Get Moon display scaling factor
    float getMoonScale(void) const {
        return moonScale;
    }

    //! Set Sun display scaling factor
    void setSunScale(float f) {
        SunScale = f;
        if (flagSunScale) getSun()->set_sphere_scale(SunScale);
    }
    
    //! Get Sun display scaling factor    
    float getSunScale(void) const {
        return SunScale;
    }
    //! Set flag for displaying clouds (planet rendering feature)
    void setFlagClouds(bool b) {
        Entity::setFlagClouds(b);
    }
    //! Get flag for displaying Atmosphere
    bool getFlagClouds(void) const {
        return Entity::getFlagClouds();
    }

    //! Get list of all the translated planets name
    vector<string> getNamesI18(void);

    //! Find and return the list of at most maxNbItem objects auto-completing the passed object I18n name
    vector<string> listMatchingObjectsI18n(const string& objPrefix, unsigned int maxNbItem) const;

    //! Set selected planet by english name or "" to select none
    void setSelected(const string& englishName) {
        setSelected(searchByEnglishName(englishName));
    }

    //! Set selected object from its pointer
    void setSelected(const Object &obj);

    //! Get selected object's pointer
    Object getSelected(void) const {
        return selected;
    }

    //~ void setFlagShaders(bool b) {
        //~ Entity::flagShaders = b;
    //~ }

	void iniModel3DObj(model3D_mgr * _myObjMgr) {
		myObjMgr = _myObjMgr;
	}
	
	//modify Planet "name" is hidden or not.
	void setPlanetHidden(string name, bool planethidden);

	//return the default label, orbit and trail color for default planet
	void iniColor(Vec3f label_color, Vec3f orbit_color, Vec3f trail_color);

	//reinitialise l'ensemble des planetes comme elles étaient au chargement initial du logiciel
	// prend en compte la taille et le flag caché ou pas
	void initialSolarSystemBodies();

private:
	
	// determine the planet type: Sun, planet, moon, dwarf, asteroid ...
	ENTITY_TYPE setPlanetType (string str);

    // load one object from a hash (returns error message if any)
    string addBody(stringHash_t & param, bool deletable);

    Entity* sun;	//return the Sun
    Entity* moon;	//return the Moon
    Entity* earth;	//return the earth

    //! The currently selected planet
    Object selected;

    // solar system related settings
    float object_scale;  // should be kept synchronized with star scale...

    bool flagMoonScale;	// say if float moonScale is used or not
    bool flagSunScale;	// say if float SunScale is used or not
    float moonScale;	// Moon scale value
    float SunScale;	// Sun scale value

    s_font* planet_name_font;
    vector<Entity*> system_planets;		// Vector containing all the bodies of the system
    bool near_lunar_eclipse(const Navigator * nav, Projector * prj);

    // draw earth shadow on moon for lunar eclipses
  //  void draw_earth_shadow(const Navigator * nav, Projector * prj);  //INUTILE

    // And sort them from the furthest to the closest to the observer
    struct bigger_distance : public binary_function<Entity*, Entity*, bool> {
        bool operator()(Entity* p1, Entity* p2) {
            return p1->get_distance() > p2->get_distance();
        }
    };

  //  s_texture *tex_earth_shadow;  // for lunar eclipses NX_INUTILE utilisée dansdraw_earth_shadow

    // Master settings
    bool flagPlanetsOrbits;
    bool flagSatellitesOrbits;
    bool flag_light_travel_time;
    bool flagHints;
    bool flagTrails;
    bool flagAxis;

	// Master color setting
	Vec3f defaultOrbitColor;
	Vec3f defaultLabelColor;
	Vec3f defaultTrailColor;

    entityHash_t entityHash;  // holds list of current entity english names
    // (with parent pointer) to enforce uniqueness

    model3D_mgr * myObjMgr;
};


#endif // _SOLARSYSTEM_H_
