/*
 * Stellarium360 astronomy simulation and visualization
 *
 * Copyright (C) 2002 Fabien Chereau
 * Copyright (C) 2009-2010 Digitalis Education Solutions, Inc.
 * Copyright (C) 2013 of the LSS team
 * Copyright (C) 2014 of the LSS Team & Association Sirius
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Stellarium360 is a free open project of the LSS team
 * See the TRADEMARKS file for free open project usage requirements.
 *
 */

// Class which handles the User Interface

#include <iostream>
#include <iomanip>
#include <algorithm>
#include "app.hpp"
#include "ui.hpp"
#include "stellastro.h"
#include "log.hpp"

////////////////////////////////////////////////////////////////////////////////
//								CLASS FUNCTIONS
////////////////////////////////////////////////////////////////////////////////

string default_landscape = "";
string current_landscape = "";
UI::UI(Core * _core, App * _app) :
		baseFont(NULL),
		courierFont(NULL),
		tuiFont(NULL),

		FlagHelp(false), FlagInfos(false), FlagConfig(false), FlagSearch(false), FlagShowTuiMenu(0),

		top_bar_ctr(NULL),
		top_bar_date_lbl(NULL),
		top_bar_hour_lbl(NULL),
		top_bar_fps_lbl(NULL),
		top_bar_appName_lbl(NULL),
		top_bar_fov_lbl(NULL),

		bt_flag_ctr(NULL),
		bt_flag_constellation_draw(NULL),
		bt_flag_constellation_name(NULL),
		bt_flag_constellation_art(NULL),
		bt_flag_azimuth_grid(NULL),
		bt_flag_equator_grid(NULL),
		bt_flag_ground(NULL),
		bt_flag_cardinals(NULL),
		bt_flag_atmosphere(NULL),
		bt_flag_nebula_name(NULL),
		bt_flag_help(NULL),
		bt_flag_equatorial_mode(NULL),
		bt_flag_config(NULL),
//		bt_flag_chart(NULL),
		bt_flag_planet(NULL),
		bt_flag_search(NULL),
		bt_script(NULL),
		bt_flag_goto(NULL),
		bt_flip_horz(NULL),
		bt_flip_vert(NULL),
		bt_flag_help_lbl(NULL),
		info_select_ctr(NULL),
		info_select_txtlbl(NULL),
		licence_win(NULL),
		licence_txtlbl(NULL),

		help_win(NULL),
		help_txtlbl(NULL),

		config_win(NULL),
		search_win(NULL),
		dialog_win(NULL),
		tui_root(NULL),

		key_Modifier(NONE) ,
		KeyTimeLeft(0)
{
	if (!_core) {
		Log.write("UI.CPP CRITICAL : In stel_ui constructor, unvalid core.",cLog::eLOG_TYPE_ERROR);
		exit(-1);
	}
	core = _core;
	app = _app;
	is_dragging = false;
	waitOnLocation = true;
	opaqueGUI = true;
	initialised = false;
}

/**********************************************************************************/
UI::~UI()
{
	delete desktop;
	desktop = NULL;
	delete baseFont;
	baseFont = NULL;
	delete baseTex;
	baseTex = NULL;
	delete flipBaseTex;
	flipBaseTex = NULL;
	delete courierFont;
	courierFont = NULL;
	delete tuiFont;
	tuiFont = NULL;
	delete tex_up;
	tex_up = NULL;
	delete tex_down;
	tex_down = NULL;
	if (tui_root) delete tui_root;
	tui_root=NULL;
	Component::deleteScissor();
}

////////////////////////////////////////////////////////////////////////////////
void UI::init(const InitParser& conf)
{

	if (initialised) {

		// delete existing objects before recreating
		if (baseFont) delete baseFont;
		if (courierFont) delete courierFont;
		if (baseTex) delete baseTex;
		if (flipBaseTex) delete flipBaseTex;
		if (tex_up) delete tex_up;
		if (tex_down) delete tex_down;
		if (desktop) delete desktop;
	}

	// Ui section
	FlagShowFps			= conf.get_boolean("gui:flag_show_fps");
	FlagShowLatLon      = conf.get_boolean("gui:flag_show_latlon");
	FlagMenu			= conf.get_boolean("gui:flag_menu");
	FlagHelp			= conf.get_boolean("gui:flag_help");
	FlagInfos			= conf.get_boolean("gui:flag_infos");
	FlagShowTopBar		= conf.get_boolean("gui:flag_show_topbar");
	FlagShowTime		= conf.get_boolean("gui:flag_show_time");
	FlagShowDate		= conf.get_boolean("gui:flag_show_date");
	FlagShowAppName		= conf.get_boolean("gui:flag_show_appname");
	FlagShowFov			= conf.get_boolean("gui:flag_show_fov");
	FlagNumberPrint		= conf.get_int("gui:flag_number_print");
	FlagShowSelectedObjectInfo = conf.get_boolean("gui:flag_show_selected_object_info");
	FontSizeGeneral		= conf.get_double ("font","font_general_size");
	FontNameGeneral        = core->getUserFontDir() +conf.get_str("font", "font_general_name");//, "DejaVuSans.ttf");  
	FlagShowScriptBar	= conf.get_boolean("gui","flag_show_script_bar",false);
	MouseCursorTimeout  = conf.get_double("gui","mouse_cursor_timeout",0);
	PosDateTime			= conf.get_int("gui","datetime_display_position");
	PosMenuM			= conf.get_int("gui","menu_display_position",120);
	FlagShowPlanetname	= conf.get_boolean("gui:flag_show_planetname");
	// Text ui section
	FontSizeMenuTui     = conf.get_double ("font","font_menutui_size");
	FlagEnableTuiMenu = conf.get_boolean("tui:flag_enable_tui_menu");
	FlagShowGravityUi = conf.get_boolean("tui:flag_show_gravity_ui");
	FlagShowTuiDateTime = conf.get_boolean("tui:flag_show_tui_datetime");
	FlagShowTuiShortObjInfo = conf.get_boolean("tui:flag_show_tui_short_obj_info");
	FlagMouseUsableInScript = conf.get_boolean("gui","flag_mouse_usable_in_script",false);
	//~ ReferenceState state;
	//~ state.show_tui_date_time = FlagShowTuiDateTime;
	//~ state.show_tui_short_obj_info = FlagShowTuiShortObjInfo;
	//~ SharedData::Instance()->References( state );

	//FontNameGeneral = core->getDataDir() + FontNameGeneral;

	// TODO: can we get rid of this second font requirement?
	FontSizeMenuGui = conf.get_double ("font","font_menugui_size");
	FontNameMenu = core->getUserFontDir()+conf.get_str("font", "font_menu_name"); //, "DejaVuSansMono.ttf");

	// Load standard font
	baseFont = new s_font(FontSizeGeneral, FontNameGeneral);
	if (!baseFont) {
		Log.write("Error while creating base font",cLog::eLOG_TYPE_ERROR);
		exit(-1);
	}

	courierFont = new s_font(FontSizeMenuGui, FontNameMenu);
	if (!courierFont) {
		Log.write("Error while creating courier font",cLog::eLOG_TYPE_ERROR);
		exit(-1);
	}

	// set up mouse cursor timeout
	MouseTimeLeft = MouseCursorTimeout*1000;

	// Create standard texture
	baseTex = new s_texture("backmenu.png", TEX_LOAD_TYPE_PNG_ALPHA);
	flipBaseTex = new s_texture("backmenu_flip.png", TEX_LOAD_TYPE_PNG_ALPHA);

	tex_up = new s_texture("up.png");
	tex_down = new s_texture("down.png");

	// Set default Painter
	Painter p(baseTex, baseFont, s_color(0.5, 0.5, 0.5), s_color(1., 1., 1.));
	Component::setDefaultPainter(p);

	Component::initScissor(core->getDisplayWidth(), core->getDisplayHeight());

	desktop = new s_gui::Container(true);
	desktop->reshape(0,0,core->getDisplayWidth(),core->getDisplayHeight());

	bt_flag_help_lbl = new Label("ERROR...");
	bt_flag_help_lbl->setPos(3,core->getDisplayHeight()-41-(int)baseFont->getDescent());
	bt_flag_help_lbl->setVisible(0);

	bt_flag_time_control_lbl = new Label("ERROR...");
	bt_flag_time_control_lbl->setPos(core->getDisplayWidth()-210,core->getDisplayHeight()-41-(int)baseFont->getDescent());
	bt_flag_time_control_lbl->setVisible(0);

	// Info on selected object
	info_select_ctr = new s_gui::Container();
	info_select_ctr->reshape(0,15,300,200);
	info_select_txtlbl = new TextLabel();
	info_select_txtlbl->reshape(5,5,550,202);
	info_select_ctr->setVisible(1);
	info_select_ctr->addComponent(info_select_txtlbl);
	info_select_ctr->setGUIColorSchemeMember(false);
	desktop->addComponent(info_select_ctr);

	// message window
	message_txtlbl = new TextLabel();
	message_txtlbl->adjustSize();
	message_txtlbl->setPos(10,10);
	message_win = new StdTransBtWin("Message", 5000);
	//message_win->setOpaque(opaqueGUI);
	message_win->reshape(core->getDisplayWidth()/2 -350, core->getDisplayHeight()/2 -100, 700, 200);
	message_win->addComponent(message_txtlbl);
	message_win->setVisible(false);
	desktop->addComponent(message_win);

	desktop->addComponent(createTopBar());
	desktop->addComponent(createFlagButtons(conf));
	desktop->addComponent(createTimeControlButtons());
	desktop->addComponent(bt_flag_help_lbl);
	desktop->addComponent(bt_flag_time_control_lbl);

	dialog_win = new StdDlgWin(APP_NAME);
	//dialog_win->setOpaque(opaqueGUI);
	dialog_win->setDialogCallback(callback<void>(this, &UI::dialogCallback));
	desktop->addComponent(dialog_win);

	desktop->addComponent(createLicenceWindow());
	desktop->addComponent(createHelpWindow());
	desktop->addComponent(createConfigWindow());
	desktop->addComponent(createSearchWindow());

	initialised = true;

	setTitleObservatoryName(getTitleWithAltitude());
    default_landscape = core->getObservatory()->get_landscape_name();
    current_landscape = core->getObservatory()->get_landscape_name();
	Log.write("Landscape : "+default_landscape ,cLog::eLOG_TYPE_INFO);
    get_a_tab_parser();
}


////////////////////////////////////////////////////////////////////////////////
void UI::show_message(string _message, int _time_out)
{
	// draws a message window to display a message to user
	// if timeout is zero, won't time out
	// otherwise use miliseconds

	// TODO figure out how to size better for varying message lengths

	message_txtlbl->setLabel(_message);
	message_txtlbl->adjustSize();
	message_win->set_timeout(_time_out);
	message_win->setVisible(1);
}

////////////////////////////////////////////////////////////////////////////////
Component* UI::createTopBar(void)
{

	top_bar_date_lbl = new Label("-", baseFont);
	top_bar_date_lbl->setPos(2,1);
	top_bar_hour_lbl = new Label("-", baseFont);
	top_bar_hour_lbl->setPos(110,1);
	top_bar_fps_lbl = new Label("-", baseFont);
	top_bar_fps_lbl->setPos(core->getDisplayWidth()-100,1);
	top_bar_fov_lbl = new Label("-", baseFont);
	top_bar_fov_lbl->setPos(core->getDisplayWidth()-220,1);
	top_bar_appName_lbl = new Label(APP_NAME, baseFont);
	top_bar_appName_lbl->setPos(core->getDisplayWidth()/2-top_bar_appName_lbl->getSizex()/2,1);
	top_bar_ctr = new FilledContainer();
	top_bar_ctr->reshape(0,0,core->getDisplayWidth(),(int)(baseFont->getLineHeight()+0.5)+5);
	top_bar_ctr->addComponent(top_bar_date_lbl);
	top_bar_ctr->addComponent(top_bar_hour_lbl);
	top_bar_ctr->addComponent(top_bar_fps_lbl);
	top_bar_ctr->addComponent(top_bar_fov_lbl);
	top_bar_ctr->addComponent(top_bar_appName_lbl);
	return top_bar_ctr;
}

////////////////////////////////////////////////////////////////////////////////
void UI::updateTopBar(void)
{
	top_bar_ctr->setVisible(FlagShowTopBar);
	if (!FlagShowTopBar) return;

	double jd = core->getJDay();

	if (FlagShowDate) {
		top_bar_date_lbl->setLabel(app->get_printable_date_local(jd));
		top_bar_date_lbl->adjustSize();
	}
	top_bar_date_lbl->setVisible(FlagShowDate);

	if (FlagShowTime) {
		top_bar_hour_lbl->setLabel(app->get_printable_time_local(jd));
		top_bar_hour_lbl->adjustSize();
	}
	top_bar_hour_lbl->setVisible(FlagShowTime);

	top_bar_appName_lbl->setVisible(FlagShowAppName);

	if (FlagShowFov) {
		stringstream wos;
		wos << "FOV=" << setprecision(3) << core->getFov() << "\u00B0";
		top_bar_fov_lbl->setLabel(wos.str());
		top_bar_fov_lbl->adjustSize();
	}
	top_bar_fov_lbl->setVisible(FlagShowFov);

	if (FlagShowFps) {
		stringstream wos;
		wos << "FPS=" << int(app->fps + 0.5);
		top_bar_fps_lbl->setLabel(wos.str());
		top_bar_fps_lbl->adjustSize();
	}
	top_bar_fps_lbl->setVisible(FlagShowFps);
}

// Create the button panel in the lower left corner
#define UI_PADDING 5
#define UI_BT 25
#define UI_SCRIPT_BAR 400
Component* UI::createFlagButtons(const InitParser &conf)
{
	int x = 0;

	bt_flag_constellation_draw = new FlagButton(false, NULL, "bt_constellations.png");
	bt_flag_constellation_draw->setOnPressCallback(callback<void>(this, &UI::cb));
	bt_flag_constellation_draw->setOnMouseInOutCallback(callback<void>(this, &UI::cbr));

	bt_flag_constellation_name = new FlagButton(false, NULL, "bt_const_names.png");
	bt_flag_constellation_name->setOnPressCallback(callback<void>(this, &UI::cb));
	bt_flag_constellation_name->setOnMouseInOutCallback(callback<void>(this, &UI::cbr));

	bt_flag_constellation_art = new FlagButton(false, NULL, "bt_constart.png");
	bt_flag_constellation_art->setOnPressCallback(callback<void>(this, &UI::cb));
	bt_flag_constellation_art->setOnMouseInOutCallback(callback<void>(this, &UI::cbr));

	bt_flag_azimuth_grid = new FlagButton(false, NULL, "bt_azgrid.png");
	bt_flag_azimuth_grid->setOnPressCallback(callback<void>(this, &UI::cb));
	bt_flag_azimuth_grid->setOnMouseInOutCallback(callback<void>(this, &UI::cbr));

	bt_flag_equator_grid = new FlagButton(false, NULL, "bt_eqgrid.png");
	bt_flag_equator_grid->setOnPressCallback(callback<void>(this, &UI::cb));
	bt_flag_equator_grid->setOnMouseInOutCallback(callback<void>(this, &UI::cbr));

	bt_flag_ground = new FlagButton(false, NULL, "bt_ground.png");
	bt_flag_ground->setOnPressCallback(callback<void>(this, &UI::cb));
	bt_flag_ground->setOnMouseInOutCallback(callback<void>(this, &UI::cbr));

	bt_flag_cardinals = new FlagButton(false, NULL, "bt_cardinal.png");
	bt_flag_cardinals->setOnPressCallback(callback<void>(this, &UI::cb));
	bt_flag_cardinals->setOnMouseInOutCallback(callback<void>(this, &UI::cbr));

	bt_flag_atmosphere = new FlagButton(false, NULL, "bt_atmosphere.png");
	bt_flag_atmosphere->setOnPressCallback(callback<void>(this, &UI::cb));
	bt_flag_atmosphere->setOnMouseInOutCallback(callback<void>(this, &UI::cbr));

	bt_flag_nebula_name = new FlagButton(false, NULL, "bt_nebula.png");
	bt_flag_nebula_name->setOnPressCallback(callback<void>(this, &UI::cb));
	bt_flag_nebula_name->setOnMouseInOutCallback(callback<void>(this, &UI::cbr));

	bt_flag_help = new FlagButton(false, NULL, "bt_help.png");
	bt_flag_help->setOnPressCallback(callback<void>(this, &UI::cb));
	bt_flag_help->setOnMouseInOutCallback(callback<void>(this, &UI::cbr));

	bt_flag_equatorial_mode = new FlagButton(false, NULL, "bt_follow.png");
	bt_flag_equatorial_mode->setOnPressCallback(callback<void>(this, &UI::cb));
	bt_flag_equatorial_mode->setOnMouseInOutCallback(callback<void>(this, &UI::cbr));

	bt_flag_config = new FlagButton(false, NULL, "bt_config.png");
	bt_flag_config->setOnPressCallback(callback<void>(this, &UI::cb));
	bt_flag_config->setOnMouseInOutCallback(callback<void>(this, &UI::cbr));



	bt_flag_planet = new FlagButton(false, NULL, "bt_planet_label.png");
	bt_flag_planet->setOnPressCallback(callback<void>(this, &UI::cb));
	bt_flag_planet->setOnMouseInOutCallback(callback<void>(this, &UI::cbr));

	bt_flag_quit = new FlagButton(true, NULL, "bt_quit.png");
	bt_flag_quit->setOnPressCallback(callback<void>(this, &UI::cb));
	bt_flag_quit->setOnMouseInOutCallback(callback<void>(this, &UI::cbr));

	bt_flag_search = new FlagButton(true, NULL, "bt_search.png");
	bt_flag_search->setOnPressCallback(callback<void>(this, &UI::cb));
	bt_flag_search->setOnMouseInOutCallback(callback<void>(this, &UI::cbr));

	bt_script = new EditBox();
	bt_script->setAutoFocus(false);
	bt_script->setSize(UI_SCRIPT_BAR-1,24);
	bt_script->setOnKeyCallback(callback<void>(this, &UI::cbEditScriptKey));
	bt_script->setOnReturnKeyCallback(callback<void>(this, &UI::cbEditScriptExecute));
	bt_script->setOnMouseInOutCallback(callback<void>(this, &UI::cbr));

	bt_flag_goto = new FlagButton(true, NULL, "bt_goto.png");
	bt_flag_goto->setOnPressCallback(callback<void>(this, &UI::cb));
	bt_flag_goto->setOnMouseInOutCallback(callback<void>(this, &UI::cbr));

	bt_flag_ctr = new FilledContainer();
	bt_flag_ctr->addComponent(bt_flag_constellation_draw);
	bt_flag_constellation_draw->setPos(x,0);
	x+=UI_BT;
	bt_flag_ctr->addComponent(bt_flag_constellation_name);
	bt_flag_constellation_name->setPos(x,0);
	x+=UI_BT;
	bt_flag_ctr->addComponent(bt_flag_constellation_art);
	bt_flag_constellation_art->setPos(x,0);
	x+=UI_BT;
	bt_flag_ctr->addComponent(bt_flag_azimuth_grid);
	bt_flag_azimuth_grid->setPos(x,0);
	x+=UI_BT;
	bt_flag_ctr->addComponent(bt_flag_equator_grid);
	bt_flag_equator_grid->setPos(x,0);
	x+=UI_BT;
	bt_flag_ctr->addComponent(bt_flag_ground);
	bt_flag_ground->setPos(x,0);
	x+=UI_BT;
	bt_flag_ctr->addComponent(bt_flag_cardinals);
	bt_flag_cardinals->setPos(x,0);
	x+=UI_BT;
	bt_flag_ctr->addComponent(bt_flag_atmosphere);
	bt_flag_atmosphere->setPos(x,0);
	x+=UI_BT;
	bt_flag_ctr->addComponent(bt_flag_planet);
	bt_flag_planet->setPos(x,0);
	x+=UI_BT;
	bt_flag_ctr->addComponent(bt_flag_nebula_name);
	bt_flag_nebula_name->setPos(x,0);
	x+=UI_BT;
	bt_flag_ctr->addComponent(bt_flag_equatorial_mode);
	bt_flag_equatorial_mode->setPos(x,0);
	x+=UI_BT;
	bt_flag_ctr->addComponent(bt_flag_goto);
	bt_flag_goto->setPos(x,0);
	x+=UI_BT;
	if (conf.get_boolean("gui","flag_show_flip_buttons",false)) {
		bt_flip_horz = new FlagButton(true, NULL, "bt_flip_horz.png");
		bt_flip_horz->setOnPressCallback(callback<void>(this, &UI::cb));
		bt_flip_horz->setOnMouseInOutCallback(callback<void>(this, &UI::cbr));
		bt_flag_ctr->addComponent(bt_flip_horz);
		bt_flip_horz->setPos(x,0);
		x+=UI_BT;
		bt_flip_vert = new FlagButton(true, NULL, "bt_flip_vert.png");
		bt_flip_vert->setOnPressCallback(callback<void>(this, &UI::cb));
		bt_flip_vert->setOnMouseInOutCallback(callback<void>(this, &UI::cbr));
		bt_flag_ctr->addComponent(bt_flip_vert);
		bt_flip_vert->setPos(x,0);
		x+=UI_BT;
	}

	x+= UI_PADDING;
	bt_flag_ctr->addComponent(bt_script);
	bt_script->setPos(x,0);
	if (!FlagShowScriptBar) {
		bt_script->setVisible(false);
	} else {
		x+=UI_SCRIPT_BAR;
		x+= UI_PADDING;
	}

	bt_flag_ctr->addComponent(bt_flag_search);
	bt_flag_search->setPos(x,0);
	x+=UI_BT;
	bt_flag_ctr->addComponent(bt_flag_config);
	bt_flag_config->setPos(x,0);
	x+=UI_BT;
//	bt_flag_ctr->addComponent(bt_flag_chart);			bt_flag_chart->setPos(x,0); x+=UI_BT;
	bt_flag_ctr->addComponent(bt_flag_help);
	bt_flag_help->setPos(x,0);
	x+=UI_BT;
	bt_flag_ctr->addComponent(bt_flag_quit);
	bt_flag_quit->setPos(x,0);
	x+=UI_BT;

	bt_flag_ctr->setOnMouseInOutCallback(callback<void>(this, &UI::bt_flag_ctrOnMouseInOut));
	bt_flag_ctr->reshape(0, core->getDisplayHeight()-25, x-1, 25);

	return bt_flag_ctr;

}

// Create the button panel in the lower right corner
Component* UI::createTimeControlButtons(void)
{
	bt_dec_time_speed = new FlagButton(false, NULL, "bt_rwd.png");
	bt_dec_time_speed->setOnPressCallback(callback<void>(this, &UI::bt_dec_time_speed_cb));
	bt_dec_time_speed->setOnMouseInOutCallback(callback<void>(this, &UI::tcbr));

	bt_real_time_speed = new FlagButton(false, NULL, "bt_realtime.png");
	bt_real_time_speed->setSize(24,24);
	bt_real_time_speed->setOnPressCallback(callback<void>(this, &UI::bt_real_time_speed_cb));
	bt_real_time_speed->setOnMouseInOutCallback(callback<void>(this, &UI::tcbr));

	bt_inc_time_speed = new FlagButton(false, NULL, "bt_fwd.png");
	bt_inc_time_speed->setOnPressCallback(callback<void>(this, &UI::bt_inc_time_speed_cb));
	bt_inc_time_speed->setOnMouseInOutCallback(callback<void>(this, &UI::tcbr));

	bt_time_now = new FlagButton(false, NULL, "bt_now.png");
	bt_time_now->setOnPressCallback(callback<void>(this, &UI::bt_time_now_cb));
	bt_time_now->setOnMouseInOutCallback(callback<void>(this, &UI::tcbr));

	bt_time_control_ctr = new FilledContainer();
	bt_time_control_ctr->addComponent(bt_dec_time_speed);
	bt_dec_time_speed->setPos(0,0);
	bt_time_control_ctr->addComponent(bt_real_time_speed);
	bt_real_time_speed->setPos(25,0);
	bt_time_control_ctr->addComponent(bt_inc_time_speed);
	bt_inc_time_speed->setPos(50,0);
	bt_time_control_ctr->addComponent(bt_time_now);
	bt_time_now->setPos(75,0);

	bt_time_control_ctr->setOnMouseInOutCallback(callback<void>(this, &UI::bt_time_control_ctrOnMouseInOut));
	bt_time_control_ctr->reshape(core->getDisplayWidth()-4*25-1, core->getDisplayHeight()-25, 4*25, 25);

	return bt_time_control_ctr;
}

void UI::bt_dec_time_speed_cb(void)
{
	app->FlagTimePause = 0;
	double s = core->getTimeSpeed();
	if (s>JD_SECOND) s/=10.;
	else if (s<=-JD_SECOND) s*=10.;
	else if (s>-JD_SECOND && s<=0.) s=-JD_SECOND;
	else if (s>0. && s<=JD_SECOND) s=0.;
	core->setTimeSpeed(s);
}

void UI::bt_inc_time_speed_cb(void)
{
	app->FlagTimePause = 0;
	double s = core->getTimeSpeed();
	if (s>=JD_SECOND) s*=10.;
	else if (s<-JD_SECOND) s/=10.;
	else if (s>=0. && s<JD_SECOND) s=JD_SECOND;
	else if (s>=-JD_SECOND && s<0.) s=0.;
	core->setTimeSpeed(s);
}

void UI::bt_real_time_speed_cb(void)
{
	app->FlagTimePause = 0;
	core->setTimeSpeed(JD_SECOND);
}

void UI::bt_time_now_cb(void)
{
	core->setJDay(NShadeDateTime::JulianFromSys());
}

////////////////////////////////////////////////////////////////////////////////
// Script edit command line

void UI::cbEditScriptInOut(void)
{
	if (bt_script->getIsMouseOver()) {
		bt_flag_help_lbl->setLabel(_("Script commander"));
	}
}

void UI::cbEditScriptKey(void)
{
	if (bt_script->getLastKey() == SDL_SCANCODE_SPACE || bt_script->getLastKey() == SDL_SCANCODE_TAB) {
		string command = bt_script->getText();
		transform(command.begin(), command.end(), command.begin(), ::tolower);
		if (bt_script->getLastKey() == SDL_SCANCODE_SPACE) command = command.substr(0,command.length()-1);
	} else if	(bt_script->getLastKey() == SDL_SCANCODE_ESCAPE) {
		bt_script->clearText();
	}
}

void UI::cbEditScriptExecute(void)
{
	unsigned long int delay;
	string command_string = bt_script->getText();
	Log.write("UI.CPP: Executing command: " + command_string ,cLog::eLOG_TYPE_INFO);
	bt_script->clearText();
	bt_script->setEditing(false);

	if (!app->commander->execute_command(command_string, delay, false))  // Not trusted to enable script path defaults
		bt_flag_help_lbl->setLabel(_("Invalid Script command"));
}

////////////////////////////////////////////////////////////////////////////////
void UI::cb(void)
{
	core->setFlagConstellationLines(bt_flag_constellation_draw->getState());
	core->setFlagConstellationNames(bt_flag_constellation_name->getState());
	core->setFlagConstellationArt(bt_flag_constellation_art->getState());
	core->setFlagAzimutalGrid(bt_flag_azimuth_grid->getState());
	core->setFlagEquatorGrid(bt_flag_equator_grid->getState());
	core->setFlagLandscape(bt_flag_ground->getState());
	core->setFlagCardinalsPoints(bt_flag_cardinals->getState());
	core->setFlagAtmosphere(bt_flag_atmosphere->getState());
	core->setFlagNebulaHints( bt_flag_nebula_name->getState() );
	if (bt_flip_horz) core->setFlipHorz( bt_flip_horz->getState() );
	if (bt_flip_vert) core->setFlipVert( bt_flip_vert->getState() );
	FlagHelp 				= bt_flag_help->getState();
	help_win->setVisible(FlagHelp);
	core->setMountMode(bt_flag_equatorial_mode->getState() ? Core::MOUNT_EQUATORIAL : Core::MOUNT_ALTAZIMUTAL);
	FlagConfig			= bt_flag_config->getState();
	core->setFlagPlanetsHints( bt_flag_planet->getState());
	config_win->setVisible(FlagConfig);

	FlagSearch			= bt_flag_search->getState();
	search_win->setVisible(FlagSearch);
	if (bt_flag_goto->getState()) core->gotoSelectedObject();
	bt_flag_goto->setState(false);

	if (!bt_flag_quit->getState()) app->quit();
}

void UI::bt_flag_ctrOnMouseInOut(void)
{
	if (bt_flag_ctr->getIsMouseOver()) bt_flag_help_lbl->setVisible(1);
	else bt_flag_help_lbl->setVisible(0);
}

void UI::bt_time_control_ctrOnMouseInOut(void)
{
	if (bt_time_control_ctr->getIsMouseOver()) bt_flag_time_control_lbl->setVisible(1);
	else bt_flag_time_control_lbl->setVisible(0);
}

void UI::cbr(void)
{
	if (bt_flag_constellation_draw->getIsMouseOver())
		bt_flag_help_lbl->setLabel(_("Constellation Lines"));
	if (bt_flag_constellation_name->getIsMouseOver())
		bt_flag_help_lbl->setLabel(_("Constellation Labels"));
	if (bt_flag_constellation_art->getIsMouseOver())
		bt_flag_help_lbl->setLabel(_("Constellation Art"));
	if (bt_flag_azimuth_grid->getIsMouseOver())
		bt_flag_help_lbl->setLabel(_("Azimuthal Grid"));
	if (bt_flag_equator_grid->getIsMouseOver())
		bt_flag_help_lbl->setLabel(_("Equatorial Grid"));
	if (bt_flag_ground->getIsMouseOver())
		bt_flag_help_lbl->setLabel(_("Ground"));
	if (bt_flag_cardinals->getIsMouseOver())
		bt_flag_help_lbl->setLabel(_("Cardinal Points"));
	if (bt_flag_atmosphere->getIsMouseOver())
		bt_flag_help_lbl->setLabel(_("Atmosphere"));
	if (bt_flag_nebula_name->getIsMouseOver())
		bt_flag_help_lbl->setLabel(_("Nebula Labels"));
	if (bt_flag_help->getIsMouseOver())
		bt_flag_help_lbl->setLabel(_("Help"));
	if (bt_flag_equatorial_mode->getIsMouseOver())
		bt_flag_help_lbl->setLabel(_("Equatorial/Altazimuthal Mount"));
	if (bt_flag_config->getIsMouseOver())
		bt_flag_help_lbl->setLabel(_("Configuration"));

	if (bt_flag_planet->getIsMouseOver())
		bt_flag_help_lbl->setLabel(_("Body Labels"));
	if (bt_flag_quit->getIsMouseOver())
#ifndef MACOSX
		bt_flag_help_lbl->setLabel(_("Quit [CTRL + Q]"));
#else
		bt_flag_help_lbl->setLabel(_("Quit [CMD + Q]"));
#endif
	if (bt_flag_search->getIsMouseOver())
		bt_flag_help_lbl->setLabel(_("Search for Object"));
	if (bt_script->getIsMouseOver())
		bt_flag_help_lbl->setLabel(_("Script Command"));
	if (bt_flag_goto->getIsMouseOver())
		bt_flag_help_lbl->setLabel(_("Track Selected Object"));
	if (bt_flip_horz && bt_flip_horz->getIsMouseOver())
		bt_flag_help_lbl->setLabel(_("Flip Horizontally"));
	if (bt_flip_vert && bt_flip_vert->getIsMouseOver())
		bt_flag_help_lbl->setLabel(_("Flip Vertically"));
}

void UI::tcbr(void)
{
	if (bt_dec_time_speed->getIsMouseOver())
		bt_flag_time_control_lbl->setLabel(_("Decrease Time Rate"));
	if (bt_real_time_speed->getIsMouseOver())
		bt_flag_time_control_lbl->setLabel(_("Real Time Rate"));
	if (bt_inc_time_speed->getIsMouseOver())
		bt_flag_time_control_lbl->setLabel(_("Increase Time Rate"));
	if (bt_time_now->getIsMouseOver())
		bt_flag_time_control_lbl->setLabel(_("Return to Current Time"));
}

// The window containing the info (licence)
Component* UI::createLicenceWindow(void)
{

// - removed escaped characters
	licence_txtlbl = new TextLabel(
	    string("                 *   " APP_NAME " - " EDITION " Edition    *\n\n") +
	    "*   Copyright 2013 of LSS team et al.\n" +
	    "*   Copyright 2003-2011 Digitalis Education Solutions, Inc. et al.\n" +
	    "*   Copyright 2000-2008 Fabien Chereau et al.\n\n" +
	    "*" + _("   Please check for newer versions and send bug reports\n\
    and comments to us at: http://lss-planetariums.info\n\n") +
	    "*   This program is free software; you can redistribute it and/or\n\
modify it under the terms of the GNU General Public License\n\
as published by the Free Software Foundation; either version 3\n\
of the License, or (at your option) any later version.\n\n" +
	    "This program is distributed in the hope that it will be useful, but\n\
WITHOUT ANY WARRANTY; without even the implied\n\
warranty of MERCHANTABILITY or FITNESS FOR A\n\
PARTICULAR PURPOSE.  See the GNU General Public\n\
License for more details.\n\n" +
	    "You should have received a copy of the GNU General Public\n\
License along with this program; if not, write to:\n" +
	    "Free Software Foundation, Inc.\n\
59 Temple Place - Suite 330\n\
Boston, MA  02111-1307, USA.\n\
http://www.fsf.org\n\n" +
	    "* Stellarium360 is a free open project\n   of the LSS team");

	licence_txtlbl->adjustSize();
	licence_txtlbl->setPos(10,10);
	licence_win = new StdBtWin(_("Information"));
	//licence_win->setOpaque(opaqueGUI);
	licence_win->reshape(core->getDisplayWidth()/2 -300, core->getDisplayHeight()/2 -270, 600, 560);
	licence_win->addComponent(licence_txtlbl);
	licence_win->setVisible(FlagInfos);

	return licence_win;
}

Component* UI::createHelpWindow(void)
{

	help_text = new HashBox(30, 0.33);
	help_text->setPos(20,20);
	help_text->setSizex(450);
	help_text->addItemList( string(" \n<b>") + string(_("Movement and Selection:")) + "</b>\n"
							+ _("Arrow Keys") + "\n" + _("Pan View") + "\n"
							+ _("Page Up") + "\n" + _("Zoom In") + "\n"
							+ _("Page Down") + "\n" + _("Zoom Out") + "\n"
							+ _("Left Click") + "\n" + _("Select Object")  + "\n"
							+ _("Right Click") + "\n" + _("Unselect")  + "\n"
							+ _("CTRL + Left Click") + "\n" + _("Unselect")  + "\n"
							+ "CTRL + Q \n" + _("Quit") + "\n"
		);

	help_win = new StdBtWin(_("Help"));
	help_win->reshape(core->getDisplayWidth()/2 -250, core->getDisplayHeight()/2 -300, 500, 600);
	help_win->addComponent(help_text);
	help_win->setVisible(FlagHelp);
	help_win->setOnHideBtCallback(callback<void>(this, &UI::help_win_hideBtCallback));
	return help_win;
}

void UI::help_win_hideBtCallback(void)
{
	help_win->setVisible(0);
}


/*******************************************************************/
void UI::draw(void)
{

	// Lines should be 1 pixel wide only and not use sky lines width
	glLineWidth(1);

	// draw first as windows should cover these up
	// also problem after 2dfullscreen with square viewport
	if (FlagShowGravityUi) draw_gravity_ui();
	if (getFlagShowTuiMenu()) draw_tui();

	// Special cool text transparency mode
	glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);

	glEnable(GL_BLEND);

	app->set2DfullscreenProjection();	// 2D coordinate
	Component::enableScissor();

	glScalef(1, -1, 1);						// invert the y axis, down is positive
	glTranslatef(0, -core->getDisplayHeight(), 0);	// move the origin from the bottom left corner to the upper left corner

	desktop->draw();

	Component::disableScissor();
	app->restoreFrom2DfullscreenProjection();	// Restore the other coordinate

}

/*******************************************************************************/
int UI::handle_move(int x, int y)
{
	// Do not allow use of mouse while script is playing
	// otherwise script can get confused
	if (app->scripts->is_playing() && ! FlagMouseUsableInScript) return 0;

	// Show cursor
	SDL_ShowCursor(1);
	MouseTimeLeft = MouseCursorTimeout*1000;

	if (desktop->onMove(x, y)) return 1;
	if (is_dragging) {
		if ((has_dragged || sqrtf((x-previous_x)*(x-previous_x)+(y-previous_y)*(y-previous_y))>4.)) {
			has_dragged = true;
			core->setFlagTracking(false);
			core->dragView(previous_x, previous_y, x, y);
			previous_x = x;
			previous_y = y;
			return 1;
		}
	}
	return 0;
}

/*******************************************************************************/
int UI::handle_clic(Uint16 x, Uint16 y, S_GUI_VALUE button, S_GUI_VALUE state)
{
	// Do not allow use of mouse while script is playing
	// otherwise script can get confused
	if (app->scripts->is_playing() && ! FlagMouseUsableInScript) return 0;

	// Make sure object pointer is turned on (script may have turned off)
	core->setFlagSelectedObjectPointer(true);

	// Show cursor
	SDL_ShowCursor(1);
	MouseTimeLeft = MouseCursorTimeout*1000;

	if (desktop->onClic((int)x, (int)y, button, state)) {
		has_dragged = false;
		is_dragging = false;
		return 1;
	}

	switch (button) {
	case S_GUI_MOUSE_RIGHT :
		break;
	case S_GUI_MOUSE_LEFT :
		if (state==S_GUI_PRESSED) {
			is_dragging = app->getFlagEnableDragMouse();
			has_dragged = false;
			previous_x = x;
			previous_y = y;
		} else {
			is_dragging = false;
		}
		break;
	case S_GUI_MOUSE_MIDDLE :
		break;
	case S_GUI_MOUSE_WHEELUP :
		core->zoomTo(core->getAimFov()-app->MouseZoom*core->getAimFov()/60., 0.2);
		return 1;
	case S_GUI_MOUSE_WHEELDOWN :
		core->zoomTo(core->getAimFov()+app->MouseZoom*core->getAimFov()/60., 0.2);
		return 1;
	default:
		break;
	}

	// Manage the event for the main window
	{
		//if (state==S_GUI_PRESSED) return 1;
		// Deselect the selected object
		if (button==S_GUI_MOUSE_RIGHT && state==S_GUI_RELEASED) {
			app->commander->execute_command("select");
			return 1;
		}
		if (button==S_GUI_MOUSE_MIDDLE && state==S_GUI_RELEASED) {
			if (core->getFlagHasSelected()) {
				core->gotoSelectedObject();
				core->setFlagTracking(true);
			}
		}
		if (button==S_GUI_MOUSE_LEFT && state==S_GUI_RELEASED && !has_dragged) {
			// CTRL + left clic = right clic for 1 button mouse
			if (SDL_GetModState() & KMOD_CTRL) {
				app->commander->execute_command("select");
				return 1;
			}

			// Try to select object at that position
			core->findAndSelect(x, y);

			// If an object was selected update informations
			if (core->getFlagHasSelected()) updateInfoSelectString();
		}
	}
	return 0;
}


/*******************************************************************************/

// LSS HANDLE KEYS

// TODO replace this with flexible keymapping feature
// odd extension to prevent compilation from makefile but inclusion in make dist
int flag_compass = 0;
int flag_triangle = 0;
int flag_creu = 0;
int flag_f9 = 0;
int flag_esclat = 0;
int flag_esclon = 0;

double esclat[5] = {90,66.55,45,23.45,0};
double esclon[5] = {0,-23.45,-45,-66.55,-90};

int analema_pointM[6] = {2,4,6,9,11,12};
int analema_pointD[6] = {12,15,14,1,3,24};

int season_pointM[4] = {3,6,9,12};
int season_pointD[4] = {20,20,22,21};

int flag_perifeli = 0;
int flag_analema_point = 0;
int flag_season_point = 0;

bool antipodes = false;
bool medium_sun = false;
bool heading_north = true;
bool heading_east = false;



/*
 * For old compatibility :
 * a_stab[i] <--- i° script to launch
 * for(i=0;i<18;i++)  a_stab[i] launch the file  (i+1).sts
 * for(i=18;i<28;i++) a_stab[i] launch the file K(i-18).sts  a_stab[18]-->K0.sts
 * for(i=28;i<50;i++) a_stab[i] launch the file M(i-28).sts  a_stab[28]---> M00.sts
 *
 * a_atab[i] is for audio & a_vtab[i] is for video
 */
int UI::handle_keys(SDL_Scancode key, Uint16 mod, Uint16 unicode, S_GUI_VALUE state)
{
    int retVal = 1;
    std::ostringstream oss;
    double latimem;

    string homeDir = getenv("HOME");
	string CDIR2;

	if (core->getFlagNav())
		CDIR2 = homeDir + "/." + APP_LOWER_NAME + "/navigation/";
	else
		CDIR2 = homeDir + "/." + APP_LOWER_NAME + "/";

    if (desktop->onKey(key, state)) //What does it do ?
        return 1;

//	printf("valeur de la touche appuyée %s\n",SDL_GetScancodeName(key));

    if (state==S_GUI_PRESSED)
    {
        if (key == SDL_SCANCODE_A && (mod & COMPATIBLE_KMOD_CTRL))
        {
            app->quit();
        }

	if(!app->scripts->is_playing())
            app->time_multiplier = 1;  // if no script in progress always real time
			
			switch (key)
			{
			case SDL_SCANCODE_GRAVE : 
									if (key_Modifier != SUPER)
									{ key_Modifier=SUPER;
									  KeyTimeLeft = 3*1000;  } 
									else key_Modifier=NONE;
									break;
			case SDL_SCANCODE_RGUI :
			case SDL_SCANCODE_LGUI : 
									if (key_Modifier != KWIN)
									{ key_Modifier=KWIN;
									KeyTimeLeft = 120*1000;}
									else key_Modifier=NONE;
									break;

			case SDL_SCANCODE_RSHIFT : 
			case SDL_SCANCODE_LSHIFT : 
									if (key_Modifier != SHIFT)
									{ key_Modifier=SHIFT;
									KeyTimeLeft = 120*1000; }
									else key_Modifier=NONE;
									break;
			case SDL_SCANCODE_RCTRL :
			case SDL_SCANCODE_LCTRL :
									key_Modifier=CTRL;
									KeyTimeLeft = 120*1000; // 2 min);
									break;

			case SDL_SCANCODE_BACKSLASH : 
						switch(key_Modifier) {
						case NONE:
									app->scripts->play_script(CDIR2+"internal/clear_mess.sts", CDIR2+"internal/"); 
									break;
						case SUPER:
									app->commander->execute_command( "flag nebula_names toggle");
									key_Modifier= NONE; break;
						case KWIN:		
									break;
						case CTRL: 
									core->setFlagNebulaNames(true);
									break;
						default: 		
									break;
						} 
						break;  

			case SDL_SCANCODE_SEMICOLON : 
						switch(key_Modifier) {
						case NONE:
									if (FlagEnableTuiMenu) setFlagShowTuiMenu(true);
									break;
						case SUPER:
									key_Modifier= NONE; break;
						case KWIN:
									break;
						case CTRL: 
									break;
						default:
									break;
						}
						break;

			case SDL_SCANCODE_COMMA : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "external_viewer filename "+CDIR2+"audio/"+a_atab[1]+" action play");
									break;
						case SUPER: 
									app->commander->execute_command( "external_viewer filename "+CDIR2+"audio/"+a_atab[5]+" action play");
									key_Modifier= NONE; break;
						case KWIN:  		
									break;
						case CTRL: 
									app->commander->execute_command( "external_viewer filename "+CDIR2+"audio/"+a_atab[9]+" action play");
									break;
						default: 		
									break;
						}
						break;

			case SDL_SCANCODE_PERIOD : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "external_viewer filename "+CDIR2+"audio/"+a_atab[2]+" action play");
									break;
						case SUPER: 
									app->commander->execute_command( "external_viewer filename "+CDIR2+"audio/"+a_atab[6]+" action play");
									key_Modifier= NONE; break;
						case KWIN:  		
									break;
						case CTRL: 
									app->commander->execute_command( "external_viewer filename "+CDIR2+"audio/"+a_atab[10]+" action play");
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_SLASH : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "external_viewer filename "+CDIR2+"audio/"+a_atab[3]+" action play");
									break;
						case SUPER: 
									app->commander->execute_command( "external_viewer filename "+CDIR2+"audio/"+a_atab[7]+" action play");
									key_Modifier= NONE; break;
						case KWIN:  		
									break;
						case CTRL: 
									app->commander->execute_command( "external_viewer filename "+CDIR2+"audio/"+a_atab[11]+" action play");
									break;
						default: 	
									break;
						}
						break;


			case  SDL_SCANCODE_A: 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "flag cardinal_points toggle");
									break;
						case SUPER:
									if (flag_creu != 1) {
										app->scripts->play_script(CDIR2+"internal/windrose.sts", CDIR2+"internal/");
									}
									else core->setLandscape(current_landscape);
									flag_creu = (flag_creu+1)%2;
									key_Modifier= NONE; break;
						case KWIN: 
									break;
						case CTRL: 
									app->quit(); break;
						default: 	
									break;
						}
						break;

			case  SDL_SCANCODE_B: 
						switch(key_Modifier) {
						case NONE:
									if (core->getMeteorsRate()==10) app->commander->execute_command("meteors zhr 10000"); 
										else app->commander->execute_command("meteors zhr 10");
									break;
						case SUPER:
									if (core->getMeteorsRate()<=10000) app->commander->execute_command("meteors zhr 150000");
										else app->commander->execute_command("meteors zhr 10");
									key_Modifier= NONE; break;
						case KWIN:	
									break;
						case CTRL:	
									break;
						default: 	
									break;
						}
						break;

			case  SDL_SCANCODE_C: 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "flag equatorial_grid toggle");
									key_Modifier= NONE; break;
						case SUPER:
									app->commander->execute_command( "flag circumpolar_circle toggle");
									key_Modifier= NONE; break;
						case KWIN:	
									if (core->getFlagNav()) app->commander->execute_command( "flag aries_line toggle");						  
									break;
						case SHIFT:	
									if (core->getFlagNav()) app->commander->execute_command( "flag greenwich_line toggle");
									break;
						case CTRL :	
									app->commander->execute_command( "flag vernal_points toggle");
									break;
						default:    
									break;
						}
						break;

			case  SDL_SCANCODE_D: 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "flag equator_line toggle"); break;
						case SUPER:
						            app->commander->execute_command( "flag tropic_lines toggle"); 
						            key_Modifier= NONE; break;
						case KWIN:
									if (app->get_m_OutputFrameNumber() == 0) 
										{  show_message(_("Now recording video frames!\nPress Win + D to stop."), 1000);
										app->set_m_OutputFrameNumber(1);
										}
									else
										{ app->set_m_OutputFrameNumber(0);
										show_message(_("Stopped recording video frames."), 3000);
										}
									 key_Modifier= NONE; break;
						case SHIFT:
									app->commander->execute_command( "flag satellites_orbits toggle");
									break;
						case CTRL :
									app->scripts->play_script(CDIR2+"internal/equator_poles.sts", CDIR2+"internal/");
									break;
						default: 	
									break;
						}
						break;

			case  SDL_SCANCODE_E: 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "flag constellation_art toggle");
									break;
						case SUPER:
					                app->commander->execute_command("deselect");
					                app->commander->execute_command("select constellation Ari pointer off");
					                app->commander->execute_command("select constellation Tau pointer off");
					                app->commander->execute_command("select constellation Gem pointer off");
					                app->commander->execute_command("select constellation Cnc pointer off");
					                app->commander->execute_command("select constellation Leo pointer off");
					                app->commander->execute_command("select constellation Vir pointer off");
					                app->commander->execute_command("select constellation Lib pointer off");
					                app->commander->execute_command("select constellation Sco pointer off");
					                app->commander->execute_command("select constellation Sgr pointer off");
					                app->commander->execute_command("select constellation Cap pointer off");
					                app->commander->execute_command("select constellation Aqr pointer off");
					                app->commander->execute_command("select constellation Psc pointer off");
					                key_Modifier= NONE; break;
						case KWIN:	
									break;
						case CTRL:	
									app->scripts->play_script(CDIR2+"internal/sky_culture3.sts", CDIR2+"internal/");
									break;
						default: 	
									break;
						}
						break;


			case SDL_SCANCODE_F  : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "flag moon_scaled toggle");
									//app->commander->execute_command( "flag sun_scaled toggle");
									break;
						case SUPER:
									app->commander->execute_command( "flag sun_scaled toggle");
									app->scripts->play_script(CDIR2+"internal/big_planets.sts", CDIR2+"internal/");
									key_Modifier= NONE; break;
						case SHIFT:
									app->scripts->play_script(CDIR2+"internal/bodies-asteroids-501ex.sts", CDIR2+"internal/");
									key_Modifier= NONE; break;
						case KWIN:
									FlagSearch = !FlagSearch;
									search_win->setVisible(FlagSearch);
									break;
						case CTRL :
									app->scripts->play_script(CDIR2+"internal/comet.sts", CDIR2+"internal/");
									break;
						default: 	
									break;
						}
						break;
									

			case SDL_SCANCODE_G : 
						if ( app->scripts->is_playing() )
							app->commander->execute_command( "script action end");
						else
							app->commander->execute_command( "timerate rate 0");
						break;

			case SDL_SCANCODE_H :
						switch(key_Modifier) {
						case NONE:
									if ( app->scripts->is_playing() ) {
										app->commander->execute_command( "script action pause");
										app->time_multiplier = 1;}
									else
										app->commander->execute_command( "timerate action pause");
									break;
						case SUPER:	
									key_Modifier= NONE; break;
						case KWIN:	
									break;
						case CTRL :
									app->scripts->play_script(CDIR2+"internal/personal.sts", CDIR2+"internal/");
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_I : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "date sidereal -1");
									break;
						case SUPER: 
									key_Modifier= NONE; break;
						case SHIFT:
									{double jd = core->getJDay();
									ln_date current_date;
									NShadeDateTime::JulianToDate(jd,&current_date);
									core->setJDay(NShadeDateTime::JulianDayFromDateTime(current_date.years,current_date.months-1,current_date.days,current_date.hours,current_date.minutes,current_date.seconds));}
									break;
						case KWIN:
									FlagHelp=!FlagHelp;
									help_win->setVisible(FlagHelp);
									break;
						case CTRL :
									app->commander->execute_command( "date relative -1");
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_J :
						switch(key_Modifier) {
						case NONE:
									if(app->scripts->is_playing())
									app->scripts->slower_script();
									//~ {app->commander->execute_command( "script action slower");
									//~ printf("time --\n");}
									else
									app->commander->execute_command( "timerate action decrement");
									break;
						case SUPER:	
									app->scripts->play_script(CDIR2+"internal/proper_demotion.sts", CDIR2+"internal/");
									key_Modifier= NONE; break;
						case SHIFT:		
									app->commander->execute_command( "date sun rise");
									app->scripts->play_script(CDIR2+"internal/date_shift_minus.sts", CDIR2+"internal/");
									break;
						case KWIN:		
									break;
						case CTRL:
									{double jd = core->getJDay();
									ln_date current_date;
									NShadeDateTime::JulianToDate(jd,&current_date);
									core->setJDay(NShadeDateTime::JulianDayFromDateTime(current_date.years-20,current_date.months,current_date.days,current_date.hours,current_date.minutes,current_date.seconds));}
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_K :
						switch(key_Modifier) {
						case NONE:
									if ( app->scripts->is_playing() ) {
										app->commander->execute_command( "script action resume");
										app->time_multiplier = 1; }
									else
										app->commander->execute_command( "timerate rate 1"); 
									break;
						case SHIFT:	
									app->commander->execute_command( "date sun meridian"); 
						default: 	
									break;
						}
						break;
			
			case SDL_SCANCODE_L :
						switch(key_Modifier) {
						case NONE:
									if(app->scripts->is_playing())
									  app->scripts->faster_script();
									else
									  app->commander->execute_command( "timerate action increment");
									break;
						case SUPER:	
									app->scripts->play_script(CDIR2+"internal/proper_motion.sts", CDIR2+"internal/");
									key_Modifier= NONE; break;
						case KWIN:		
									break;
						case SHIFT:		
									app->commander->execute_command( "date sun set"); 
									app->scripts->play_script(CDIR2+"internal/date_shift_plus.sts", CDIR2+"internal/");
									break;
						case CTRL :
									{double jd = core->getJDay();
									ln_date current_date;
									NShadeDateTime::JulianToDate(jd,&current_date);
									core->setJDay(NShadeDateTime::JulianDayFromDateTime(current_date.years+20,current_date.months,current_date.days,current_date.hours,current_date.minutes,current_date.seconds));}
									break;
						default: 		
									break;
						}
						break;

			case SDL_SCANCODE_M : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "external_viewer filename "+CDIR2+"audio/"+a_atab[0]+" action play");
									break;
						case SUPER: 
									app->commander->execute_command( "external_viewer filename "+CDIR2+"audio/"+a_atab[4]+" action play");
									key_Modifier= NONE; break;
						case KWIN:  		
									break;
						case CTRL : 
									app->commander->execute_command( "external_viewer filename "+CDIR2+"audio/"+a_atab[8]+" action play");
						default: 		
									break;
						}
						break;

			case SDL_SCANCODE_N : 
						switch(key_Modifier) {
						case NONE:
									if ( system("killall vlc 2>/dev/null")) printf("error stop vlc music\n");
									app->audio->haltMusic();
									break;
						case SUPER:
									app->scripts->play_script(CDIR2+"internal/chut.sts", CDIR2+"internal/");
									key_Modifier= NONE; break;
						case KWIN:
									app->scripts->play_script(CDIR2+"internal/navigation.sts", CDIR2+"internal/");
									break;
						case CTRL:
									break;
						default: 		
									break;
						}
						break;

			case SDL_SCANCODE_O : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "date sidereal 1");
									break;
						case SUPER: 		
									key_Modifier= NONE; break;
						case KWIN:		
									break;
						case SHIFT:
									{double jd = core->getJDay();
									ln_date current_date;
									NShadeDateTime::JulianToDate(jd,&current_date);
									core->setJDay(NShadeDateTime::JulianDayFromDateTime(current_date.years,current_date.months+1,current_date.days,current_date.hours,current_date.minutes,current_date.seconds));}
									break;
						case CTRL : 
									app->commander->execute_command( "date relative 1");
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_P : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "date sidereal 7");
									break;
						case SUPER:		
									key_Modifier= NONE; break;
						case SHIFT:
									{double jd = core->getJDay();
									ln_date current_date;
									NShadeDateTime::JulianToDate(jd,&current_date);
									core->setJDay(NShadeDateTime::JulianDayFromDateTime(current_date.years+1,current_date.months,current_date.days,current_date.hours,current_date.minutes,current_date.seconds));}
									break;
						case KWIN:	
									get_a_tab_parser();
									break;
						case CTRL : 
									app->commander->execute_command( "date relative 7");
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_Q : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "flag constellation_drawing toggle");
									break;
						case SUPER:	
									app->commander->execute_command("set sky_culture western-asterisms");
									key_Modifier= NONE; break;
						case KWIN:	
									break;
						case CTRL : 
									app->scripts->play_script(CDIR2+"internal/sky_culture1.sts", CDIR2+"internal/");
									break;
						default: 		
									break;
						}
						break;


			case SDL_SCANCODE_R: 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "flag constellation_boundaries toggle");
									break;
						case SUPER:
									app->commander->execute_command("flag zodiac toggle");
									key_Modifier= NONE; break;
						case KWIN:
									if (app->scripts->is_recording()) {
									  app->commander->execute_command( "script action cancelrecord");
									  show_message(_("Command recording stopped."), 3000);
									} else {
									  app->commander->execute_command( "script action record");
									  if (app->scripts->is_recording()) {
										show_message(string( _("Recording commands to script file:\n")
										                      + app->scripts->get_record_filename() + "\n\n"
										                      + _("Hit it again to stop.\n")), 4000);
									  } else {
										show_message(_("Error: Unable to open script file to record commands."), 3000);
									  }
									}
									break;
						case CTRL : 
									app->scripts->play_script(CDIR2+"internal/sky_culture4.sts", CDIR2+"internal/");
									break;
						default: 		
									break;
						}
						break;

			case SDL_SCANCODE_S: 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "flag ecliptic_line toggle");
									break;
						case SUPER:	
									app->commander->execute_command( "flag precession_circle toggle");
									key_Modifier= NONE; break;
						case KWIN:	
									{string tempName = app->getNextScreenshotFilename();
									app->writeScreenshot(tempName);}
									break;
						case SHIFT:
									app->commander->execute_command( "flag planets_orbits toggle");
									break;
						case CTRL : 
									app->scripts->play_script(CDIR2+"internal/ecliptic_poles.sts", CDIR2+"internal/");
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_T : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command("flag object_trails toggle");
									break;
						case SUPER:
									app->commander->execute_command( "flag suntrace toggle");
									key_Modifier= NONE; break;
						case KWIN:
  									app->commander->execute_command( "suntrace action clear");
									break;
						case SHIFT:
									app->commander->execute_command( "suntrace pen toggle");
									break;
						case CTRL : 
									app->scripts->play_script(CDIR2+"internal/trace.sts", CDIR2+"internal/");
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_U : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "date sidereal -7");
									key_Modifier= NONE; break;
									break;
						case SUPER:	
									key_Modifier= NONE; break;
						case SHIFT:
									{double jd = core->getJDay();
									ln_date current_date;
									NShadeDateTime::JulianToDate(jd,&current_date);
									core->setJDay(NShadeDateTime::JulianDayFromDateTime(current_date.years-1,current_date.months,current_date.days,current_date.hours,current_date.minutes,current_date.seconds));}
									break;
						case KWIN:	
									break;
						case CTRL : 
									app->commander->execute_command( "date relative -7");
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_V : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "flag show_tui_datetime toggle");
									// keep these in sync.  Maybe this should just be one flag.
									//if(FlagShowTuiDateTime) app->commander->execute_command( "flag show_tui_short_obj_info on");
									//else app->commander->execute_command( "flag show_tui_short_obj_info off");
									break;
						case SUPER:
									app->commander->execute_command( "flag show_tui_short_obj_info toggle");
									key_Modifier= NONE; break;
						case CTRL : 
									app->commander->execute_command( "flag show_latlon toggle");
									break;
						case KWIN:	
									break;
						default: 	
									break;
						}
						break;
            

			case SDL_SCANCODE_W : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "flag constellation_names toggle");
									break;
						case SUPER:
									app->commander->execute_command("flag zenith_line toggle");
									key_Modifier= NONE; break;
						case KWIN:	
									break;
						case CTRL : 
									app->scripts->play_script(CDIR2+"internal/sky_culture2.sts", CDIR2+"internal/");
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_X : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "flag meridian_line toggle");
									break;
						case SUPER:
									app->commander->execute_command("flag azimuthal_grid toggle");
									key_Modifier= NONE; break;
						case KWIN:	
									break;
						case SHIFT : 
									app->commander->execute_command("flag planets_axis toggle");
									break;
						case CTRL : 
									app->scripts->play_script(CDIR2+"internal/mire.sts", CDIR2+"internal/");
									break;
						default: 	
									break;
						}
						break;


			case SDL_SCANCODE_Y : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command("flag analemma_line toggle");
									break;
						case SUPER:
									app->commander->execute_command("flag galactic_center toggle");
									key_Modifier= NONE; break;
						case KWIN:	
									break;

						case CTRL : 
									app->commander->execute_command("select planet home_planet pointer off");
									app->commander->execute_command("flag track_object on");
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_Z : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "flag atmosphere toggle");
									break;
						case SUPER:	
									app->commander->execute_command( "flag landscape toggle");
 									key_Modifier= NONE; break;  						
						case SHIFT : 
									app->scripts->play_script(CDIR2+"internal/panorama2.sts", CDIR2+"internal/");
									//current_landscape = core->getObservatory()->get_landscape_name();
									break;
						case KWIN:	
									break;
						case CTRL : 
									app->commander->execute_command( "flag clouds toggle");
									app->scripts->play_script(CDIR2+"internal/panorama4.sts", CDIR2+"internal/");
									break;
						default:
									break;
						}
						break;

			case SDL_SCANCODE_0 : //certified
						switch(key_Modifier) {
						case NONE:
									//~ latimem = core->getObservatory()->get_latitude();
									//~ latimem = latimem + 45;
									//~ oss << "moveto lat " << latimem << " duration 7";
									//~ app->commander->execute_command(oss.str());
									core->moveRelLatObserver(45,7000);  //latitude , duration
									break;
						case SUPER:
									app->commander->execute_command("moveto lat 90 duration 5");									
									key_Modifier= NONE; break;
						case KWIN:	
									break;
						case SHIFT : 
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[27], CDIR2+"fscripts/"); 
									break;
						case CTRL : 
									//~ latimem = core->getObservatory()->get_latitude();
									//~ latimem = latimem + 30;
									//~ oss << "moveto lat " << latimem << " duration 7";
									//~ app->commander->execute_command(oss.str());
									core->moveRelLatObserver(30,5000);  //latitude , duration
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_1 : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command("flag star_names toggle");
									break;
						case SUPER:
									app->scripts->play_script(CDIR2+"internal/white_room.sts", CDIR2+"internal/");
									key_Modifier= NONE; break;
						case KWIN:	
									break;
						case CTRL:	
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[12], CDIR2+"fscripts/"); 
									break;
						case SHIFT : 
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[18], CDIR2+"fscripts/");
									break;
						default: 	
									break;
						}
						break;


			case SDL_SCANCODE_2 : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command("flag planet_names toggle");
									break;
						case SUPER:	
									app->commander->execute_command("flag planet_orbits toggle");
									key_Modifier= NONE; break;
						case KWIN:		break;
						case CTRL:	
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[13], CDIR2+"fscripts/"); 
									break;
						case SHIFT : 
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[19], CDIR2+"fscripts/");
									break;
						default: 		
									break;
						}
						break;


			case SDL_SCANCODE_3 : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "flag nebula_hints toggle");
									break;
						case SUPER:
									app->scripts->play_script(CDIR2+"internal/deepsky_drawings.sts", CDIR2+"internal/");
									key_Modifier= NONE; break;
						case KWIN:		
									app->commander->execute_command( "flag dso_pictograms toggle");
									break;
						case CTRL:
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[14], CDIR2+"fscripts/"); 
									break;
						case SHIFT : 
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[20], CDIR2+"fscripts/");
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_4 : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "flag fog toggle");
									break;
						case SUPER:
									app->scripts->play_script(CDIR2+"internal/orange_fog.sts", CDIR2+"internal/");
									key_Modifier= NONE; break;
						case KWIN:	
									break;
						case CTRL:
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[15], CDIR2+"fscripts/"); 
									break;
						case SHIFT : 
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[21], CDIR2+"fscripts/");
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_5 : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "flag planets toggle");
									break;
						case SUPER:	
									app->commander->execute_command( "body action clear");
									key_Modifier= NONE; break;
						case KWIN:	
									break;
						case CTRL:
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[16], CDIR2+"fscripts/"); 
									break;
						case SHIFT : 
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[22], CDIR2+"fscripts/");
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_6 : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "flag stars toggle");
									break;
						case SUPER:	
									app->commander->execute_command("deselect");
									key_Modifier= NONE; break;
						case KWIN:	
									break;
						case CTRL:
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[17], CDIR2+"fscripts/"); 
									break;
						case SHIFT : 
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[23], CDIR2+"fscripts/");
									break;
						default: 	
									break;
						}
						break;
            
			case SDL_SCANCODE_7 : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "flag milky_way toggle");
									break;
						case SUPER:
									app->scripts->play_script(CDIR2+"internal/milkyway.sts", CDIR2+"internal/");
									key_Modifier= NONE; break;
						case KWIN:	
									break;
						case CTRL:
									break;
						case SHIFT : 
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[24], CDIR2+"fscripts/");
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_8 : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command("flag nebulae toggle");
									break;
						case SUPER:	
									app->commander->execute_command( "nebula action clear");
									key_Modifier= NONE; break;
						case KWIN:
									break;
						case CTRL:
									app->commander->execute_command( "flag nebula_names toggle");
									break;
						case SHIFT: 
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[25], CDIR2+"fscripts/");
									break;
						default:
									break;
						}
						break;

			case SDL_SCANCODE_9 : 
						switch(key_Modifier) {
						case NONE:
									//~ latimem = core->getObservatory()->get_latitude();
									//~ latimem = latimem - 45;
									//~ oss << "moveto lat " << latimem << " duration 7";
									//~ app->commander->execute_command(oss.str());
									core->moveRelLatObserver(-45,7000);  //latitude , duration
									break;
						case SUPER:
									app->commander->execute_command("moveto lat -90 duration 5");
									key_Modifier= NONE; break;
						case KWIN:	
									break;
						case SHIFT : 
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[26], CDIR2+"fscripts/");
									break;
						case CTRL : 
									//~ latimem = core->getObservatory()->get_latitude();
									//~ latimem = latimem - 30;
									//~ oss << "moveto lat " << latimem << " duration 7";
									//~ app->commander->execute_command(oss.str());
									core->moveRelLatObserver(-30,5000);  //latitude , duration
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_MINUS : 
						switch(key_Modifier) {
						case NONE:
									if (core->getFlagManualAutoZoom()) app->commander->execute_command( "zoom auto out manual 1");
									else app->commander->execute_command( "zoom auto initial");
									break;
						case SUPER:
									app->scripts->play_script(CDIR2+"internal/zoom.sts", CDIR2+"internal/");
									key_Modifier= NONE; break;
						case KWIN:	
									break;
						case SHIFT : 
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[28], CDIR2+"fscripts/");
									break;
						case CTRL : 
									app->commander->execute_command( "zoom auto out");
									app->commander->execute_command( "zoom fov 60 duration 5");
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_EQUALS : 
						switch(key_Modifier) {
						case NONE:
									if(core->getFlagManualAutoZoom()) app->commander->execute_command( "zoom auto in manual 1");
									else app->commander->execute_command( "zoom auto in");
									break;
						case SUPER:
									app->commander->execute_command("moveto lat default lon default alt default");
									key_Modifier= NONE; break;
						case KWIN:	
									break;
						case SHIFT : 
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[29], CDIR2+"fscripts/");
									break;
						case CTRL : 
									app->commander->execute_command( "zoom auto in");
									app->commander->execute_command( "zoom fov 1 duration 5");
									break;
						default: 	
									break;
						}
						break;


			case SDL_SCANCODE_SPACE :	
						switch(key_Modifier) {
						case NONE:
									if ( app->scripts->is_playing() ) {
										app->commander->execute_command( "script action pause");
										app->time_multiplier = 1;}
									else
										app->commander->execute_command( "timerate action pause");
									break;
						case SUPER:		
									key_Modifier= NONE; break;
						case KWIN:	
									break;
						case CTRL :
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_NONUSBACKSLASH : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "flag landscape toggle");
									break;
						case SUPER:
									app->scripts->play_script(CDIR2+"internal/panorama1.sts", CDIR2+"internal/");
									current_landscape = core->getObservatory()->get_landscape_name();
									key_Modifier= NONE; break;
						case KWIN:	
									break;
						case SHIFT : 
									app->scripts->play_script(CDIR2+"internal/panorama3.sts", CDIR2+"internal/");
									current_landscape = core->getObservatory()->get_landscape_name();
									break;
						case CTRL : 
									core->setLandscape(default_landscape);
									current_landscape = core->getObservatory()->get_landscape_name();
									flag_compass = 0;
									flag_triangle = 0;
									break;
						default: 	
									break;
						}
						break;

                
			case SDL_SCANCODE_LEFTBRACKET : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( string("flag track_object toggle"));
									break;
						case SUPER:	
									key_Modifier= NONE; break;
						case KWIN:	
									break;
						case SHIFT:	
									app->commander->execute_command("position action save"); 
									break;
						case CTRL :
									{string planet = core->getSelectedPlanetEnglishName();
									if(planet!="") app->commander->execute_command( string("set home_planet ") + planet + string(" duration 10"));} 
									break;
						default: 	
									break;
						}
						break;


			case  SDL_SCANCODE_RIGHTBRACKET : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command("date load preset");
									break;
						case SUPER:
									app->init();
									app->scripts->play_script(CDIR2+"internal/initial.sts", CDIR2+"internal/"); 
									key_Modifier= NONE; break;
						case SHIFT: 
									app->commander->execute_command("date load current");
									app->commander->execute_command("date sun set"); 
									app->commander->execute_command("date relative 0.07"); 
								        break;
						case KWIN:	
									break;
						case CTRL:
									{double jd = core->getJDay();
									ln_date current_date,temps;
									NShadeDateTime::JulianToDate(jd,&current_date);
									temps=current_date;
									//temps.hours=current_date.hours;
									//temps.minutes=current_date.minutes;
									//temps.seconds=current_date.seconds;
									app->commander->execute_command("date load preset");
									jd = core->getJDay();
									NShadeDateTime::JulianToDate(jd,&current_date);
									current_date.hours=temps.hours;
									current_date.minutes=temps.minutes;
									current_date.seconds=temps.seconds;
									core->setJDay(NShadeDateTime::JulianDayFromDateTime(current_date.years,current_date.months,current_date.days,current_date.hours,current_date.minutes,current_date.seconds));}
									break;
						default: 	
									break;
						}
						break;


			case  SDL_SCANCODE_APOSTROPHE : 
						switch(key_Modifier) {
						case NONE:
									core->setFlagLockSkyPosition(!core->getFlagLockSkyPosition());
									break;
						case SHIFT: 
									app->commander->execute_command("position action save"); 
									break;
						case SUPER:
									app->scripts->play_script(CDIR2+"internal/clear_mess.sts", CDIR2+"internal/"); 
									key_Modifier= NONE; break; 
						case KWIN:
									break;
						case CTRL : 	
									break;
						default:
									break;
						}
						break;	

			case SDL_SCANCODE_ESCAPE:
					                // RFE 1310384, ESC closes dialogs
					                // close search mode
					                FlagSearch=false;
					                search_win->setVisible(FlagSearch);
					
					                // close config dialog
					                FlagConfig = false;
					                config_win->setVisible(FlagConfig);
					
					                // close help dialog
					                FlagHelp = false;
					                help_win->setVisible(FlagHelp);
					
					                // close information dialog
					                FlagInfos = false;
					                licence_win->setVisible(FlagInfos);
					                // END RFE 1310384
							break;

			case SDL_SCANCODE_INSERT:
									app->commander->execute_command("position action save");
									break;
			case SDL_SCANCODE_DELETE:
									app->commander->execute_command("position action load");
									break;
			case SDL_SCANCODE_HOME:
									app->init();
									app->scripts->play_script(CDIR2+"internal/initial.sts", CDIR2+"internal/"); 
									break;
			case SDL_SCANCODE_END:
									app->scripts->play_script(CDIR2+"internal/initial_night.sts", CDIR2+"internal/"); 
									break;
			case SDL_SCANCODE_RETURN:
									app->commander->execute_command("deselect");
									break;
			case SDL_SCANCODE_KP_1 : 
						switch(key_Modifier) {
						case NONE:
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[29], CDIR2+"fscripts/");
									break;
						case SUPER:	
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[39], CDIR2+"fscripts/");
									key_Modifier= NONE; break;
						case SHIFT:	
									core->moveHeadingRelative(-0.2);
									//app->commander->execute_command( string("set heading -180 duration 10"));
									break;
						case KWIN:	
									break;
						case CTRL: 
									core->moveHeadingRelative(-1);									
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_KP_2 : 
						switch(key_Modifier) {
						case NONE:
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[30], CDIR2+"fscripts/");
									break;
						case SUPER:	
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[40], CDIR2+"fscripts/");
									key_Modifier= NONE; break;
						case KWIN:
									break;
						case SHIFT:
									core->moveRelLatObserver(-0.5,0.1);  //latitude , duration
									break;
						case CTRL:
									break;
						default:
									break;
						}
						break;

			case SDL_SCANCODE_KP_3 : 
						switch(key_Modifier) {
						case NONE:
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[31], CDIR2+"fscripts/");
									break;
						case SUPER:	
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[41], CDIR2+"fscripts/");
									key_Modifier= NONE; break;
						case SHIFT:	
									latimem = core->getObservatory()->get_altitude();
									latimem = latimem/2;
									oss << "moveto altitude " << latimem << " duration 1";
									app->commander->execute_command(oss.str());        
									break;
						case KWIN:	
									break;
						case CTRL: 	
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_KP_4 : 
						switch(key_Modifier) {
						case NONE:
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[32], CDIR2+"fscripts/");
									break;
						case SUPER:	
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[42], CDIR2+"fscripts/");
									key_Modifier= NONE; break;
						case SHIFT:	
									core->moveRelLonObserver(-0.5,0.1);  //longitude , duration
									break;
						case KWIN:	
									break;
						case CTRL: 	
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_KP_5 : 
						switch(key_Modifier) {
						case NONE:
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[33], CDIR2+"fscripts/");
									break;
						case SUPER:	
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[43], CDIR2+"fscripts/");
									key_Modifier= NONE; break;
						case KWIN:
									break;
						case SHIFT:		
									{string newplanet = core->getSelectedPlanetEnglishName();
									if (newplanet!="") app->commander->execute_command( string("set home_planet ") + newplanet);}
									break;
						case CTRL:
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_KP_6 : 
						switch(key_Modifier) {
						case NONE:
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[34], CDIR2+"fscripts/");
									break;
						case SUPER:	
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[44], CDIR2+"fscripts/");
									key_Modifier= NONE; break;
						case SHIFT:	
									core->moveRelLonObserver(0.5,0.1);  //longitude , duration
									break;
						case KWIN:
									break;
						case CTRL:
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_KP_7 : 
						switch(key_Modifier) {
						case NONE:
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[35], CDIR2+"fscripts/");
									break;
						case SUPER:	
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[45], CDIR2+"fscripts/");
									key_Modifier= NONE; break;
						case SHIFT:	
									core->moveHeadingRelative(0.2);
									//app->commander->execute_command( string("set heading 180 duration 10"));
									break;
						case KWIN:	
									break;
						case CTRL : 
									core->moveHeadingRelative(1);									
									break;
						default:
									break;
						}
						break;

			case SDL_SCANCODE_KP_8 : 
						switch(key_Modifier) {
						case NONE:
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[36], CDIR2+"fscripts/");
									break;
						case SUPER:	
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[46], CDIR2+"fscripts/");
									key_Modifier= NONE; break;
						case SHIFT:	
									core->moveRelLatObserver(0.5,0.1);  //latitude , duration
									break;
						case KWIN:
									break;
						case CTRL : 
									break;
						default:
									break;
						}
						break;

			case SDL_SCANCODE_KP_9 : 
						switch(key_Modifier) {
						case NONE:
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[37], CDIR2+"fscripts/");
									break;
						case SUPER:	
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[47], CDIR2+"fscripts/");
									key_Modifier= NONE; break;
						case SHIFT:	
									latimem = core->getObservatory()->get_altitude();
									latimem = latimem*2;
									oss << "moveto altitude " << latimem << " duration 1";
									app->commander->execute_command(oss.str());        
									break;
						case KWIN:	
									app->commander->execute_command( "flag galactic_grid toggle");  
									break;
						case CTRL : 
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_KP_0 : 
						switch(key_Modifier) {
						case NONE:
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[28], CDIR2+"fscripts/");
									break;
						case SUPER:
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[38], CDIR2+"fscripts/");
									key_Modifier= NONE; break;
						case SHIFT:
									app->commander->execute_command( string("flag track_object toggle"));
									break;
						case KWIN:
									break;
						case CTRL:
									break;
						default:
									break;
						}
						break;

			case SDL_SCANCODE_KP_PERIOD : 
						switch(key_Modifier) {
						case NONE:
									app->scripts->play_script(CDIR2+"fscripts/M20.sts", CDIR2+"fscripts/");
									break;
						case SUPER:	
									app->scripts->play_script(CDIR2+"fscripts/M21.sts", CDIR2+"fscripts/");
									key_Modifier= NONE; break;
						case SHIFT:
									{ string coreplanet = core->getSelectedPlanetEnglishName();
									  if (coreplanet!="") {
									    app->commander->execute_command( string("set home_planet ") + coreplanet);
									    app->commander->execute_command( string("select planet ')+coreplanet+string(' pointer off"));
									    app->commander->execute_command( string("wait duration 0.01"));
									    app->commander->execute_command( string("body name _ action drop"));
									    app->commander->execute_command( string("body action load name _ parent Sun radius 1 oblateness 0.0 albedo 0.0 lighting false halo false color 1.0,1.0,1.0 rot_periode 1000000000 tex_map none.png coord_func ')+coreplanet+string('_special"));
									    app->commander->execute_command( string("set home_planet _ duration 0"));
									    app->commander->execute_command( string("select planet _ pointer off"));
									    app->commander->execute_command( string("flag track_object on"));
									  }
									} 
									break;
						case KWIN:
									break;
						case CTRL : 
									break;
						default:
									break;
						}
						break;

			case SDL_SCANCODE_KP_PLUS : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( string("set home_planet Earth"));        
									app->commander->execute_command( string("moveto lat default lon default alt default"));        
									app->commander->execute_command( string("zoom auto out manual 1"));        
									break;
						case SUPER:	
									latimem = core->getObservatory()->get_latitude();
									latimem = -latimem;
									oss << "moveto lat " << latimem;
									latimem = core->getObservatory()->get_longitude();
									latimem = 180+latimem;
									oss << " lon " << latimem << " duration 0";
									app->commander->execute_command(oss.str());
									if (!antipodes) {
									    app->scripts->play_script(CDIR2+"internal/antipodes.sts", CDIR2+"internal/");
									} else {
									    core->setLandscape(current_landscape);
									}
									antipodes = !antipodes;
									key_Modifier= NONE; break;
						case KWIN:
									break;
						case CTRL : 
									latimem = core->getObservatory()->get_latitude();
									latimem = -latimem;
									oss << "moveto lat " << latimem << " duration 0";
									app->commander->execute_command(oss.str());        
									break;
						default:
									break;
						}
						break;

			case SDL_SCANCODE_KP_MINUS : 
						switch(key_Modifier) {
						case NONE:
									SDL_ShowCursor(1);
									SDL_WarpMouseInWindow( core->get_Window(), core->getDisplayWidth() /2, core->getDisplayHeight() /2);
									break;
						case SUPER:
									SDL_ShowCursor(0);
									SDL_WarpMouseInWindow( core->get_Window(), core->getDisplayWidth() /2, core->getDisplayHeight());
									key_Modifier= NONE; break;
						case KWIN:	
									break;
						case CTRL : 
									FlagConfig=!FlagConfig;
									config_win->setVisible(FlagConfig);
									break;
						default:
									break;
						}
						break;

			case SDL_SCANCODE_KP_MULTIPLY : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command("audio volume increment");
									break;
						case SUPER:	
									key_Modifier= NONE; break;
						case KWIN:
									break;
						case CTRL: 
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_KP_DIVIDE : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command("audio volume decrement");
									break;
						case SUPER:
									key_Modifier= NONE; break;
						case KWIN:	
									break;
						case CTRL : 
									break;
						default: 	
									break;
						}
						break;


			case  SDL_SCANCODE_F1:  
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "external_viewer filename "+CDIR2+"videos/"+a_vtab[0]+" action play");
									break;
						case SUPER:
									app->commander->execute_command("external_viewer filename "+CDIR2+"videos/"+a_vtab[12]+" action play");
									key_Modifier= NONE; break;
						case SHIFT:
									app->commander->execute_command("external_viewer filename "+CDIR2+"videos/"+a_vtab[24]+" action play");
									break;
						case CTRL :
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[0], CDIR2+"fscripts/");
									break;
						default: 	
									break;
						}
						break;


			case SDL_SCANCODE_F2 :
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "external_viewer filename "+CDIR2+"videos/"+a_vtab[1]+" action play");
									break;
						case SUPER:
									app->commander->execute_command("external_viewer filename "+CDIR2+"videos/"+a_vtab[13]+" action play");
									key_Modifier= NONE; break;
						case SHIFT:
									app->commander->execute_command("external_viewer filename "+CDIR2+"videos/"+a_vtab[25]+" action play");
									break;
						case CTRL :
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[1], CDIR2+"fscripts/");
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_F3 : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "external_viewer filename "+CDIR2+"videos/"+a_vtab[2]+" action play");
									break;
						case SUPER:
									app->commander->execute_command("external_viewer filename "+CDIR2+"videos/"+a_vtab[14]+" action play");
									key_Modifier= NONE; break;
						case SHIFT:
									app->commander->execute_command("external_viewer filename "+CDIR2+"videos/"+a_vtab[26]+" action play");
									break;
						case CTRL :
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[2], CDIR2+"fscripts/");
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_F4:
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "external_viewer filename "+CDIR2+"videos/"+a_vtab[3]+" action play");
									break;
						case SUPER:
									app->commander->execute_command("external_viewer filename "+CDIR2+"videos/"+a_vtab[15]+" action play");
									key_Modifier= NONE; break;
						case SHIFT:
									app->commander->execute_command("external_viewer filename "+CDIR2+"videos/"+a_vtab[27]+" action play");
									break;
						case CTRL :
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[3], CDIR2+"fscripts/");
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_F5: 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "external_viewer filename "+CDIR2+"videos/"+a_vtab[4]+" action play");
									break;
						case SUPER:
									app->commander->execute_command("external_viewer filename "+CDIR2+"videos/"+a_vtab[16]+" action play");
									key_Modifier= NONE; break;
						case SHIFT:
									app->commander->execute_command("external_viewer filename "+CDIR2+"videos/"+a_vtab[28]+" action play");
									break;
						case CTRL :
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[4], CDIR2+"fscripts/");
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_F6 : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "external_viewer filename "+CDIR2+"videos/"+a_vtab[5]+" action play");
									break;
						case SUPER:
									app->commander->execute_command("external_viewer filename "+CDIR2+"videos/"+a_vtab[17]+" action play");
									key_Modifier= NONE; break;
						case SHIFT:
									app->commander->execute_command("external_viewer filename "+CDIR2+"videos/"+a_vtab[29]+" action play");
									break;
						case CTRL :
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[5], CDIR2+"fscripts/");
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_F7 :
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "external_viewer filename "+CDIR2+"videos/"+a_vtab[6]+" action play");
									break;
						case SUPER:
									app->commander->execute_command("external_viewer filename "+CDIR2+"videos/"+a_vtab[18]+" action play");
									key_Modifier= NONE; break;
						case SHIFT:
									app->commander->execute_command("external_viewer filename "+CDIR2+"videos/"+a_vtab[30]+" action play");
									break;
						case CTRL :
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[6], CDIR2+"fscripts/");
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_F8 : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "external_viewer filename "+CDIR2+"videos/"+a_vtab[7]+" action play");
									break;
						case SUPER:
									app->commander->execute_command("external_viewer filename "+CDIR2+"videos/"+a_vtab[19]+" action play");
									key_Modifier= NONE; break;
						case SHIFT:
									app->commander->execute_command("external_viewer filename "+CDIR2+"videos/"+a_vtab[31]+" action play");
									break;
						case CTRL :
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[7], CDIR2+"fscripts/");
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_F9 : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "external_viewer filename "+CDIR2+"videos/"+a_vtab[8]+" action play");
									break;
						case SUPER:
									app->commander->execute_command("external_viewer filename "+CDIR2+"videos/"+a_vtab[20]+" action play");
									key_Modifier= NONE; break;
						case SHIFT:
									app->commander->execute_command("external_viewer filename "+CDIR2+"videos/"+a_vtab[32]+" action play");
									break;
						case CTRL :
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[8], CDIR2+"fscripts/");
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_F10 : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "external_viewer filename "+CDIR2+"videos/"+a_vtab[9]+" action play");
									break;
						case SUPER:
									app->commander->execute_command("external_viewer filename "+CDIR2+"videos/"+a_vtab[21]+" action play");
									key_Modifier= NONE; break;
						case SHIFT:
									app->commander->execute_command("external_viewer filename "+CDIR2+"videos/"+a_vtab[33]+" action play");
									break;
						case CTRL :
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[9], CDIR2+"fscripts/");
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_F11 : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "external_viewer filename "+CDIR2+"videos/"+a_vtab[10]+" action play");
									break;
						case SUPER:
									app->commander->execute_command("external_viewer filename "+CDIR2+"videos/"+a_vtab[22]+" action play");
									key_Modifier= NONE; break;
						case SHIFT:
									if (core->getFlagNav()) {
									                if (!app->scripts->is_playing()) {
									                    cout << flag_compass;
									                    cout <<  " -> " << flag_compass << endl;
									                    string s;
									                    stringstream out;
									                    out << flag_compass+1;
									                    s = out.str();
									                    if (flag_compass == 4) {
									                        core->setLandscape(current_landscape);
									                    }
									                    else app->scripts->play_script(CDIR2+"internal/windrose/0"+s+".sts",CDIR2+"internal/windrose/");
									                    flag_compass = (flag_compass+1)%5;
									                    flag_triangle = 0;
									                    flag_f9 = 0;
									                }
									} else app->commander->execute_command("external_viewer filename "+CDIR2+"videos/"+a_vtab[34]+" action play");
									break;
						case CTRL :
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[10], CDIR2+"fscripts/");
									break;
						default: 	
									break;
						}
						break;

			case SDL_SCANCODE_F12 : 
						switch(key_Modifier) {
						case NONE:
									app->commander->execute_command( "external_viewer filename "+CDIR2+"videos/"+a_vtab[11]+" action play");
									break;
						case SUPER:
									app->commander->execute_command("external_viewer filename "+CDIR2+"videos/"+a_vtab[23]+" action play");
									key_Modifier= NONE; break;
						case SHIFT:
									app->commander->execute_command("external_viewer filename "+CDIR2+"videos/"+a_vtab[35]+" action play");
									break;
						case CTRL :
									app->scripts->play_script(CDIR2+"fscripts/"+a_stab[11], CDIR2+"fscripts/");
									break;
						default: 	
									break;
						}
						break;

			retVal=1;
            
			default: break;
			//~ }
	}  // end S_GUI_PRESSED
}
    if (state==S_GUI_RELEASED)
    {
		switch (key)
			{
			case SDL_SCANCODE_RCTRL : 
			case SDL_SCANCODE_LCTRL : 
									KeyTimeLeft=0;
									key_Modifier=NONE;
									break;
			case SDL_SCANCODE_RSHIFT :			
			case SDL_SCANCODE_LSHIFT :
									KeyTimeLeft=0;
									key_Modifier=NONE;
			case SDL_SCANCODE_LGUI :
			case SDL_SCANCODE_RGUI :
									KeyTimeLeft=0;
									key_Modifier=NONE;
									break;

			default:				break;
			}
	}
    
    return retVal;
}

// Update changing values
void UI::gui_update_widgets(int delta_time)
{
	updateTopBar();

	// handle mouse cursor timeout
	if (MouseCursorTimeout > 0) {
		if (MouseTimeLeft > delta_time) MouseTimeLeft -= delta_time;
		else {
			// hide cursor
			MouseTimeLeft = 0;
			SDL_ShowCursor(0);
		}
	}

	// handle key_Modifier cursor timeout
	if (key_Modifier != NONE) {
		if (KeyTimeLeft > delta_time) KeyTimeLeft -= delta_time;
		else {
			// done
			//	cout << "KWIN timed out\n";
			key_Modifier = NONE;
		}
	}

	// update message win
	message_win->update(delta_time);


	if (FlagShowSelectedObjectInfo && core->getFlagHasSelected()) {
		info_select_ctr->setVisible(true);
		updateInfoSelectString();
	} else
		info_select_ctr->setVisible(false);

	bt_flag_ctr->setVisible(FlagMenu);
	bt_time_control_ctr->setVisible(FlagMenu);

	bt_flag_constellation_draw->setState(core->getFlagConstellationLines());
	bt_flag_constellation_name->setState(core->getFlagConstellationNames());
	bt_flag_constellation_art->setState(core->getFlagConstellationArt());
	bt_flag_azimuth_grid->setState(core->getFlagAzimutalGrid());
	bt_flag_equator_grid->setState(core->getFlagEquatorGrid());
	bt_flag_ground->setState(core->getFlagLandscape());
	bt_flag_cardinals->setState(core->getFlagCardinalsPoints());
	bt_flag_atmosphere->setState(core->getFlagAtmosphere());
	bt_flag_nebula_name->setState(core->getFlagNebulaHints());
	bt_flag_help->setState(help_win->getVisible());
	bt_flag_equatorial_mode->setState(core->getMountMode()==Core::MOUNT_EQUATORIAL);
	bt_flag_config->setState(config_win->getVisible());
//	bt_flag_chart->setState(app->getVisionModeChart());
	bt_flag_planet->setState(core->getFlagPlanetsHints());
	bt_flag_search->setState(search_win->getVisible());
	bt_flag_goto->setState(false);
	if (bt_flip_horz) bt_flip_horz->setState(core->getFlipHorz());
	if (bt_flip_vert) bt_flip_vert->setState(core->getFlipVert());

	bt_real_time_speed->setState(fabs(core->getTimeSpeed()-JD_SECOND)<0.000001);
	bt_inc_time_speed->setState((core->getTimeSpeed()-JD_SECOND)>0.0001);
	bt_dec_time_speed->setState((core->getTimeSpeed()-JD_SECOND)<-0.0001);
	// cache last time to prevent to much slow system call
	static double lastJD = 0;
	if (fabs(lastJD-core->getJDay())>JD_SECOND/4) {
		bt_time_now->setState(fabs(core->getJDay()-NShadeDateTime::JulianFromSys())<JD_SECOND);
		lastJD = core->getJDay();
	}
	if (config_win->getVisible()) updateConfigForm();
}

// Update the infos about the selected object in the TextLabel widget
void UI::updateInfoSelectString(void)
{
	//~ if (app->getVisionModeNight()) {
		//~ info_select_txtlbl->setTextColor(Vec3f(1.0,0.2,0.2));
	//~ } else {
		info_select_txtlbl->setTextColor(core->getSelectedObjectInfoColor());
	//~ }
	info_select_txtlbl->setLabel(core->getSelectedObjectInfo());
}

void UI::setTitleObservatoryName(const string& name)
{
	if (name == "")
		top_bar_appName_lbl->setLabel(APP_NAME);
	else {
	  top_bar_appName_lbl->setLabel(string(APP_NAME) + " (" + name + ")");
	}
	top_bar_appName_lbl->setPos(core->getDisplayWidth()/2-top_bar_appName_lbl->getSizex()/2,1);
}

string UI::getTitleWithAltitude(void)
{
	return core->getObservatory()->getHomePlanetNameI18n() +
	       ", " + core->getObservatory()->get_name() +
	       " @ " + Utility::doubleToString(core->getObservatory()->get_altitude()) + "m";
}

void UI::setColorScheme(const string& skinFile, const string& section)
{
	if (!desktop) return;

	InitParser conf;
	conf.load(skinFile);

	s_color GuiBaseColor		= Utility::str_to_vec3f(conf.get_str(section, "gui_base_color", "0.3,0.4,0.7"));
	s_color GuiTextColor		= Utility::str_to_vec3f(conf.get_str(section, "gui_text_color", "0.7,0.8,0.9"));

	desktop->setColorScheme(GuiBaseColor, GuiTextColor);
}

void UI::get_a_tab_parser()
{
	InitParser conf;
	string configDir= AppSettings::Instance()->getConfigDir().c_str();


	conf.load(configDir + "allocate.ini");
	Log.write("Loading " + configDir + "allocate.ini", cLog::eLOG_TYPE_INFO);

	a_atab[0]= conf.get_str("audio:audio01");
	a_atab[1]= conf.get_str("audio:audio02");
	a_atab[2]= conf.get_str("audio:audio03");
	a_atab[3]= conf.get_str("audio:audio04");
	a_atab[4]= conf.get_str("audio:audio05");
	a_atab[5]= conf.get_str("audio:audio06");
	a_atab[6]= conf.get_str("audio:audio07");
	a_atab[7]= conf.get_str("audio:audio08");
	a_atab[8]= conf.get_str("audio:audio09");
	a_atab[9]= conf.get_str("audio:audio10");
	a_atab[10]= conf.get_str("audio:audio11");
	a_atab[11]= conf.get_str("audio:audio12");

	a_vtab[0]= conf.get_str("video:video01");
	a_vtab[1]= conf.get_str("video:video02");
	a_vtab[2]= conf.get_str("video:video03");
	a_vtab[3]= conf.get_str("video:video04");
	a_vtab[4]= conf.get_str("video:video05");
	a_vtab[5]= conf.get_str("video:video06");
	a_vtab[6]= conf.get_str("video:video07");
	a_vtab[7]= conf.get_str("video:video08");
	a_vtab[8]= conf.get_str("video:video09");
	a_vtab[9]= conf.get_str("video:video10");
	a_vtab[10]= conf.get_str("video:video11");
	a_vtab[11]= conf.get_str("video:video12");
	a_vtab[12]= conf.get_str("video:video13");
	a_vtab[13]= conf.get_str("video:video14");
	a_vtab[14]= conf.get_str("video:video15");
	a_vtab[15]= conf.get_str("video:video16");
	a_vtab[16]= conf.get_str("video:video17");
	a_vtab[17]= conf.get_str("video:video18");
	a_vtab[18]= conf.get_str("video:video19");
	a_vtab[19]= conf.get_str("video:video20");
	a_vtab[20]= conf.get_str("video:video21");
	a_vtab[21]= conf.get_str("video:video22");
	a_vtab[22]= conf.get_str("video:video23");
	a_vtab[23]= conf.get_str("video:video24");

	a_stab[0]= conf.get_str("script:script01");
	a_stab[1]= conf.get_str("script:script02");
	a_stab[2]= conf.get_str("script:script03");
	a_stab[3]= conf.get_str("script:script04");
	a_stab[4]= conf.get_str("script:script05");
	a_stab[5]= conf.get_str("script:script06");
	a_stab[6]= conf.get_str("script:script07");
	a_stab[7]= conf.get_str("script:script08");
	a_stab[8]= conf.get_str("script:script09");
	a_stab[9]= conf.get_str("script:script10");
	a_stab[10]= conf.get_str("script:script11");
	a_stab[11]= conf.get_str("script:script12");
	a_stab[12]= conf.get_str("script:script13");
	a_stab[13]= conf.get_str("script:script14");
	a_stab[14]= conf.get_str("script:script15");
	a_stab[15]= conf.get_str("script:script16");
	a_stab[16]= conf.get_str("script:script17");
	a_stab[17]= conf.get_str("script:script18");
	a_stab[18]= conf.get_str("script:script19");
	a_stab[19]= conf.get_str("script:script20");
	a_stab[20]= conf.get_str("script:script21");
	a_stab[21]= conf.get_str("script:script22");
	a_stab[22]= conf.get_str("script:script23");
	a_stab[23]= conf.get_str("script:script24");
	a_stab[24]= conf.get_str("script:script25");
	a_stab[25]= conf.get_str("script:script26");
	a_stab[26]= conf.get_str("script:script27");
	a_stab[27]= conf.get_str("script:script28");
	a_stab[28]= conf.get_str("script:script29");
	a_stab[29]= conf.get_str("script:script30");
	a_stab[30]= conf.get_str("script:script31");
	a_stab[31]= conf.get_str("script:script32");
	a_stab[32]= conf.get_str("script:script33");
	a_stab[33]= conf.get_str("script:script34");
	a_stab[34]= conf.get_str("script:script35");
	a_stab[35]= conf.get_str("script:script36");
	a_stab[36]= conf.get_str("script:script37");
	a_stab[37]= conf.get_str("script:script38");
	a_stab[38]= conf.get_str("script:script39");
	a_stab[39]= conf.get_str("script:script40");
	a_stab[40]= conf.get_str("script:script41");
	a_stab[41]= conf.get_str("script:script42");
	a_stab[42]= conf.get_str("script:script43");
	a_stab[43]= conf.get_str("script:script44");
	a_stab[44]= conf.get_str("script:script45");
	a_stab[45]= conf.get_str("script:script46");
	a_stab[46]= conf.get_str("script:script47");
	a_stab[47]= conf.get_str("script:script48");
	a_stab[48]= conf.get_str("script:script49");
	a_stab[49]= conf.get_str("script:script50");
}

void UI::setFlagShowTuiMenu(const bool flag)
{

	if (flag && !FlagShowTuiMenu) {
		tuiUpdateIndependentWidgets();
	}

	FlagShowTuiMenu = flag;
}
